/***************************************************************************

  CList.c

  The List native class.

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CLIST_C

#include "common.h"
#include "error.h"
#include "array.h"

#include "Date.h"
#include "Variant.h"
#include "gambas_api.h"

#include "CList.h"


PRIVATE int (*compare_func)();


PRIVATE void clist_release(CLIST *ob)
{
  long i;

  for (i = 0; i < ARRAY_count(ob->data); i++)
    VARIANT_free(&ob->data[i]);
}


PRIVATE void clist_clear(CLIST *ob)
{
  clist_release(ob);
  ARRAY_delete(&ob->data);
  ARRAY_create_with_size(&ob->data, sizeof(VARIANT), 8);
}


PRIVATE void clist_insert(CLIST *ob, void *value, long index)
{
  VARIANT *data;

  data = (VARIANT *)ARRAY_insert(&ob->data, index);
  GB_Store(T_VARIANT, value, data);
}


BEGIN_METHOD_VOID(CLIST_new)

  CLIST *ob = OBJECT(CLIST);

  ob->type = T_VOID;

  ARRAY_create_with_size(&ob->data, sizeof(VARIANT), 8);

END_METHOD


BEGIN_METHOD_VOID(CLIST_free)

  CLIST *ob = OBJECT(CLIST);

  clist_release(ob);
  ARRAY_delete(&ob->data);

END_METHOD


BEGIN_METHOD_VOID(CLIST_clear)

  clist_clear(OBJECT(CLIST));

END_METHOD


BEGIN_PROPERTY(CLIST_count)

  GB_ReturnInt(ARRAY_count(OBJECT(CLIST)->data));

END_PROPERTY


BEGIN_METHOD(CLIST_add, GB_VARIANT value; long index)

  long index;

  if (GB_IsMissing(2))
    index = -1;
  else
    index = PARAM(index);

  clist_insert(OBJECT(CLIST), &PARAM(value), index);

END_METHOD


BEGIN_METHOD(CLIST_push, GB_VARIANT value)

  VARIANT *data;

  data = (VARIANT *)ARRAY_insert(&(OBJECT(CLIST)->data), -1);
  GB_Store(T_VARIANT, &PARAM(value), data);

END_METHOD


BEGIN_METHOD(CLIST_remove, long index)

  CLIST *ob = OBJECT(CLIST);
  long index = PARAM(index);

  if ((index < 0) || (index >= ARRAY_count(ob->data)))
    THROW(E_ARG);

  VARIANT_free(&ob->data[index]);
  ARRAY_remove(&ob->data, index);

END_METHOD


BEGIN_METHOD_VOID(CLIST_pop)

  CLIST *ob = OBJECT(CLIST);
  long index = ARRAY_count(ob->data) - 1;

  if (index < 0)
    THROW(E_ARG);

  GB_ReturnPtr(T_VARIANT, &ob->data[index]);
  VARIANT_free(&ob->data[index]);
  ARRAY_remove(&ob->data, index);

END_METHOD


BEGIN_METHOD(CLIST_get, long index)

  CLIST *ob = OBJECT(CLIST);
  long index = PARAM(index);

  if ((index < 0) || (index > ARRAY_count(ob->data)))
    THROW(E_BOUND);

  GB_ReturnPtr(T_VARIANT, &ob->data[index]);

END_METHOD


BEGIN_METHOD(CLIST_put, GB_VARIANT value; long index)

  CLIST *ob = OBJECT(CLIST);
  long index = PARAM(index);

  if ((index < 0) || (index > ARRAY_count(ob->data)))
    THROW(E_BOUND);

  GB_Store(T_VARIANT, &PARAM(value), &ob->data[index]);

END_METHOD


BEGIN_METHOD_VOID(CLIST_next)

  CLIST *ob = OBJECT(CLIST);
  long *index = (long *)GB_GetEnum();

  if (*index >= ARRAY_count(ob->data))
    GB_StopEnum();
  else
  {
    GB_ReturnPtr(T_VARIANT, &ob->data[*index]);
    (*index)++;
  }

END_METHOD


BEGIN_METHOD(CLIST_from_string, GB_STRING str; GB_STRING sep; GB_STRING esc)

  CLIST *ob = OBJECT(CLIST);
  char *str = GB_ToZeroString(PARAM(str));
  char *sep = GB_IsMissing(2) ? "," : GB_ToZeroString(PARAM(sep));
  char *esc = GB_IsMissing(3) ? "" : GB_ToZeroString(PARAM(esc));

  char c;
  char *p;
  long len;
  GB_VARIANT value;
  char *elt;
  boolean escape;

  value.type = GB_T_STRING;

  clist_clear(ob);

  p = str;
  escape = FALSE;

  for(;;)
  {
    c = *str;

    if (p == str && c != 0 && index(esc, c))
    {
      escape = TRUE;
      goto _NEXT;
    }

    if (escape)
    {
      if (c != 0 && index(esc, c) == NULL)
        goto _NEXT;
      else
        c = *(++str);
    }

    if (c != 0 && index(sep, c) == NULL)
      goto _NEXT;


    if (escape)
    {
      len = str - p - 2;
      if (len > 0)
      {
        str[-1] = 0;
        elt = p + 1;
      }
    }
    else
    {
      len = str - p;
      if (len > 0)
      {
        *str = 0;
        elt = p;
      }
    }

    value._string.value = (len <= 0) ? NULL : elt;

    clist_insert(ob, &value, -1);

    if (c == 0)
      break;

    p = str + 1;
    escape = FALSE;

  _NEXT:

    str++;
  }

  GB_ReturnInt(ARRAY_count(ob->data));

END_METHOD


int clist_compare_integer(VARIANT *a, VARIANT *b)
{
  long va, vb;

  if (a->type == T_BYTE)
    va = *((byte *)a->value);
  else if (a->type == T_SHORT)
    va = *((short *)a->value);
  else if (a->type == T_INTEGER)
    va = *((long *)a->value);
  else
    return 1;

  if (b->type == T_BYTE)
    vb = *((byte *)b->value);
  else if (b->type == T_SHORT)
    vb = *((short *)b->value);
  else if (b->type == T_INTEGER)
    vb = *((long *)b->value);
  else
    return -1;

  return (va < vb) ? -1 : ((va > vb) ? 1 : 0);
}


int clist_compare_float(VARIANT *a, VARIANT *b)
{
  double va, vb;

  if (a->type != T_FLOAT)
    return 1;

  if (b->type != T_FLOAT)
    return -1;

  va = *((double *)a->value);
  vb = *((double *)b->value);

  return (va < vb) ? -1 : ((va > vb) ? 1 : 0);
}


int clist_compare_string(VARIANT *a, VARIANT *b)
{
  char *va, *vb;

  if (a->type != T_STRING)
    return 1;

  if (b->type != T_STRING)
    return -1;

  va = *((char **)a->value);
  vb = *((char **)b->value);

  return (*compare_func)(va, vb);
}


int clist_compare_date(VARIANT *a, VARIANT *b)
{
  if (a->type != T_DATE)
    return 1;

  if (b->type != T_DATE)
    return -1;

  /* Ca marche  cause de la similitude entre VARIANT et VALUE !! */
  return DATE_comp_value((VALUE *)a, (VALUE *)b);
}


BEGIN_METHOD(CLIST_sort, long opt)

  VARIANT *array = OBJECT(CLIST)->data;
  int (*compare)();
  long opt = GB_IsMissing(1) ? 0 : PARAM(opt);

  if (ARRAY_count(array) <= 1)
    return;


  switch(array->type)
  {
    case T_STRING:
      if (opt == 1)
        compare_func = strcasecmp;
      else if (opt == 2)
        compare_func = strcoll;
      else
        compare_func = strcmp;

      compare = clist_compare_string;
      break;

    case T_BYTE: case T_SHORT: case T_INTEGER:
      compare = clist_compare_integer;
      break;

    case T_FLOAT:
      compare = clist_compare_float;
      break;

    case T_DATE:
      compare = clist_compare_date;
      break;

    default:
      THROW(E_TYPE, "String, Number or Date", TYPE_get_name(array->type));
  }

  qsort(array, ARRAY_count(array), sizeof(*array), compare);

END_METHOD



/*
BEGIN_METHOD_VOID(collection_clear)

  CCOLLECTION *col = OBJECT(CCOLLECTION);

  collection_free(NULL, col);
  HASH_TABLE_create(&col->hash_table, TYPE_sizeof(col->type));

END_METHOD


BEGIN_METHOD_VOID(collection_first)

  CCOL_ENUM *enum_state = (CCOL_ENUM *)GB_GetEnum();

  enum_state->i = (-1);
  enum_state->node = NULL;

END_METHOD


BEGIN_METHOD_VOID(collection_next)

  void *value;
  HASH_TABLE *hash_table = OBJECT(CCOLLECTION)->hash_table;
  CCOL_ENUM *enum_state = (CCOL_ENUM *)GB_GetEnum();

  value = HASH_TABLE_next(hash_table, &enum_state->i, &enum_state->node);

  if (value == NULL)
    GB_StopEnum();
  else
    GB_ReturnPtr(OBJECT(CCOLLECTION)->type, value);

END_METHOD
*/

PUBLIC GB_DESC NATIVE_List[] =
{
  GB_DECLARE("Array", sizeof(CLIST)),

  GB_METHOD("_new", NULL, CLIST_new, NULL),
  GB_METHOD("_free", NULL, CLIST_free, NULL),

  GB_PROPERTY_READ("Count", "i", CLIST_count),
  GB_PROPERTY_READ("Length", "i", CLIST_count),

  GB_METHOD("Add", NULL, CLIST_add, "(Value)v[(Index)i]"),
  GB_METHOD("Remove", NULL, CLIST_remove, "(Index)i"),
  GB_METHOD("Clear", NULL, CLIST_clear, NULL),

  GB_METHOD("FromString", "i", CLIST_from_string, "(String)s[(Separator)s(Escape)s]"),
  GB_METHOD("Sort", NULL, CLIST_sort, "[(Compare)i]"),

  GB_METHOD("_next", "v", CLIST_next, NULL),
  GB_METHOD("_get", "v", CLIST_get, "i"),
  GB_METHOD("_put", NULL, CLIST_put, "vi"),

  GB_END_DECLARE
};


/*!---------

\HELP
c-list
class

\IN
native

\TITLE
List

\DESC:fr
Cette classe implmente une liste indxe de taille variable aux
lments quelconques.

\REMARK:fr
Les objets de cette classe peuvent tre numrs.

----------*/

