/***************************************************************************

  CProcess.c

  The process management class

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CPROCESS_C

#include "common.h"

#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "gambas_api.h"
#include "gambas.h"
#include "stream.h"
#include "array.h"
#include "exec.h"
#include "trace.h"

#include "CProcess.h"

extern char **environ;

DECLARE_EVENT(EVENT_Write);
DECLARE_EVENT(EVENT_Error);
DECLARE_EVENT(EVENT_Kill);

PRIVATE CPROCESS *ProcessList = NULL;

PRIVATE void callback_write(int fd, int type, CPROCESS *process)
{
  /*fprintf(stderr, ">> Write\n"); fflush(stderr);*/
  char buffer[256];
  int n;
  GB_STRING str;

  n = read(fd, buffer, sizeof(buffer));

  if (n <= 0)
  {
    GB_Raise(process, EVENT_Kill, NULL);
    return;
  }

  str.addr = buffer;
  str.len = n;

  GB_Raise(process, EVENT_Write, (char *)&str);

  /*printf("(%d) %.*s\n", n, n, buffer);*/

  /*fprintf(stderr, "<< Write\n"); fflush(stderr);*/
}


PRIVATE void callback_error(int fd, int type, CPROCESS *process)
{
  /*fprintf(stderr, ">> Write\n"); fflush(stderr);*/
  char buffer[256];
  int n;
  GB_STRING str;

  n = read(fd, buffer, sizeof(buffer));

  if (n <= 0)
  {
    close(process->err);
    return;
  }

  str.addr = buffer;
  str.len = n;

  GB_Raise(process, EVENT_Error, (char *)&str);

  /*fprintf(stderr, "<< Write\n"); fflush(stderr);*/
}

/*
PRIVATE CFILE *create_stream_from(int fd, const char *mode)
{
  CFILE *file = CFILE_new();

  if (0)
  {
    file->stream.type = &STREAM_buffer;
    file->stream.buffer.file = fdopen(fd, mode);
    file->stream.buffer.size = 0;

    if (file->stream.buffer.file == NULL)
      THROW_SYSTEM(errno, NULL);
  }
  else
  {
    file->stream.type = &STREAM_direct;
    file->stream.direct.fd = fd;
    file->stream.buffer.size = 0;
  }

  GB_Ref(file);

  return file;
}
*/

PRIVATE void stop_process(CPROCESS *process)
{
  /*
  fprintf(stderr, "stop_process %d\n", process->pid);
  fflush(NULL);
  */

  close(process->in);
  process->pid = 0;
  process->running = FALSE;
}


PRIVATE void signal_child(int sig)
{
  pid_t pid;
  int status;
  CPROCESS *process;

  signal(SIGCHLD, signal_child);

  /*
  fprintf(stderr, "SIGCHLD\n");
  fflush(NULL);
  */

  for(;;)
  {
    pid = wait3(&status, WNOHANG, NULL);
    if (pid < 0)
      break;

    for (process = ProcessList; process; process = process->next)
    {
      if (process->pid == pid)
      {
        process->ret = status;

        /*printf("Process %d has returned %d\n", process->pid, status);*/
        fflush(NULL);

        stop_process(process);
      }
    }
  }
}


BEGIN_METHOD_VOID(CPROCESS_init)

  signal(SIGCHLD, signal_child);

END_METHOD


BEGIN_METHOD_VOID(CPROCESS_new)

  THIS->cmd = NULL;
  THIS->watch = GB_WATCH_NONE;
  THIS->pid = 0;
  THIS->in = 0;
  THIS->out = 0;
  THIS->err = 0;
  THIS->ret = 0;

  if (ProcessList)
    ProcessList->prev = THIS;

  THIS->next = ProcessList;
  THIS->prev = NULL;

  ProcessList = THIS;

END_METHOD


BEGIN_METHOD_VOID(CPROCESS_free)

  if (THIS->running)
    stop_process(THIS);

  GB_Watch(THIS->out, GB_WATCH_NONE, NULL, 0);
  GB_Watch(THIS->err, GB_WATCH_NONE, NULL, 0);

  GB_FreeString(THIS->cmd);
  /*GB_StoreObject(NULL, (void **)&THIS->param);*/

  close(THIS->in);
  close(THIS->out);
  close(THIS->err);

  if (THIS->prev)
    THIS->prev->next = THIS->next;

  if (THIS->next)
    THIS->next->prev = THIS->prev;

  if (THIS == ProcessList)
    ProcessList = THIS->next;

END_METHOD


BEGIN_PROPERTY(CPROCESS_signaled)

  if (READ_PROPERTY)
    GB_ReturnBoolean(TRACE.signaled);
  else
    TRACE.signaled = PROPERTY(char) != 0;

END_PROPERTY


BEGIN_PROPERTY(CPROCESS_command)

  if (READ_PROPERTY)
    GB_ReturnZeroString(THIS->cmd);
  else
    GB_StoreString(PROPERTY(GB_STRING), &(THIS->cmd));

END_PROPERTY


BEGIN_PROPERTY(CPROCESS_watch)

  if (READ_PROPERTY)
    GB_ReturnBoolean(THIS->watch);
  else
  {
    if (!THIS->running)
      THIS->watch = PROPERTY(char);
  }

END_PROPERTY


BEGIN_METHOD_VOID(CPROCESS_run)

  int fdin[2], fdout[2], fderr[2];
  pid_t pid;
  char *argv[] = { "sh", "-c", NULL, NULL };

  if (pipe(fdin) != 0 || pipe(fdout) != 0 || pipe(fderr) != 0)
    THROW_SYSTEM(errno, NULL);

  argv[2] = THIS->cmd;

  if (argv[2] == NULL)
    return;

  /*fprintf(stderr, "exec %s\n", argv[2]);
  fflush(NULL);*/

  pid = fork();
  if (pid == (-1))
    THROW_SYSTEM(errno, NULL);

  if (pid == 0)
  {
    /*
    pause();

    fprintf(stderr, "child got SIGUSR1\n");
    fflush(NULL);
    */

    close(fdin[1]);
    close(fdout[0]);
    close(fderr[0]);

    if ((dup2(fdin[0], STDIN_FILENO) != -1)
        && (dup2(fdout[1], STDOUT_FILENO) != -1)
        && (dup2(fderr[1], STDERR_FILENO) != -1))
    {
      execvp(argv[0], argv);
    }

    exit(127);
  }
  else
  {
    THIS->running = TRUE;
    THIS->pid = pid;

    /*printf("Running process %d\n", pid);
    fflush(NULL);*/

    close(fdin[0]);
    close(fdout[1]);
    close(fderr[1]);

    /*fcntl(fdout[0], F_SETFL, fcntl(fdout[0], F_GETFL) | O_NONBLOCK);*/
    /*fcntl(fderr[0], F_SETFL, fcntl(fderr[0], F_GETFL) | O_NONBLOCK);*/

    THIS->in = fdin[1];
    THIS->out = fdout[0];
    THIS->err = fderr[0];

    GB_Watch(THIS->out, GB_WATCH_READ, (void *)callback_write, (long)THIS);
    GB_Watch(THIS->err, GB_WATCH_READ, (void *)callback_error, (long)THIS);

    /*
    fprintf(stderr, "Send signal to child\n");
    fflush(NULL);

    kill(pid, SIGUSR1);
    */
  }

END_METHOD


BEGIN_METHOD_VOID(CPROCESS_kill)

  if (!THIS->running)
    return;

  kill(THIS->pid, SIGKILL);
  /*stop_process(THIS);*/

END_METHOD


BEGIN_METHOD_VOID(CPROCESS_signal)

  if (!THIS->running)
    return;

  /*
  printf("Send SIGUSR1 to process %d\n", THIS->pid);
  fflush(NULL);*/
  kill(THIS->pid, SIGUSR1);

END_METHOD


BEGIN_METHOD(CPROCESS_send, GB_STRING text)

  STREAM stream;

  if (!THIS->running)
    return;

  stream.type = &STREAM_direct;
  stream.direct.fd = THIS->in;
  stream.direct.size = 0;

  STREAM_write(&stream, PARAM(text).addr, PARAM(text).len);

END_METHOD


PUBLIC GB_DESC NATIVE_Process[] =
{
  GB_DECLARE("Process", sizeof(CPROCESS)),

  GB_STATIC_PROPERTY("Signaled", "b", CPROCESS_signaled),

  GB_PROPERTY("Command", "s", CPROCESS_command),
  GB_PROPERTY("Watch", "b", CPROCESS_watch),

  GB_STATIC_METHOD("_init", NULL, CPROCESS_init, NULL),
  /*GB_STATIC_METHOD("_exit", NULL, CPROCESS_exit, NULL),*/
  GB_METHOD("_new", NULL, CPROCESS_new, NULL),
  GB_METHOD("_free", NULL, CPROCESS_free, NULL),

  GB_METHOD("Run", NULL, CPROCESS_run, NULL),
  GB_METHOD("Kill", NULL, CPROCESS_kill, NULL),
  GB_METHOD("Signal", NULL, CPROCESS_signal, NULL),
  GB_METHOD("Send", NULL, CPROCESS_send, "(Input)s"),

  GB_EVENT("Write", NULL, "s", &EVENT_Write),
  GB_EVENT("Error", NULL, "s", &EVENT_Error),
  GB_EVENT("Kill", NULL, NULL, &EVENT_Kill),

  GB_END_DECLARE
};

