/***************************************************************************

  error.c

  Errors management routines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define _ERROR_C

#include "common.h"
#include <stdarg.h>

#include "buffer.h"
#include "error.h"
#include "debug.h"

#include "exec.h"
#include "trace.h"
#include "gambas_api.h"


PUBLIC long ERROR_code = 0;

PRIVATE char *ERROR_Message[256] =
{
  /*  0 E_UNKNOWN */ "Unknown error #%d",
  /*  1 E_MEMORY */ "Out of memory",
  /*  2 E_CLASS */ "Cannot load class '%s': %s%s",
  /*  3 E_STACK */ "Stack overflow",
  /*  4 E_NEPARAM */ "Not enough arguments",
  /*  5 E_TMPARAM */ "Too many arguments",
  /*  6 E_TYPE */ "Type mismatch: wanted %s, got %s instead",
  /*  7 E_OVERFLOW */ "Overflow",
  /*  8 E_ILLEGAL */ "Illegal instruction",
  /*  9 E_NFUNC */ "Not a function",
  /* 10 E_CSTATIC */ "Class '%s' is static",
  /* 11 E_NSYMBOL */ "Unknown symbol '%s' in class '%s'",
  /* 12 E_NOBJECT */ "Not an object",
  /* 13 E_NULL */ "Null object",
  /* 14 E_STATIC */ "'%s.%s' is static",
  /* 15 E_NREAD */ "'%s.%s' is write only",
  /* 16 E_NWRITE */ "'%s.%s' is read only",
  /* 17 E_NPROPERTY */ "'%s.%s' is not a property",
  /* 18 E_NRETURN */ "No return value",
  /* 19 E_MATH*/ "Mathematic error",
  /* 20 E_ARG */ "Bad argument",
  /* 21 E_BOUND */ "Out of bounds",
  /* 22 E_NDIM */ "Bad number of dimensions",
  /* 23 E_NARRAY */ "Not an array",
  /* 24 E_MAIN */ "No startup method",
  /* 25 E_NNEW */ "No instanciation method",
  /* 26 E_ZERO */ "Division by zero",
  /* 27 E_LIBRARY */ "Cannot load library '%s': %s",
  /* 28 E_EVENT */ "Bad event handler in %s: %s()",
  /* 29 E_IOBJECT */ "Invalid object",
  /* 30 E_FROMLIB */ "",
  /* 31 E_ENUM */ "Not an enumeration",
  /* 32 E_BADKEY */ "Invalid key '%.*s'",
  /* 33 E_DATE */ "Invalid date",
  /* 34 E_BADPATH */ "Invalid path",
  /* 35 E_OPEN */ "Cannot open file '%s': %s",
  /* 36 E_PROJECT */ "Bad project file: line %d: %s",
  /* 37 E_FULL */ "Device is full",
  /* 38 E_EXIST */ "File already exists: %s",
  /* 39 E_EOF */ "End of file",
  /* 40 E_FORMAT */ "Bad format string",
  /* 41 E_DYNAMIC */ "'%s.%s' is not static",
  /* 42 E_SYSTEM */ "System error. %s",
  /* 43 E_ACCESS */ "Access forbidden",
  /* 44 E_TOOLONG */ "File name is too long",
  /* 45 E_NEXIST */ "File or directory does not exist: %s",
  /* 46 E_DIR */ "File is a directory: %s",
  /* 47 E_READ */ "Read error",
  /* 48 E_WRITE */ "Write error",
  /* 49 E_NDIR */ "Not a directory: %s",
  /* 50 E_REGEXP */ "Bad regular expression: %s",
  /* 51 E_ARCH */ "Bad archive: %s",
  /* 52 E_REGISTER */ "Cannot register class '%s'",
  /* 53 E_CLOSED */ "File is closed",
  /* 54 E_VIRTUAL */ "Bad use of virtual class",
  NULL
};

PRIVATE ERROR *ERROR_current = NULL;
PRIVATE boolean NoError = FALSE;
PRIVATE char *MessageBuffer = NULL;


PRIVATE void ERROR_clear(void)
{
  errno = 0;
  if (MessageBuffer != NULL)
    BUFFER_delete(&MessageBuffer);
}


PUBLIC void ERROR_enter(ERROR *err)
{
  Clear(err);
  err->prev = ERROR_current;
  ERROR_current = err;
  ERROR_clear();
}


PUBLIC void ERROR_leave(ERROR *err)
{
  if (err->prev != ERROR_LEAVE_DONE)
  {
    /*ERROR_panic("ERROR_leave already done");*/

    ERROR_current = err->prev;
    err->prev = ERROR_LEAVE_DONE;
  }
}


PUBLIC char *ERROR_get(void)
{
  /*
  if (code > 0 && code < 256)
    return strerror(code);
  else
    return ERROR_Message[code - 256];
  */
  return strerror(errno);
}

PRIVATE void add_error_message(char *msg, boolean newline)
{
  int len;

  if (NoError)
    ERROR_panic("Cannot throw error. %s", msg);

  NoError = TRUE;

  if (!MessageBuffer)
    BUFFER_create(&MessageBuffer);

  len = strlen(msg);

  if (len)
    BUFFER_add(&MessageBuffer, msg, len);
  /*
  if (msg[len - 1] != '.')
    BUFFER_add(&MessageBuffer, ".", 1);
  */
  if (newline)
    BUFFER_add(&MessageBuffer, "\n", 1);

  NoError = FALSE;
}


PUBLIC void PROPAGATE()
{
  ERROR *err;

  if (ERROR_current == NULL)
    ERROR_panic("Cannot propagate error. No error handler.");

  err = ERROR_current;
  ERROR_leave(ERROR_current);
  longjmp(err->env, 1);
}


PRIVATE void throw_error(char *msg) NORETURN;

PRIVATE void throw_error(char *msg)
{
  add_error_message(msg, FALSE);

  /*
  if (errno)
  {
    add_error_message(": ", FALSE);
    add_error_message(strerror(errno), FALSE);
    errno = 0;
  }
  */

  add_error_message("", TRUE);

  PROPAGATE();
}


PUBLIC void THROW(long code, ...)
{
  va_list args;
  int n;
  char buf[512];

  #if DEBUG_ERROR
  printf("THROW %d\n", code);
  #endif

  if (FP != NULL)
  {
    sprintf(buf, "[%s] ", TRACE_get_current_position());
    add_error_message(buf, FALSE);
  }

  sprintf(buf, "#%ld: ", code);
  add_error_message(buf, FALSE);

  va_start(args, code);

  if (code > 0 && code < 255)
    n = vsnprintf(buf, sizeof(buf), ERROR_Message[code], args);
  else
    n = snprintf(buf, sizeof(buf), ERROR_Message[0], code);

  ERROR_code = code;
  throw_error(buf);
}


PUBLIC void THROW_SYSTEM(int err, char *path)
{
  switch(err)
  {
    case ENOENT:
      THROW(E_NEXIST, path);

    case EISDIR:
      THROW(E_DIR, path);

    case ENOTDIR:
      THROW(E_NDIR, path);

    case ENOMEM:
      THROW(E_MEMORY);

    case EACCES:
      THROW(E_ACCESS);

    case ENAMETOOLONG:
      THROW(E_TOOLONG);

    case ENOSPC:
      THROW(E_FULL);

    case EEXIST:
      THROW(E_EXIST);

    default:
      THROW(E_SYSTEM, strerror(err));
  }
}


PUBLIC void THROW_LIBRARY()
{
  int n;
  char buf[512];
  char *msg = GAMBAS_Error;

  if (FP != NULL)
  {
    sprintf(buf, "[%s] ", TRACE_get_current_position());
    add_error_message(buf, FALSE);
  }

  sprintf(buf, "#%d: ", E_FROMLIB);
  add_error_message(buf, FALSE);

  if ((long)msg > 0 && (long)msg < 255)
    n = snprintf(buf, sizeof(buf), ERROR_Message[(long)msg]);
  else
    n = snprintf(buf, sizeof(buf), msg);

  ERROR_code = E_FROMLIB;
  throw_error(msg);
}


PUBLIC void ERROR_panic(char *error, ...)
{
  va_list args;

  va_start(args, error);

  fflush(NULL);

  fprintf(stderr, "\n"
                  "** INTERNAL ERROR **\n"
                  );
  vfprintf(stderr, error, args);
  fprintf(stderr, "\n");
  if (MessageBuffer)
  {
    fprintf(stderr, "\n");
    ERROR_print();
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "** Program aborting... Sorry... :-(\n\n");
  abort();
}


PUBLIC void ERROR_print_at(FILE *where)
{
  fprintf(where, "%.*s", (int)BUFFER_length(MessageBuffer), MessageBuffer);
}

PUBLIC void ERROR_print(void)
{
  ERROR_print_at(stderr);
}


