/***************************************************************************

  exec.c

  Subroutines for the interpreter : executing methods, native methods,
  the NEW operator, the casting operator, etc.

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/


#define __EXEC_C

#include "common.h"
#include "error.h"
#include "type.h"

#include <unistd.h>
#include <sys/time.h>

#include "limit.h"
#include "subr.h"
#include "stack.h"
#include "trace.h"

#include "String.h"
#include "Date.h"

#include "CCollection.h"

#include "gambas_api.h"
#include "exec.h"

/* Contexte  sauvegarder */
PUBLIC STACK_CONTEXT EXEC_current = { 0 };

/* Opcode de l'instruction en cours */
PUBLIC PCODE EXEC_code;

/* Pointeur de pile */
PUBLIC VALUE *SP = NULL;

/* Valeur de retour de la fonction */
PRIVATE VALUE RETURN;
PUBLIC VALUE *RP = &RETURN;

/* Stockage temporaire */
PUBLIC VALUE TEMP;

/* Mode dbogage */
PUBLIC boolean EXEC_debug = FALSE;

/* Temps au dmarrage */
PUBLIC double EXEC_start_time;
/* Indique la fin d'une numration */
PUBLIC boolean EXEC_enum_stop = FALSE;
/* Etat de l'numration en cours */
PUBLIC void *EXEC_enum_data;

/* Excution d'une archive */
PUBLIC boolean EXEC_arch = FALSE;

/* Envoi d'un signal au parent pendant le dbogage */
PUBLIC boolean EXEC_signal = FALSE;

PUBLIC EXEC_HOOK EXEC_Hook = { NULL };

#if 0
/* Constantes intgres */
PUBLIC long EXEC_const[] =
{
  0, /* erreur */
  0, /* gbCompBinary */
  1, /* gbCompCase */
  2, /* gbCompLang */
  4, /* gbSearchBackward */
  8,  /* gbSearchWord */

  1, /* gbRead */
  2, /* gbWrite */
  4, /* gbAppend */
  4, /* gbCreate */
  8, /* gbDirect */
  16, /* gbBinary */

  0, /* gbNull */
  1, /* gbBoolean */
  2, /* gbByte */
  3, /* gbShort */
  4, /* gbInteger */
  5, /* gbLong */
  6, /* gbFloat */
  7, /* gbDate */
  8, /* gbString */
  11, /* gbVariant */
  15, /* gbObject */

  0, /* gbFile */
  1, /* gbDirectory */
  2, /* gbDevice */
  3, /* gbPipe */
  4, /* gbSocket */
  5, /* gbLink */
  8, /* gbHidden */

  14, /* gbVoid */

};
#endif




PUBLIC void EXEC_init(void)
{
  struct timeval tv;

  PC = NULL;
  BP = NULL;
  OP = NULL;
  CP = NULL;

  if (gettimeofday(&tv, NULL) == 0)
    EXEC_start_time = (double)tv.tv_sec + (double)tv.tv_usec / 1E6;
  else
    EXEC_start_time = 0.0;

  DATE_init();
}


#define print_register() \
  printf("| SP = %d  PP = %d  BP = %d  FP = %p  PC = %p  EC = %p\n", SP - (VALUE *)STACK_base, PP - (VALUE *)STACK_base, BP - (VALUE *)STACK_base, FP, PC, EC)


PUBLIC void EXEC_enter(VALUE_FUNCTION *value, int nparam)
{
  CLASS *class = value->class;
  unsigned long func_id;

  FUNCTION *func;
  int i;

  #if DEBUG_STACK
  printf("\n| >> EXEC_enter(%s, %ld, %d)\n", class->name, value->function, nparam);
  print_register();
  #endif

  func_id = value->function;

  if ((func_id & 0xFFFFE000) != 0)
    func_id = (int)((CLASS_DESC_METHOD *)func_id)->exec;

  func = &class->load->func[(int)func_id];


  /* vrification des paramtres */

  if (nparam < func->npmin)
    THROW(E_NEPARAM);
  else if (nparam > func->n_param)
    THROW(E_TMPARAM);

  /* paramtres obligatoires et rservation */

  for (i = 0; i < func->npmin; i++)
  {
    VALUE_conv(SP - nparam + i, func->param[i].type);
    BORROW(SP - nparam + i);
  }

  if (func->npmin < func->n_param)
  {
    /* paramtres facultatifs, mais passs  la fonction */

    for (i = func->npmin; i < nparam; i++)
    {
      if (SP[- nparam + i].type == T_VOID)
        SP[- nparam + i]._void.ptype = func->param[i].type;
      else
      {
        VALUE_conv(SP - nparam + i, func->param[i].type);
        BORROW(SP - nparam + i);
      }
    }

    /* paramtres facultatifs non passs  la fonction */

    if (nparam < func->n_param)
    {
      STACK_check(func->n_param - nparam);

      for (i = nparam; i < func->n_param; i++)
      {
        SP->type = T_VOID;
        SP->_void.ptype = func->param[i].type;
        SP++;
      }
    }
  }

  /* sauvegarde du contexte */

  STACK_push_frame(&EXEC_current);

  /* Vrification de la pile */

  STACK_check(func->stack_usage);

  /* entre de fonction */

  BP = SP;
  FP = func;
  PC = func->code;
  OP = value->object;
  CP = class;
  EP = NULL;
  if (func->error)
  {
    #if DEBUG_ERROR
      printf("EXEC_enter: EC = PC + %d\n", func->error);
    #endif
    EC = PC + func->error;
  }
  else
    EC = NULL;

  /* On rfrence l'objet pour qu'il ne soit pas dtruit pendant l'excution de la mthode */
  OBJECT_REF(OP, "EXEC_enter");

  /*printf("PC = %p  nparam = %d\n", PC, FP->n_param);*/

  /* Initialisation des variables locales */

  if (func->n_local)
  {
    for (i = 0; i < func->n_local; i++)
    {
      VALUE_class_default(class, SP, func->local[i].type);
      SP++;
    }
  }

  /* Initialisation des variables de controles */

  if (func->n_ctrl)
  {
    for (i = 0; i < func->n_ctrl; i++)
    {
      SP->type = T_VOID;
      SP++;
    }
  }

  RP->type = T_VOID;

  #if DEBUG_STACK
  printf("| << EXEC_enter()\n");
  print_register();
  #endif
}



PUBLIC void EXEC_leave(VALUE *ret)
{
  int nparam;
  boolean drop; /*output*/

#if DEBUG_STACK
  printf("| >> EXEC_leave\n");
  print_register();
#endif

  /* Je garde cette bourde en commentaires.
     La honte...

  for (i = 0; i < (SP - BP); i++)
  {
    SP--;
    #if 1
    printf("Release local %p (0x%08lX)\n", SP, (ulong)SP->type);
    #endif
    RELEASE(SP);
  }

  */


  /* Libration des variables locales et de controle */

  SP = BP;
  nparam = FP->n_local + FP->n_ctrl;

#if DEBUG_REF
  printf("EXEC_leave: local %d\n", nparam);
#endif

  while (nparam)
  {
    RELEASE(SP);
    SP++;
    nparam--;
  }

  /* Libration des paramtres */

  nparam = FP->n_param;

#if DEBUG_REF
  printf("EXEC_leave: param %d\n", nparam);
#endif

  while (nparam)
  {
    RELEASE(&BP[-nparam]);
    nparam--;
  }

  /* On libre l'objet reserv dans EXEC_enter() */
  OBJECT_UNREF((void **)&OP, "EXEC_leave");

  /* restitution du contexte */

  SP = BP;

  /* On rcupre le nombre de paramtres pousss sur la pile */
  nparam = FP->n_param;

  STACK_pop_frame(&EXEC_current);

  /*printf("PC = %p  nparam = %d\n", PC, nparam);*/

  if (PC)
  {
    drop = PCODE_is_void(*PC);
    SP -= nparam + 1;
    /*output = PCODE_is_output(*PC);*/
  }
  else
  {
    drop = TRUE;
    SP -= nparam;
  }

  /*
    if (PCODE_is(*PC, C_CALL))
    {
  */

  /*
  **  Si on manipule un type VOID, la fonction
  **  de conversion VALUE_conv() dclenchera
  **  l'erreur E_NRETURN !
  */

  /*
  if (!drop && RP->type == T_VOID)
    THROW(E_NRETURN);
  */

  /*
    Attention ! le RELEASE(RP) conjugu au UNBORROW(RP)
    peut faire descendre le compteur de rfrence d'un objet  -1 !
  */

  /*
    NOTE: les paramtres input-output ont le mme problme que la
    valeur de retour. Ils peuvent contenir des rfrences dsalloues !
    A CORRIGER ! En attendant, ils sont dsactivs.
  */

  /*
  SP--;

#if DEBUG_REF
        printf("EXEC_leave: param string %d ??\n", nparam);
#endif

  while (nparam > 0)
  {
    SP--;
    RELEASE_STRING(SP);
    nparam--;
  }
  */

#if DEBUG_REF
  printf("EXEC_leave: return\n");
#endif

  if (!drop)
  {
    UNBORROW(RP);
    *SP++ = *RP;
    RP->type = T_VOID;
  }
  else
    RELEASE(RP);

/* output = TRUE

      }
      else
      {
        if (!drop)
        {
          UNBORROW(RP);
          SP[-(nparam + 1)] = *RP;
          RP->type = T_VOID;
        }
        else
          RELEASE(RP);
      }
    }
  }
*/

#if DEBUG_STACK
  printf("| << EXEC_leave()\n");
  print_register();
  printf("\n");
#endif
}


PUBLIC void EXEC_function(VALUE_FUNCTION *func, int nparam)
{
  /*PCODE *save = PC;*/
  boolean retry;

  /* ncessaire, car *PC est examin par EXEC_leave pour savoir si on attend une valeur de retour */
  STACK_push_frame(&EXEC_current);

  PC = NULL;

  EXEC_enter(func, nparam);

  if (PC != NULL)
  {
    do
    {
      TRY
      {
        EXEC_loop();
        retry = FALSE;
      }
      CATCH
      {
        if (EP != NULL)
        {
          #if DEBUG_ERROR
          printf("#1 EP = %p\n", EP);
          #endif

          while (SP > EP)
          {
            SP--;
            RELEASE_STRING(SP);
          }

          PC = EC;
          retry = TRUE;
        }
        else if (EC != NULL)
        {
          #if DEBUG_ERROR
          printf("#2 EC = %p\n", EC);
          #endif

          PC = EC;
          EC = NULL;
          retry = TRUE;
        }
        else
        {
          #if DEBUG_ERROR
          printf("#3\n");
          #endif

          if (EXEC_debug && !STACK_has_error_handler())
          {
            TRACE_main(TRUE);
            retry = TRUE;
          }
          else
          {
            while (PC != NULL && EC == NULL)
              EXEC_leave(NULL);

            if (PC == NULL)
            {
              ERROR_print();
              abort();
              /*retry = FALSE;*/
            }

            PC = EC;
            EC = NULL;
            retry = TRUE;
          }
        }
      }
      END_TRY

      #if DEBUG_ERROR
      if (retry)
        printf("retry %p\n", PC);
      #endif
    }
    while (retry);
  }

  STACK_pop_frame(&EXEC_current);
  /*PC = save;*/
}


PUBLIC boolean EXEC_call_native(void (*exec)(), void *object, char *param, TYPE type)
{
  TYPE save_type = GAMBAS_ReturnType;

  GAMBAS_ReturnType = type;
  GAMBAS_Error = NULL;

  OBJECT_REF(object, "EXEC_call_native");
  (*exec)(object, param);
  OBJECT_UNREF(&object, "EXEC_call_native");

  GAMBAS_ReturnType = save_type;

  return (GAMBAS_Error != NULL);
}


PUBLIC void EXEC_native(VALUE_FUNCTION *func, int nparam, boolean drop)
{
  char *param;
  CLASS_DESC_METHOD *desc = (CLASS_DESC_METHOD *)func->function;
  int i;
  VALUE *value;
  char param_buffer[nparam * 12];
  boolean error;

  /* vrification des paramtres */

  if (nparam < desc->npmin)
    THROW(E_NEPARAM);
  else if (nparam > desc->npmax)
    THROW(E_TMPARAM);

  param = param_buffer;

  /* Transcription des paramtres et mmorisation des chaines  librer */


  GAMBAS_MissingParam = 0xFFFFFFFF;

  for (i = 0, value = &SP[-nparam]; i < nparam; i++, value++)
  {
    if (value->type != T_VOID)
    {
      /*npreal = i + 1;*/
      GAMBAS_MissingParam &= ~(1L << i);
      VALUE_put(value, param, desc->signature[i]);
    }

    if (TYPE_is_string(desc->signature[i]))
      param += sizeof(GB_STRING);
    else
      param += TYPE_sizeof(desc->signature[i]);
  }

  error = EXEC_call_native(desc->exec, func->object, param_buffer, desc->type);


  /* Libration des chaines passes en paramtres */

  while (nparam)
  {
    nparam--;
    SP--;
    RELEASE_STRING(SP);
  }

  /* Si la description de la fonction se trouve sur la pile */
  SP--;
  if ((void *)SP  != (void *)func)
    SP++;
  /*
  #if DEBUG_STACK
  else
    printf("** SP != func SP = %p func = %p **\n>", SP, func);
  #endif
  */

  if (!drop)
  {
    if (desc->type == T_VOID)
    {
      SP->type = T_VOID;
      SP->_void.ptype = T_NULL;
    }
    else
      *SP = TEMP;
    SP++;
  }

  if (error)
    THROW_LIBRARY();
}


PUBLIC void EXEC_object(VALUE *SP, CLASS **class, OBJECT **object, boolean *defined)
{
  if (SP->type == T_CLASS)
  {
    *class = SP->_class.class;
    *object = NULL;
    *defined = TRUE;

    CLASS_load(*class);
  }
  else
  {
    if (SP->type == T_OBJECT)
    {
      *object = SP->_object.object;
      *class = OBJECT_class(*object);
      *defined = FALSE;
    }
    else if (TYPE_is_object(SP->type))
    {
      *object = SP->_object.object;
      *class = SP->_object.class;
      *defined = TRUE;
    }
    else if (SP->type == T_VARIANT)
    {
      if (SP->_variant.vtype == T_OBJECT)
      {
        *object = *((void **)SP->_variant.value);
        *class = OBJECT_class(*object);
        *defined = FALSE;
      }
      else if (TYPE_is_object(SP->_variant.vtype))
      {
        *object = *((void **)SP->_variant.value);
        *class = (CLASS *)SP->_variant.vtype;
        *defined = FALSE;
      }
      else
        THROW(E_NOBJECT);
    }
    else
      THROW(E_NOBJECT);

    if (*object == NULL)
      THROW(E_NULL);

    CLASS_load(*class);

    if ((*class)->check(*object))
      THROW(E_IOBJECT);
  }
}


PUBLIC void EXEC_public(CLASS *class, void *object, char *name, int nparam)
{
  CLASS_DESC *desc;
  VALUE_FUNCTION func;

  desc = CLASS_get_symbol_desc_kind(class, name, (object != NULL) ? CD_METHOD : CD_STATIC_METHOD, 0);

  if (desc == NULL)
    return;

  func.class = class;
  func.object = object;
  func.function = (long)desc;

  if (CLASS_is_native(class))
    EXEC_native(&func, nparam, TRUE);
  else
    EXEC_function(&func, nparam);
}



PUBLIC boolean EXEC_spec(int special, CLASS *class, void *object, int nparam, boolean drop, boolean use_stack)
{
  CLASS_DESC *desc = class->special[special];
  VALUE_FUNCTION func, *exec;

  if (desc == NULL)
    return TRUE;

  if (CLASS_DESC_get_type(desc) == CD_STATIC_METHOD)
  {
    if (object != NULL)
      return TRUE;
  }
  else
  {
    if (object == NULL)
      THROW(E_NOBJECT);
  }

  if (use_stack)
    exec = (VALUE_FUNCTION *)(SP - nparam - 1);
  else
    exec = &func;

  exec->class = class;
  exec->object = object;

  if (CLASS_is_native(class))
  {
    exec->function = (long)desc;
    EXEC_native(exec, nparam, drop);
  }
  else
  {
    exec->function = (long)desc->method.exec;
    EXEC_function(exec, nparam);
  }

  return FALSE;
}


/* PRIVATE void dump(int np) */
/* { */
/*   int i; */
/*  */
/*   for (i = 1; i <= np; i++) */
/*     printf("SP[%d] = %d  ", -i, SP[-i].type); */
/*  */
/*   printf("\n"); */
/* } */
/*  */

PUBLIC void EXEC_special_inheritance(int special, CLASS *class, OBJECT *object, int nparam, boolean drop)
{
  /*CLASS **pher;
  boolean dup;
  VALUE *save_SP;
  int i;*/
  VALUE_FUNCTION exec;

  if (class->parent != NULL)
    EXEC_special_inheritance(special, class->parent, object, 0, drop);

  /*
  pher = CLASS_get_inheritance(class);

  while (*pher)
  {
  */
    /*
    dup = ((nparam > 0) && (pher[1] != NULL));

    if (dup)
    {
      save_SP = SP;
      SP =  STACK_check(SP, nparam);

      for (i = 1; i <= nparam; i++)
        SP[-i] = save_SP[-i];
    }
    */

    /* Initialisation des objets de classes non natives */

    if (special == SPEC_NEW)
    {
      if (!CLASS_is_native(class))
      {
        exec.class = class;
        exec.object = object;
        exec.function = FUNC_INIT_DYNAMIC;

        EXEC_function(&exec, 0);
      }
    }

    EXEC_special(special, class, object, nparam, drop);

    /*
    if (dup)
      SP = save_SP;
    */
    /*
    pher++;
  }
  */
}


PUBLIC void EXEC_new(void)
{
  CLASS *class;
  int np;
  boolean event;
  void *object;
  char *name = NULL;

  np = *PC & 0xFF;
  event = np & CODE_NEW_EVENT;
  np &= 0x3F;

  /* Instanciation */

  SP -= np;

  if (SP->type != T_CLASS)
    THROW(E_ILLEGAL);

  class = SP->_class.class;

  SP += np;
  CLASS_load(class);
  SP -= np;

  if (class->new == NULL)
    THROW(E_CSTATIC, class->name);

  SP += np;

  if (event)
  {
    SP--;

    if (!TYPE_is_string(SP->type))
      THROW(E_TYPE, "String", TYPE_get_name(SP->type));

    STRING_copy_from_value(&name, SP);
    (*class->new)(&object, class, name, ((OP == NULL) ? (OBJECT *)CP : (OBJECT *)OP));
    STRING_free(&name);
    np -= 2;
  }
  else
  {
    (*class->new)(&object, class, name, ((OP == NULL) ? (OBJECT *)CP : (OBJECT *)OP));
    np--;
  }

  /*OBJECT_REF(object, "EXEC_new");*/

  EXEC_special_inheritance(SPEC_NEW, class, object, np, TRUE);

  /*
  pher = CLASS_get_inheritance(class);

  while (*pher)
  {
    if (!CLASS_is_native(*pher))
    {
      exec.class = *pher;
      exec.object = object;
      exec.function = FUNC_INIT_DYNAMIC;

      EXEC_function(&exec, 0);
   }

    pher++;
  }
  */

  /* On retourne l'objet cr */

  SP--;
  SP->_object.class = class;
  SP->_object.object = object;
  SP++;

  /* L'objet a t cre avec un nombre de rfrence gal  1.
     On remet ce nombre  0 maintenant que l'objet est prt.
     Mais on ne le dtruit pas ! */

  OBJECT_UNREF_KEEP(&object, "EXEC_new");
}


PUBLIC void EXEC_class(CLASS *class, int nparam, boolean drop)
{
  if (EXEC_special(SPEC_CALL, class, NULL, nparam, drop))
  {
    if (nparam < 1)
      THROW(E_NEPARAM);
    else if (nparam > 1)
      THROW(E_TMPARAM);

    VALUE_conv(SP - 1, (TYPE)class);
  }
}


PUBLIC void EXEC_nop(void)
{
  printf("* NOP *\n");
}


PUBLIC void EXEC_ILLEGAL(void)
{
  THROW(E_ILLEGAL);
}
