/***************************************************************************

  exec_loop.c

  The interpreter main loop

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __EXEC_LOOP_C

#include "common.h"
#include "error.h"
#include "type.h"
#include "debug.h"

#include "subr.h"
#include "pcode.h"
#include "stack.h"
#include "trace.h"
#include "event.h"
#include "value.h"

#include "String.h"
#include "gambas_api.h"

#include "exec.h"


#define SUBR_is EXEC_nop

#define SUBR_replace EXEC_nop
#define SUBR_if EXEC_nop
#define SUBR_choose EXEC_nop
#define SUBR_beep EXEC_nop


PRIVATE EXEC_FUNC SubrTable[] =
{
  /* 28 */  SUBR_eq,            SUBR_ne,            SUBR_gt,            SUBR_le,
  /* 2C */  SUBR_lt,            SUBR_ge,            SUBR_is,            SUBR_case,
  /* 30 */  SUBR_add,           SUBR_sub,           SUBR_mul,           SUBR_div,
  /* 34 */  SUBR_neg,           SUBR_quo,           SUBR_rem,           SUBR_pow,
  /* 38 */  SUBR_and,           SUBR_or,            SUBR_not,           SUBR_xor,
  /* 3C */  SUBR_cat,           SUBR_like,          SUBR_file,          EXEC_ILLEGAL,

  SUBR_left,       /* 00 */
  SUBR_mid,        /* 01 */
  SUBR_right,      /* 02 */
  SUBR_len,        /* 03 */
  SUBR_space,      /* 04 */
  SUBR_string,     /* 05 */
  SUBR_trim,       /* 06 */
  SUBR_upper,      /* 07 */
  SUBR_lower,      /* 08 */
  SUBR_chr,        /* 09 */
  SUBR_asc,        /* 10 */
  SUBR_instr,      /* 11 */
  SUBR_instr,      /* 12 */
  SUBR_subst,      /* 13 */
  SUBR_replace,    /* 14 */
  EXEC_ILLEGAL,    /* 15 */
  EXEC_ILLEGAL,    /* 16 */

  SUBR_abs,        /* 17 */
  SUBR_int,        /* 18 */

  SUBR_math,       /* 19 */

  SUBR_sgn,        /* 20 */
  SUBR_fix,        /* 21 */

  SUBR_pi,         /* 22 */
  SUBR_round,      /* 23 */
  SUBR_randomize,  /* 24 */
  SUBR_rnd,        /* 25 */
  SUBR_min_max,    /* 26 */
  SUBR_min_max,    /* 27 */

  SUBR_if,         /* 28 */
  SUBR_choose,     /* 29 */
  EXEC_ILLEGAL,    /* 30 */
  EXEC_ILLEGAL,    /* 31 */
  EXEC_ILLEGAL,    /* 32 */
  EXEC_ILLEGAL,    /* 33 */
  EXEC_ILLEGAL,    /* 34 */

  SUBR_bit,        /* 35 */
  EXEC_ILLEGAL,    /* 36 */
  EXEC_ILLEGAL,    /* 37 */
  EXEC_ILLEGAL,    /* 38 */
  EXEC_ILLEGAL,    /* 39 */
  EXEC_ILLEGAL,    /* 40 */
  EXEC_ILLEGAL,    /* 41 */

  SUBR_is_type,    /* 42 */
  SUBR_type,       /* 43 */

  SUBR_conv,       /* 44 */

  SUBR_bin,        /* 45 */
  SUBR_hex,        /* 46 */
  SUBR_val,        /* 47 */
  SUBR_str,        /* 48 */
  SUBR_format,     /* 49 */

  SUBR_timer,      /* 50 */
  SUBR_now,        /* 51 */

  SUBR_year,       /* 52 */

  SUBR_date,       /* 53 */
  SUBR_time,       /* 54 */

  SUBR_error,      /* 55 */

  SUBR_shell,      /* 56 */
  SUBR_message,    /* 57 */
  SUBR_beep,       /* 58 */
  SUBR_wait,       /* 59 */

  SUBR_open,       /* 60 */
  SUBR_close,      /* 61 */
  SUBR_input,      /* 62 */
  SUBR_print,      /* 63 */
  SUBR_read,       /* 64 */
  SUBR_write,      /* 65 */
  SUBR_eof,        /* 66 */
  SUBR_lof,        /* 67 */
  SUBR_seek,       /* 68 */
  SUBR_kill,       /* 69 */
  SUBR_mkdir,      /* 70 */
  SUBR_rmdir,      /* 71 */
  SUBR_exist,      /* 72 */
  SUBR_rename,     /* 73 */
  SUBR_stat,       /* 74 */
  SUBR_linput,     /* 75 */
  EXEC_ILLEGAL,    /* 76 */
  EXEC_ILLEGAL,    /* 77 */
  SUBR_dir,        /* 78 */
  SUBR_flush,      /* 79 */

  EXEC_ILLEGAL,    /* 80 */
  EXEC_ILLEGAL,    /* 81 */
  EXEC_ILLEGAL,    /* 82 */
  EXEC_ILLEGAL,    /* 83 */
  EXEC_ILLEGAL,    /* 84 */
  EXEC_ILLEGAL,    /* 85 */
  EXEC_ILLEGAL,    /* 86 */
};


PUBLIC void EXEC_loop(void)
{
  static void *jump_table[256] =
  {
    /* 00 NOP             */  &&_NEXT,
    /* 01 PUSH LOCAL      */  &&_PUSH_LOCAL,
    /* 02 PUSH ARRAY      */  &&_PUSH_ARRAY,
    /* 03 PUSH UNKNOWN    */  &&_PUSH_UNKNOWN,
    /* 04 illegal         */  &&_ILLEGAL,
    /* 05 illegal         */  &&_ILLEGAL,
    /* 06 PUSH SPECIAL    */  &&_PUSH_SPECIAL,
    /* 07 PUSH EVENT      */  &&_PUSH_EVENT,
    /* 08 END             */  &&_TODO,
    /* 09 POP LOCAL       */  &&_POP_LOCAL,
    /* 0A POP ARRAY       */  &&_POP_ARRAY,
    /* 0B POP UNKNOWN     */  &&_POP_UNKNOWN,
    /* 0C POP OPTIONAL    */  &&_POP_OPTIONAL,
    /* 0D POP CTRL        */  &&_POP_CTRL,
    /* 0E BREAK           */  &&_BREAK,
    /* 0F RETURN          */  &&_RETURN,
    /* 10 PUSH SHORT      */  &&_PUSH_SHORT,
    /* 11 PUSH INTEGER    */  &&_PUSH_INTEGER,
    /* 12 PUSH CHAR       */  &&_PUSH_CHAR,
    /* 13 PUSH ME         */  &&_PUSH_ME,
    /* 14 PUSH NULL       */  &&_PUSH_NULL,
    /* 15 PUSH VOID       */  &&_PUSH_VOID,
    /* 16 PUSH RETURN     */  &&_PUSH_RETURN,
    /* 17 PUSH BOOLEAN    */  &&_PUSH_BOOLEAN,
    /* 18 PUSH LAST       */  &&_PUSH_LAST,
    /* 19 EVENT           */  &&_EVENT,
    /* 1A TRY             */  &&_TRY,
    /* 1B END TRY         */  &&_END_TRY,
    /* 1C CATCH           */  &&_CATCH,
    /* 1D DUP             */  &&_DUP,
    /* 1E DROP            */  &&_DROP,
    /* 1F NEW             */  &&_NEW,
    /* 20 JUMP            */  &&_JUMP,
    /* 21 JUMP IF TRUE    */  &&_JUMP_IF_TRUE,
    /* 22 JUMP IF FALSE   */  &&_JUMP_IF_FALSE,
    /* 23 CALL            */  &&_CALL,
    /* 24 JUMP FIRST      */  &&_JUMP_FIRST,
    /* 25 JUMP NEXT       */  &&_JUMP_NEXT,
    /* 26 FIRST           */  &&_ENUM_FIRST,
    /* 27 NEXT            */  &&_ENUM_NEXT,
    /* 28                 */  &&_SUBR,
    /* 29                 */  &&_SUBR,
    /* 2A                 */  &&_SUBR,
    /* 2B                 */  &&_SUBR,
    /* 2C                 */  &&_SUBR,
    /* 2D                 */  &&_SUBR,
    /* 2E                 */  &&_SUBR,
    /* 2F                 */  &&_SUBR,
    /* 30                 */  &&_SUBR,
    /* 31                 */  &&_SUBR,
    /* 32                 */  &&_SUBR,
    /* 33                 */  &&_SUBR,
    /* 34                 */  &&_SUBR,
    /* 35                 */  &&_SUBR,
    /* 36                 */  &&_SUBR,
    /* 37                 */  &&_SUBR,
    /* 38                 */  &&_SUBR,
    /* 39                 */  &&_SUBR,
    /* 3A                 */  &&_SUBR,
    /* 3B                 */  &&_SUBR,
    /* 3C                 */  &&_SUBR,
    /* 3D                 */  &&_SUBR,
    /* 3E                 */  &&_SUBR,
    /* 3F                 */  &&_SUBR,
    /* 40                 */  &&_SUBR,
    /* 41                 */  &&_SUBR,
    /* 42                 */  &&_SUBR,
    /* 43                 */  &&_SUBR,
    /* 44                 */  &&_SUBR,
    /* 45                 */  &&_SUBR,
    /* 46                 */  &&_SUBR,
    /* 47                 */  &&_SUBR,
    /* 48                 */  &&_SUBR,
    /* 49                 */  &&_SUBR,
    /* 4A                 */  &&_SUBR,
    /* 4B                 */  &&_SUBR,
    /* 4C                 */  &&_SUBR,
    /* 4D                 */  &&_SUBR,
    /* 4E                 */  &&_SUBR,
    /* 4F                 */  &&_SUBR,
    /* 50                 */  &&_SUBR,
    /* 51                 */  &&_SUBR,
    /* 52                 */  &&_SUBR,
    /* 53                 */  &&_SUBR,
    /* 54                 */  &&_SUBR,
    /* 55                 */  &&_SUBR,
    /* 56                 */  &&_SUBR,
    /* 57                 */  &&_SUBR,
    /* 58                 */  &&_SUBR,
    /* 59                 */  &&_SUBR,
    /* 5A                 */  &&_SUBR,
    /* 5B                 */  &&_SUBR,
    /* 5C                 */  &&_SUBR,
    /* 5D                 */  &&_SUBR,
    /* 5E                 */  &&_SUBR,
    /* 5F                 */  &&_SUBR,
    /* 60                 */  &&_SUBR,
    /* 61                 */  &&_SUBR,
    /* 62                 */  &&_SUBR,
    /* 63                 */  &&_SUBR,
    /* 64                 */  &&_SUBR,
    /* 65                 */  &&_SUBR,
    /* 66                 */  &&_SUBR,
    /* 67                 */  &&_SUBR,
    /* 68                 */  &&_SUBR,
    /* 69                 */  &&_SUBR,
    /* 6A                 */  &&_SUBR,
    /* 6B                 */  &&_SUBR,
    /* 6C                 */  &&_SUBR,
    /* 6D                 */  &&_SUBR,
    /* 6E                 */  &&_SUBR,
    /* 6F                 */  &&_SUBR,
    /* 70                 */  &&_SUBR,
    /* 71                 */  &&_SUBR,
    /* 72                 */  &&_SUBR,
    /* 73                 */  &&_SUBR,
    /* 74                 */  &&_SUBR,
    /* 75                 */  &&_SUBR,
    /* 76                 */  &&_SUBR,
    /* 77                 */  &&_SUBR,
    /* 78                 */  &&_SUBR,
    /* 79                 */  &&_SUBR,
    /* 7A                 */  &&_SUBR,
    /* 7B                 */  &&_SUBR,
    /* 7C                 */  &&_SUBR,
    /* 7D                 */  &&_SUBR,
    /* 7E                 */  &&_SUBR,
    /* 7F                 */  &&_SUBR,
    /* 80                 */  &&_SUBR,
    /* 81                 */  &&_SUBR,
    /* 82                 */  &&_SUBR,
    /* 83                 */  &&_SUBR,
    /* 84                 */  &&_SUBR,
    /* 85                 */  &&_SUBR,
    /* 86                 */  &&_SUBR,
    /* 87                 */  &&_SUBR,
    /* 88                 */  &&_SUBR,
    /* 89                 */  &&_SUBR,
    /* 8A                 */  &&_SUBR,
    /* 8B                 */  &&_SUBR,
    /* 8C                 */  &&_SUBR,
    /* 8D                 */  &&_SUBR,
    /* 8E                 */  &&_SUBR,
    /* 8F                 */  &&_SUBR,
    /* 90                 */  &&_SUBR,
    /* 91                 */  &&_SUBR,
    /* 92                 */  &&_SUBR,
    /* 93                 */  &&_SUBR,
    /* 94                 */  &&_SUBR,
    /* 95                 */  &&_SUBR,
    /* 96                 */  &&_SUBR,
    /* 97                 */  &&_SUBR,
    /* 98                 */  &&_SUBR,
    /* 99                 */  &&_SUBR,
    /* 9A                 */  &&_SUBR,
    /* 9B                 */  &&_SUBR,
    /* 9C                 */  &&_SUBR,
    /* 9D                 */  &&_ILLEGAL,
    /* 9E                 */  &&_ILLEGAL,
    /* 9F                 */  &&_ILLEGAL,
    /* A0 ADD QUICK       */  &&_TODO,
    /* A1 ADD QUICK       */  &&_TODO,
    /* A2 ADD QUICK       */  &&_TODO,
    /* A3 ADD QUICK       */  &&_TODO,
    /* A4 ADD QUICK       */  &&_TODO,
    /* A5 ADD QUICK       */  &&_TODO,
    /* A6 ADD QUICK       */  &&_TODO,
    /* A7 ADD QUICK       */  &&_TODO,
    /* A8 ADD QUICK       */  &&_TODO,
    /* A9 ADD QUICK       */  &&_TODO,
    /* AA ADD QUICK       */  &&_TODO,
    /* AB ADD QUICK       */  &&_TODO,
    /* AC ADD QUICK       */  &&_TODO,
    /* AD ADD QUICK       */  &&_TODO,
    /* AE ADD QUICK       */  &&_TODO,
    /* AF ADD QUICK       */  &&_TODO,
    /* B0 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B1 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B2 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B3 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B4 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B5 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B6 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B7 PUSH CLASS      */  &&_PUSH_CLASS,
    /* B8 PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* B9 PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* BA PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* BB PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* BC PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* BD PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* BE PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* BF PUSH FUNCTION   */  &&_PUSH_FUNCTION,
    /* C0 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C1 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C2 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C3 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C4 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C5 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C6 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C7 PUSH DYNAMIC    */  &&_PUSH_DYNAMIC,
    /* C8 PUSH STATIC     */  &&_PUSH_STATIC,
    /* C9 PUSH STATIC     */  &&_PUSH_STATIC,
    /* CA PUSH STATIC     */  &&_PUSH_STATIC,
    /* CB PUSH STATIC     */  &&_PUSH_STATIC,
    /* CC PUSH STATIC     */  &&_PUSH_STATIC,
    /* CD PUSH STATIC     */  &&_PUSH_STATIC,
    /* CE PUSH STATIC     */  &&_PUSH_STATIC,
    /* CF PUSH STATIC     */  &&_PUSH_STATIC,
    /* D0 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D1 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D2 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D3 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D4 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D5 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D6 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D7 POP DYNAMIC     */  &&_POP_DYNAMIC,
    /* D8 POP STATIC      */  &&_POP_STATIC,
    /* D9 POP STATIC      */  &&_POP_STATIC,
    /* DA POP STATIC      */  &&_POP_STATIC,
    /* DB POP STATIC      */  &&_POP_STATIC,
    /* DC POP STATIC      */  &&_POP_STATIC,
    /* DD POP STATIC      */  &&_POP_STATIC,
    /* DE POP STATIC      */  &&_POP_STATIC,
    /* DF POP STATIC      */  &&_POP_STATIC,
    /* E0 PUSH CONST      */  &&_PUSH_CONST,
    /* E1 PUSH CONST      */  &&_PUSH_CONST,
    /* E2 PUSH CONST      */  &&_PUSH_CONST,
    /* E3 PUSH CONST      */  &&_PUSH_CONST,
    /* E4 PUSH CONST      */  &&_PUSH_CONST,
    /* E5 PUSH CONST      */  &&_PUSH_CONST,
    /* E6 PUSH CONST      */  &&_PUSH_CONST,
    /* E7 PUSH CONST      */  &&_PUSH_CONST,
    /* E8 PUSH CONST      */  &&_PUSH_CONST,
    /* E9 PUSH CONST      */  &&_PUSH_CONST,
    /* EA PUSH CONST      */  &&_PUSH_CONST,
    /* EB PUSH CONST      */  &&_PUSH_CONST,
    /* EC PUSH CONST      */  &&_PUSH_CONST,
    /* ED PUSH CONST      */  &&_PUSH_CONST,
    /* EE PUSH CONST      */  &&_PUSH_CONST,
    /* EF PUSH CONST      */  &&_PUSH_CONST,
    /* F0 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F1 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F2 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F3 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F4 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F5 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F6 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F7 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F8 PUSH QUICK      */  &&_PUSH_QUICK,
    /* F9 PUSH QUICK      */  &&_PUSH_QUICK,
    /* FA PUSH QUICK      */  &&_PUSH_QUICK,
    /* FB PUSH QUICK      */  &&_PUSH_QUICK,
    /* FC PUSH QUICK      */  &&_PUSH_QUICK,
    /* FD PUSH QUICK      */  &&_PUSH_QUICK,
    /* FE PUSH QUICK      */  &&_PUSH_QUICK,
    /* FF PUSH QUICK      */  &&_PUSH_QUICK
  };

  short ind;
  ushort uind;
  VALUE *val = NULL;
  TYPE type;

  void _pop_ctrl(short ind)
  {
    val = &BP[ind];
    RELEASE(val);

    SP--;

    if (SP->type == T_CSTRING)
      STRING_new_value(SP, SP->_string.addr + SP->_string.start, SP->_string.len);
    else if (SP->type == T_VARIANT && SP->_variant.vtype == T_CSTRING)
    {
      STRING_new((char **)SP->_variant.value, *((char **)SP->_variant.value), STRING_length(*((char **)SP->_variant.value)));
      SP->_variant.vtype = T_STRING;
    }

    *val = *SP;
    BORROW(val);
  }

  void _pop_value()
  {
    if (SP->type == T_CSTRING)
      STRING_new_value(SP, SP->_string.addr + SP->_string.start, SP->_string.len);
    else if (SP->type == T_VARIANT && SP->_variant.vtype == T_CSTRING)
    {
      STRING_new((char **)SP->_variant.value, *((char **)SP->_variant.value), STRING_length(*((char **)SP->_variant.value)));
      SP->_variant.vtype = T_STRING;
    }

    RELEASE(val);
    *val = *SP;
    BORROW(val);
  }

  goto _MAIN;

/*-----------------------------------------------*/

_NEXT3:

  PC++;

_NEXT2:

  PC++;

_NEXT:

  PC++;

/*-----------------------------------------------*/

_MAIN:

  /*
  if (EXEC_stop_next)
  {
    EXEC_stop_next = FALSE;
    TRACE_next();
  }

  if (EXEC_stop_next_line)
  {
    if ((EXEC_stop_function == NULL || FP == EXEC_stop_function)
        && (PC < EXEC_begin_line || PC >= EXEC_end_line))
    {
      EXEC_stop_next_line = FALSE;
      TRACE_main();
    }
  }
  */

/*_MAIN_BREAK:*/

#if DEBUG_PCODE
    TRACE_where();
#if DEBUG_STACK
      printf("[%d] ", SP - (VALUE *)STACK_base);
#endif
    if (*PC >> 8)
      PCODE_dump(PC - FP->code, PC);
    fflush(stdout);
#endif

  EXEC_code = *PC;
  goto *jump_table[EXEC_code >> 8];

/*-----------------------------------------------*/

_TODO:

  printf("** Not yet implemented\n");
  goto _NEXT;

/*-----------------------------------------------*/

_ILLEGAL:

  THROW(E_ILLEGAL);

/*-----------------------------------------------*/

_BREAK:

  ind = GET_XX();

  if (ind == 0)
  {
    if (!TRACE.stop)
      goto _NEXT;

    if (TRACE.fp != NULL)
    {
      if (BP > TRACE.bp)
        goto _NEXT;

      /*
      if (TRACE.fp != FP)
        goto _NEXT;
      */

      /*printf("BP = %p  bp = %p\n", BP, TRACE.bp);*/
    }
  }

  TRACE_breakpoint(ind);
  goto _NEXT;

/*-----------------------------------------------*/

_SUBR:

  (*SubrTable[(EXEC_code >> 8) - 0x28])();
  if (!PCODE_is_void(EXEC_code)) SP++;

  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_LOCAL:

  ind = GET_XX();

  /*
  if (ind >= 0)
    *SP = BP[ind];
  else
    *SP = PP[ind];
  */

  *SP = BP[ind];

  if (SP->type == T_STRING)
    SP->type = T_CSTRING;
  else if (SP->type == T_VARIANT && SP->_variant.vtype == T_STRING)
    SP->_variant.vtype = T_CSTRING;

  SP++;

  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_ARRAY:

  EXEC_push_array();
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_UNKNOWN:

  EXEC_push_unknown();
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_SPECIAL:

  EXEC_push_special();
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_EVENT:

  {
    OBJECT_EVENT *ev;
    OBJECT *parent;

    ind = GET_XX();
    ev = OBJECT_event(OP);

    ind = ev->event[ind];
    parent = ev->parent;

    if (parent == NULL || ind == 0)
    {
      /* fonction ne faisant rien si ce n'est renvoyer une valeur par dfaut */
      SP->_function.object = NULL;
      SP->_function.class = NULL;
      SP->_function.function = (long)(CP->load->event[GET_XX()].type);
      /* ??? Pourquoi n'utilise-t-on pas desc ? */
    }
    else
    {
      ind--;

      SP->type = T_FUNCTION;

      if (parent->class == CLASS_Class)
      {
        SP->_function.object = NULL;
        SP->_function.class = (CLASS *)parent;
        SP->_function.function = (long)((CLASS *)parent)->table[ind].desc;
      }
      else
      {
        SP->_function.object = parent;
        SP->_function.class = parent->class;
        SP->_function.function = (long)parent->class->table[ind].desc;
      }
    }

    SP->type = T_FUNCTION;
    SP++;
  }

  goto _NEXT;

/*-----------------------------------------------*/

_POP_LOCAL:

  ind = GET_XX();

  /*
  if (ind >= 0)
    val = &BP[ind];
  else
    val = &PP[ind];
  */

  val = &BP[ind];

  SP--;
  VALUE_conv(SP, val->type);

  _pop_value();

  goto _NEXT;

/*-----------------------------------------------*/

_POP_CTRL:

  ind = GET_XX();

  _pop_ctrl(GET_XX());

  goto _NEXT;

/*-----------------------------------------------*/

_POP_ARRAY:

  EXEC_pop_array();
  goto _NEXT;

/*-----------------------------------------------*/

_POP_UNKNOWN:

  EXEC_pop_unknown();
  goto _NEXT;

/*-----------------------------------------------*/

_POP_OPTIONAL:

  ind = GET_XX();

  /*
  if (ind >= 0)
    val = &BP[ind];
  else
    val = &PP[ind];
  */

  val = &BP[ind];

  if (val->type == T_VOID)
  {
    SP--;
    if (SP->type == T_VOID)
      VALUE_default(SP, val->_void.ptype);
    else
      VALUE_conv(SP, val->_void.ptype);
    /* RELEASE(val); Pas ncessaire */
    *val = *SP;
    BORROW(val);
  }
  else
    SP--;

  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_SHORT:

  SP->type = T_INTEGER;
  PC++;
  SP->_integer.value = *((short *)PC);
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_INTEGER:

  SP->type = T_INTEGER;
  PC++;
  SP->_integer.value = *((long *)PC);
  SP++;
  goto _NEXT2;

/*-----------------------------------------------*/

_PUSH_CHAR:

  ind = GET_XX();

  STRING_char_value(SP, (char)ind);
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_ME:

  if (OP == NULL)
  {
    SP->type = T_CLASS;
    SP->_class.class = CP;
  }
  else
  {
    SP->_object.class = CP;
    SP->_object.object = OP;
  }
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_NULL:

  SP->type = T_NULL;
  SP->_integer.value = 0;
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_EVENT:

  GAMBAS_DoNotRaiseEvent = (GET_XX() == 0);
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_RETURN:

  *SP++ = *RP;
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_BOOLEAN:

  ind = GET_XX();
  SP->type = T_BOOLEAN;
  SP->_integer.value = (ind != 0) ? -1 : 0;
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_LAST:

  SP->type = T_OBJECT;
  SP->_object.object = EVENT_Last;
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_VOID:

  SP->type = T_VOID;
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_DUP:

  *SP = SP[-1];
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_DROP:

  ind = EXEC_code & 0x3F;

  while (ind > 0)
  {
    SP--;
    RELEASE_STRING(SP);
    ind--;
  }
  
  goto _NEXT;

/*-----------------------------------------------*/

_NEW:

  EXEC_new();
  goto _NEXT;

/*-----------------------------------------------*/

_JUMP:

  PC += (signed short)PC[1] + 2;
  goto _MAIN;


/*-----------------------------------------------*/

_JUMP_IF_TRUE:

  SP--;
  VALUE_conv(SP, T_BOOLEAN);
  if (SP->_boolean.value)
  {
    PC += (signed short)PC[1] + 2;
    goto _MAIN;
  }

  goto _NEXT2;

/*-----------------------------------------------*/

_JUMP_IF_FALSE:

  SP--;
  VALUE_conv(SP, T_BOOLEAN);
  if (SP->_boolean.value == 0)
  {
    PC += (signed short)PC[1] + 2;
    goto _MAIN;
  }

  goto _NEXT2;

/*-----------------------------------------------*/

_RETURN:

  val = RP;

  ind = GET_XX();

  if (ind == 1)
  {
    SP--;
    VALUE_conv(SP, FP->type);

    val->type = T_NULL;

    _pop_value();

    EXEC_leave(SP);
  }
  else
  {
    VALUE_default(val, FP->type);
    EXEC_leave(NULL);
  }

  if (PC == NULL)
    return;

  goto _NEXT;

/*-----------------------------------------------*/

_CALL:

  ind = GET_3X();
  val = &SP[-(ind + 1)];
  /*EXEC_FUNCTION exec_func;*/

  if (val->type == T_CLASS)
  {
    EXEC_class(val->_class.class, ind, PCODE_is_void(EXEC_code));
    goto _NEXT;
  }

  if (!TYPE_is_function(val->type))
    THROW(E_NFUNC);

  if (val->_function.class == NULL)  /* Evnement qui ne fait rien */
  {
    while (ind > 0)
    {
      SP--;
      RELEASE_STRING(SP);
      ind--;
    }

    SP--;

    if (!PCODE_is_void(EXEC_code))
    {
      VALUE_default(SP, (TYPE)(val->_function.function));
      SP++;
    }

    goto _NEXT;
  }

  if ((val->_function.function & 0xFFFF0000)
      && ((long)((CLASS_DESC_METHOD *)val->_function.function)->exec & 0xFFFF0000))
  {
    EXEC_native(&val->_function, ind, PCODE_is_void(EXEC_code));
    goto _NEXT;
  }

  EXEC_enter(&val->_function, ind);
  goto _MAIN;

/*-----------------------------------------------*/

_JUMP_FIRST:

  ind = GET_XX();

  _pop_ctrl(ind + 1);
  _pop_ctrl(ind);

  PC[1] &= 0xFF00;
  goto _NEXT;

/*-----------------------------------------------*/

_JUMP_NEXT:

  ind = PC[-1] & 0xFF;

  {
    static void *jn_jump[5] = { &&_JN_START, &&_JN_NEXT_1, &&_JN_NEXT_2, &&_JN_NEXT_3, &&_JN_NEXT_4 };

    VALUE *end = &BP[ind];
    VALUE *inc = end + 1;

    ind = PC[2] & 0xFF;

    /*
    if (ind >= 0)
      val = &BP[ind];
    else
      val = &PP[ind];
    */
    val = &BP[ind];

    ind = GET_XX();
    goto *jn_jump[ind];

  _JN_START:

    type = val->type;

    VALUE_conv(end, type);
    VALUE_conv(inc, type);

    if (TYPE_is_integer(type))
    {
      if (inc->_integer.value > 0)
      {
        *PC |= 1;
        goto _JN_TEST_1;
      }
      else
      {
        *PC |= 2;
        goto _JN_TEST_2;
      }
    }
    else if (TYPE_is_float(type))
    {
      if (inc->_float.value > 0)
      {
        *PC |= 3;
        goto _JN_TEST_3;
      }
      else
      {
        *PC |= 4;
        goto _JN_TEST_4;
      }
    }
    else
      THROW(E_TYPE, "Number", TYPE_get_name(type));

  _JN_NEXT_1:

    val->_integer.value += inc->_integer.value;

  _JN_TEST_1:

    if (val->_integer.value <= end->_integer.value)
      goto _NEXT3;

    goto _JN_END;

  _JN_NEXT_2:

    val->_integer.value += inc->_integer.value;

  _JN_TEST_2:

    if (val->_integer.value >= end->_integer.value)
      goto _NEXT3;

    goto _JN_END;

  _JN_NEXT_3:

    val->_float.value += inc->_float.value;

  _JN_TEST_3:

    if (val->_float.value <= end->_float.value)
      goto _NEXT3;

    goto _JN_END;

  _JN_NEXT_4:

    val->_float.value += inc->_float.value;

  _JN_TEST_4:

    if (val->_float.value >= end->_float.value)
      goto _NEXT3;

    goto _JN_END;

  _JN_END:

    PC += (signed short)PC[1] + 2;
    goto _MAIN;
  }

/*-----------------------------------------------*/

_ENUM_FIRST:

  _pop_ctrl(GET_XX());
  EXEC_enum_first();
  goto _NEXT;

/*-----------------------------------------------*/

_ENUM_NEXT:

  if (EXEC_enum_next())
  {
    SP--;
    RELEASE_STRING(SP);

    goto _JUMP;
  }

  goto _NEXT2;

/*-----------------------------------------------*/

_PUSH_CLASS:

  {
    CLASS *class = CP->load->class_ref[GET_7XX()];

    /*
    if (GET_XX() == 11)
      printf("class = %s\n", class->name);
    */

    /* CLASS_load(class); effectu dans exec_object() et exec_new() */

    SP->type = T_CLASS;
    SP->_class.class = class;
    SP++;
  }
  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_FUNCTION:

  uind = GET_7XX();

  SP->type = T_FUNCTION;
  SP->_function.class = CP;
  SP->_function.object = OP;
  SP->_function.function = uind;
  SP++;

  goto _NEXT;

/*-----------------------------------------------*/

  {
    CLASS_VAR *var;
    char *addr;

_PUSH_DYNAMIC:

    var = &CP->load->dyn[GET_7XX()];

    if (OP == NULL)
      THROW(E_ILLEGAL);

    addr = &OP[var->pos];
    goto __READ;

_PUSH_STATIC:

    var = &CP->load->stat[GET_7XX()];
    addr = (char *)CP->stat + var->pos;
    goto __READ;

__READ:

    VALUE_class_read(CP, SP, addr, var->type);

    SP++;
    goto _NEXT;


_POP_DYNAMIC:

    var = &CP->load->dyn[GET_7XX()];

    if (OP == NULL)
      THROW(E_ILLEGAL);

    addr = &OP[var->pos];
    goto __WRITE;

_POP_STATIC:

    var = &CP->load->stat[GET_7XX()];
    addr = (char *)CP->stat + var->pos;
    goto __WRITE;

__WRITE:

    SP--;
    VALUE_class_write(CP, SP, addr, var->type);

    goto _NEXT;

  }

/*-----------------------------------------------*/

_PUSH_CONST:

  ind = (short)GET_XXX();

  if (ind >= 0)
  {
    CLASS_CONST *cc = &CP->load->cst[ind];

    if (cc->type == T_FLOAT)
    {
      SP->type = cc->type;
      SP->_float.value = cc->_float.value;
    }
    else if (cc->type == T_STRING)
    {
      SP->type = T_CSTRING;
      SP->_string.addr = cc->_string.addr;
      SP->_string.start = 0;
      SP->_string.len = cc->_string.len;
    }
    else
    {
      SP->type = cc->type;
      SP->_integer.value = cc->_integer.value;
    }
  }
  else
  {
    EXEC_ILLEGAL();
  }

  SP++;

  goto _NEXT;

/*-----------------------------------------------*/

_PUSH_QUICK:

  SP->type = T_INTEGER;
  SP->_integer.value = GET_XXX();
  SP++;
  goto _NEXT;

/*-----------------------------------------------*/

_TRY:

  ERROR_code = 0;
  EP = SP;
  EC = PC + (signed short)PC[1] + 2;

  #if DEBUG_ERROR
  printf("TRY %p\n", EC);
  #endif

  goto _NEXT2;

/*-----------------------------------------------*/

_END_TRY:

  #if DEBUG_ERROR
  printf("END TRY %p\n", PC);
  #endif

  EP = NULL;
  if (FP->error)
    EC = FP->code + FP->error;
  else
    EC = NULL;
  goto _NEXT;

/*-----------------------------------------------*/

_CATCH:

  if (EC == NULL)
    goto _NEXT;
  else
    goto _RETURN;
}

