/***************************************************************************

  file.c

  The file management routines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __FILE_C

#include "common.h"
#include "error.h"
#include "value.h"
#include "limit.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <dirent.h>

#include "exec.h"
#include "project.h"
#include "gambas.h"
#include "regexp.h"
#include "String.h"

#include "file.h"

PUBLIC FILE_STAT FILE_stat_info = { 0 };

PRIVATE char file_buffer[MAX_PATH + 1];
PRIVATE int file_buffer_length;

PRIVATE DIR *file_dir = NULL;
PRIVATE char *file_pattern = NULL;


PRIVATE void dir_exit(void)
{
  if (file_dir != NULL)
  {
    closedir(file_dir);
    file_dir = NULL;
  }

  STRING_free(&file_pattern);
}


PUBLIC void FILE_exit(void)
{
  dir_exit();
}


PUBLIC char *FILE_cat(char *path, ...)
{
  char *p;
  va_list args;
  int len;
  boolean end_slash = FALSE;
  boolean add_slash = FALSE;

  va_start(args, path);

  p = file_buffer;

  if (path != file_buffer)
    *p = 0;

  for(;;)
  {
    if (*path == '/' && p != file_buffer)
      path++;

    len = strlen(path);
    if (add_slash)
      len++;

    if (len > 0)
    {
      if ((p + len) > &file_buffer[MAX_PATH])
        return NULL;

      if (p != path)
      {
        if (add_slash)
          p = stradd(p, "/");

          p = stradd(p, path);
      }
      else
        p += len;

      end_slash = (p[-1] == '/');
    }

    path = va_arg(args, char *);
    if (path == NULL)
      break;

    add_slash = ((!end_slash) && (*path != 0) && (*path != '/'));
  }

  file_buffer_length = p - file_buffer;
  return file_buffer;
}


PUBLIC int FILE_buffer_length(void)
{
  if (file_buffer_length < 0)
    file_buffer_length = strlen(file_buffer);

  return file_buffer_length;
}


PUBLIC char *FILE_get_dir(char *path)
{
  char *p;

  if (path == NULL)
    return NULL;

  if (file_buffer != path)
    strcpy(file_buffer, path);

  p = rindex(file_buffer, '/');

  if (p == NULL)
    *file_buffer = 0;
  else
    *p = 0;

  file_buffer_length = -1;
  return file_buffer;
}


PUBLIC char *FILE_get_name(char *path)
{
  char *p;

  p = rindex(path, '/');
  if (p)
    return &p[1];
  else
    return path;
}


PUBLIC char *FILE_get_ext(char *path)
{
  char *p;

  p = rindex(path, '/');
  if (p)
    path = &p[1];

  p = rindex(path, '.');
  if (p == NULL)
    return &path[strlen(path)];
  else
    return p + 1;
}


PUBLIC char *FILE_set_ext(char *path, char *ext)
{
  char *p;

  if (path != file_buffer)
  {
    strcpy(file_buffer, path);
    path = file_buffer;
  }

  p = FILE_get_ext(path);

  if (ext == NULL || *ext == 0)
    *p = 0;
  else
    strcpy(p, ext);

  file_buffer_length = -1;
  return path;
}


PUBLIC char *FILE_get_basename(char *path)
{
  char *p;

  path = FILE_get_name(path);

  if (file_buffer != path)
    strcpy(file_buffer, path);

  p = rindex(file_buffer, '.');
  if (p)
    *p = 0;

  file_buffer_length = -1;

  return file_buffer;
}


PUBLIC char *FILE_get(char *path)
{
  if (!FILE_is_absolute(path) && !EXEC_arch)
    path = FILE_cat(PROJECT_path, path, NULL);

  return path;
}


PUBLIC boolean FILE_exist(char *path)
{
  struct stat buf;

  return (stat(path, &buf) == 0);
}


PUBLIC void FILE_stat(char *path)
{
  struct stat buf;

  if (lstat(path, &buf) != 0)
    THROW_SYSTEM(errno, path);

  if (S_ISREG(buf.st_mode))
    FILE_stat_info.type = GB_STAT_FILE;
  else if (S_ISDIR(buf.st_mode))
    FILE_stat_info.type = GB_STAT_DIRECTORY;
  else if (S_ISCHR(buf.st_mode) || S_ISBLK(buf.st_mode))
    FILE_stat_info.type = GB_STAT_DEVICE;
  else if (S_ISFIFO(buf.st_mode))
    FILE_stat_info.type = GB_STAT_PIPE;
  else if (S_ISSOCK(buf.st_mode))
    FILE_stat_info.type = GB_STAT_SOCKET;
  else if (S_ISLNK(buf.st_mode))
    FILE_stat_info.type = GB_STAT_LINK;

  FILE_stat_info.mode = buf.st_mode & 07777;
  FILE_stat_info.size = (long)buf.st_size;
  FILE_stat_info.time = (long)buf.st_mtime;
  FILE_stat_info.hidden = (*FILE_get_name(path) == '.');
}


PUBLIC void FILE_dir_first(char *path, char *pattern, long len_pattern)
{
  dir_exit();

  file_dir = opendir(path);
  if (file_dir == NULL)
    THROW_SYSTEM(errno, path);

  STRING_new(&file_pattern, pattern, len_pattern);
}


PUBLIC char *FILE_dir_next(void)
{
  struct dirent *entry;

  if (file_dir == NULL)
    return NULL;

  for(;;)
  {
    entry = readdir(file_dir);
    if (entry == NULL)
    {
      dir_exit();
      return NULL;
    }

    if ((strcmp(entry->d_name, ".") == 0) || (strcmp(entry->d_name, "..") == 0))
      continue;

    if (file_pattern == NULL)
      break;

    if (REGEXP_match(file_pattern, STRING_length(file_pattern), entry->d_name, strlen(entry->d_name)))
      break;
  }

  return entry->d_name;
}


PUBLIC void FILE_unlink(char *path)
{
  if (!FILE_is_absolute(path))
    THROW(E_ACCESS);

  if (unlink(path) != 0)
    THROW_SYSTEM(errno, path);
}


PUBLIC void FILE_rmdir(char *path)
{
  if (!FILE_is_absolute(path))
    THROW(E_ACCESS);

  if (rmdir(path) != 0)
    THROW_SYSTEM(errno, path);
}


PUBLIC void FILE_mkdir(char *path)
{
  if (!FILE_is_absolute(path))
    THROW(E_ACCESS);

  if (mkdir(path, 0777) != 0)
    THROW_SYSTEM(errno, path);
}


PUBLIC void FILE_rename(char *src, char *dst)
{
  if (!FILE_is_absolute(src) || !FILE_is_absolute(dst))
    THROW(E_ACCESS);

  if (FILE_exist(dst))
    THROW(E_EXIST, dst);

  if (rename(src, dst) != 0)
    THROW_SYSTEM(errno, dst);
}


PUBLIC char *FILE_getcwd(char *subdir)
{
  if (getcwd(file_buffer, PATH_MAX) == NULL)
    return NULL;

  file_buffer_length = strlen(file_buffer);

  if (subdir != NULL)
    return FILE_cat(file_buffer, subdir, NULL);
  else
    return file_buffer;
}


PUBLIC char *FILE_readlink(char *link)
{
  int len = readlink(link, file_buffer, MAX_PATH);

  if (len < 0)
    return NULL;

  file_buffer[len] = 0;
  file_buffer_length = len;
  return file_buffer;

}
