/***************************************************************************

  gambas_api.c

  Gambas API for external libraries

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __GAMBAS_API_C

#include "common.h"
#include "error.h"
#include "memory.h"

#include <stdarg.h>

#include "class.h"
#include "exec.h"
#include "event.h"
#include "stack.h"
#include "stream.h"
#include "library.h"
#include "watch.h"

#include "Object.h"
#include "String.h"
#include "Date.h"

#include "gambas.h"
#include "gambas_api.h"


typedef
  struct {
    OBJECT *object;
    CLASS_DESC_METHOD *desc;
    }
  GB_API_FUNCTION;


PUBLIC void *GAMBAS_Api[] =
{
  (void *)GB_VERSION,

  (void *)GB_GetInterface,

  (void *)GB_Declare,
  (void *)GB_Hook,

  (void *)GB_GetFunction,
  (void *)GB_Call,
  (void *)GB_Raise,
  (void *)GB_CanRaise,

  (void *)GB_Error,

  (void *)GB_GetClass,
  (void *)GB_FindClass,
  (void *)GB_Is,
  (void *)GB_Ref,
  (void *)GB_Unref,
  (void *)GB_UnrefKeep,
  (void *)GB_Detach,
  (void *)GB_Attach,
  (void *)GB_New,
  (void *)GB_CheckObject,

  (void *)GB_GetEnum,
  (void *)GB_StopEnum,

  (void *)GB_Return,
  (void *)GB_ReturnInt,
  (void *)GB_ReturnBoolean,
  (void *)GB_ReturnDate,
  (void *)GB_ReturnObject,
  (void *)GB_ReturnNull,
  (void *)GB_ReturnPtr,

  (void *)GB_ReturnString,
  (void *)GB_ReturnZeroString,
  (void *)GB_ReturnCopyString,
  (void *)GB_ReturnNewString,

  (void *)GB_NewString,
  (void *)GB_CopyString,
  (void *)GB_FreeString,
  (void *)GB_ToZeroString,

  (void *)GB_IsMissing,
  (void *)GB_GetReturnType,

  (void *)GB_LoadFile,
  (void *)GB_ReleaseFile,

  (void *)GB_Store,
  (void *)GB_StoreString,
  (void *)GB_StoreObject,

  (void *)GB_Watch,

  NULL
};

PUBLIC TYPE GAMBAS_ReturnType;
PUBLIC unsigned long GAMBAS_MissingParam;
PUBLIC char *GAMBAS_Error = NULL;
PUBLIC boolean GAMBAS_DoNotRaiseEvent = FALSE;

#define NBR_ZERO_STRING 8
PRIVATE char *gambas_zero_string[NBR_ZERO_STRING] = { NULL };


PRIVATE void free_zero_string(void)
{
  int i;

  for (i = 0; i < NBR_ZERO_STRING; i++)
  {
    STRING_free(&gambas_zero_string[i]);
    gambas_zero_string[i] = NULL;
  }
}

PUBLIC void GAMBAS_exit(void)
{
  free_zero_string();
}


/*!------------------------------------------------------------

\HELP
gb-get-interface
function

\IN
gambas

\TITLE
GB.GetInterface

\DESC:fr

Rcupre l'interface de programmation d'une extension afin de l'utiliser
dans une autre extension.

\SYNTAX:fr

int <B>GB.GetInterface</B>(char *<I>extension</I>, long <I>version</I>, void *<I>interface</I>)

\ARG:fr

<TR><TD><I>extension</I></TD><TD>Nom de l'extension.</TD><TR>
<TR><TD><I>version</I></TD><TD>Version attendue de l'extension.</TD><TR>
<TR><TD><I>interface</I></TD><TD>Adresse de la structure o sera stocke l'interface.</TD><TR>

\RETURN
La fonction retourne 0 si tout s'est bien pass, et une valeur non nulle en cas de problmes
(Par exemple, si l'extension est introuvable).

\EXAMPLE

#include\<gambas.h\>
#include\<gbqt.h\>

...

QT__INTERFACE QT;

...

int GB__INIT(void)
{
  ...

  GB.GetInterface("QT", 1, &QT);

  ...
}

-------------------------------------------------------------*/

PUBLIC int GB_GetInterface(char *library, long version, void *iface)
{
  if (LIBRARY_get_interface(library, version, iface))
    ERROR_panic("Cannot find interface of library '%s'", library);

  return FALSE;
}


/*!------------------------------------------------------------

\HELP
gb-hook
function

\IN
gambas

\TITLE
GB.Hook

\DESC:fr

Permet d'intercepter certaines des fonctionnalits de l'interprteur.

\SYNTAX:fr

void *<B>GB.Hook</B>(int <I>type</I>, void *<I>fonction</I>)

\ARG:fr

<TR><TD><I>type</I></TD><TD>Type de fonctionnalit  driver.</TD><TR>
<TR><TD><I>fonction</I></TD><TD>Pointeur sur la fonction d'interception.</TD><TR>

\RETURN
La fonction retourne le pointeur sur la fonction d'interception pralablement
mise en place.

\REMARK

<P>Voici les diffrentes fonctionnalits interceptables, selon la valeur du paramtre
<I>type</I> :</P>

<UL>
  <LI>_gb-hook-main_
  <LI>_gb-hook-loop_
  <LI>_gb-hook-wait_
  <LI>_gb-hook-message_
  <LI>_gb-hook-beep_
  <LI>_gb-hook-watch_
</UL>

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
gb-hook-main
function

\IN
gambas

\TITLE
GB\_HOOK\_MAIN

\DESC:fr

Cette fonction est appele pour interprter les arguments de la ligne de commande.

\SYNTAX:fr

void <I>hook\_main</I>(int *<I>argc, char **argv)

\ARG:fr

<TR><TD><I>argc</I></TD><TD>Pointeur sur une variable contenant le nombre d'arguments.</TD><TR>
<TR><TD><I>argv</I></TD><TD>Pointeur sur un tableau de chanes de caractres
reprsentant chaque argument.</TD><TR>

\REMARK
La fonction peut ainsi modifier <I>argc</I> et <I>argv</I> de manire  retirer
les paramtres qu'elle a su interprter. Les paramtres restants sont passs au
programme interprt qui peut les rcuprer en utilisant la proprit _app-args_
de la classe _app_.

\SEE
_gb-hook_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
gb-hook-loop
function

\IN
gambas

\TITLE
GB\_HOOK\_LOOP

\DESC:fr

Cette fonction est appele pour grer la boucle d'vnements.

\SYNTAX:fr

int <I>hook\_loop</I>(void)

\RETURN
Cette fonction ne se termine que pour quitter le programme. Sa valeur de retour sera
la valeur renvoye par la fonction <I>main</I> de l'interprteur.

\REMARK
Il ne peut y avoir qu'une et une seule boucle d'vnement. Celle-ci se dclenche
aprs que la fonction <I>Main()</I> du programme GAMBAS se soit termine.

<P>L'interprteur possde une boucle d'vnement par dfaut qui ne sert qu' surveiller
les descripteurs de fichiers. Cette boucle se termine lorsqu'il n'y a plus de
descripteurs de fichiers  surveiller.</P>

\SEE
_gb-hook_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
gb-hook-wait
function

\IN
gambas

\TITLE
GB\_HOOK\_WAIT

\DESC:fr

Cette fonction est appele pour suspendre le programme pendant
la dure spcifie. Pendant ce laps de temps, la main est rendue
 la boucle de traitement des vnements.

\SYNTAX:fr

void <I>hook\_wait</I>(long <I>dure</I>)

\ARG:fr

<TR><TD><I>dure</I></TD><TD>Dure de l'attente en millisecondes.</TD><TR>

\REMARK
si <I>dure</I> vaut 0, alors le programme est suspendue jusqu' ce qu'un
vnement survienne.

\SEE
_gb-hook_

-------------------------------------------------------------*/


PUBLIC void *GB_Hook(int type, void *hook)
{
  void *old_hook;

  if ((type < GB_HOOK_MAIN) || (type > GB_HOOK_WATCH))
    return NULL;

  type--;
  old_hook = ((void **)&EXEC_Hook)[type];
  ((void **)&EXEC_Hook)[type] = hook;

  return old_hook;
}


/*!------------------------------------------------------------

\HELP
gb-declare
function

\IN
gambas

\TITLE
GB.Declare

\DESC:fr

Dclare  l'interprteur l'implmentation de nouvelles classes
natives.

\SYNTAX:fr

void <B>GB.Declare</B>(GAMBAS__DESC **<I>desc</I>, ...)

\ARG:fr

<TR><TD><I>desc</I></TD><TD>Liste de descriptions de classes se terminant par un pointeur <I>NULL</I>.</TD><TR>

-------------------------------------------------------------*/

PUBLIC void GB_Declare(void *desc, ...)
{
  va_list args;

  va_start(args, desc);

  while (desc != NULL)
  {
    if (CLASS_register(desc) == NULL)
      THROW(E_REGISTER, *((char **)desc));

    desc = va_arg(args, void *);
  }
}


/*!------------------------------------------------------------

\HELP
gb-raise
function

\IN
gambas

\TITLE
GB.Raise

\DESC:fr

Dclenche un vnenement.

\SYNTAX:fr

int <B>GB.Raise</B>(void *<I>objet</I>, int <I>vnement</I>, char *<I>paramtres</I>)

\ARG:fr

<TR><TD><I>objet</I></TD><TD>Pointeur sur l'objet qui recevra l'vnement.</TD><TR>
<TR><TD><I>vnement</I></TD><TD>Numro de l'vnenement, rcupr lors de la dclaration de la classe.</TD><TR>
<TR><TD><I>paramtres</I></TD><TD>Pointeur sur une structure contenant les paramtres
de l'vnement tel que dcrit dans la dclaration de la classe, ou <I>NULL</I> si l'vnement
ne reoit aucun paramtre.</TD><TR>

-------------------------------------------------------------*/

PRIVATE void call_method(void *object, CLASS_DESC_METHOD *desc, char *param)
{
  VALUE_FUNCTION func;
  VALUE *save_SP;
  int nparam;
  int i;

  if (OBJECT_is_class(object))
  {
    func.object = NULL;
    func.class = (CLASS *)object;
  }
  else
  {
    func.object = object;
    func.class = OBJECT_class(object);
  }

  func.function = (long)desc;

  save_SP = SP;
  nparam = desc->npmin;

  STACK_check(nparam);

  for (i = 0; i < nparam; i++)
  {
    if (TYPE_is_string(desc->signature[i]))
    {
      SP->type = T_CSTRING;
      SP->_string.addr = ((GB_STRING *)param)->addr;
      SP->_string.start = 0;
      SP->_string.len = ((GB_STRING *)param)->len;
      param += sizeof(GB_STRING);
    }
    else
    {
      VALUE_read(SP, param, desc->signature[i]);
      param += TYPE_sizeof(desc->signature[i]);
    }
    SP++;
  }

  EXEC_function(&func, nparam);

  SP = save_SP;
}


PUBLIC int GB_CanRaise(void *object, int event_id)
{
  OBJECT *parent;
  short *event_tab;
  int func_id;

  if (object == NULL)
    return FALSE;

  event_tab = OBJECT_event(object)->event;
  parent = OBJECT_event(object)->parent;

  if (parent == NULL)
    return FALSE;

  func_id = event_tab[event_id];

  if (func_id == 0)
    return FALSE;

  return TRUE;
}


PUBLIC int GB_Raise(void *object, int event_id, char *param)
{
  OBJECT *parent;
  CLASS *class;
  int func_id;
  CLASS_DESC_METHOD *desc;
  short *event_tab;


  /*MEMORY_check_ptr(object);*/

  if (GAMBAS_DoNotRaiseEvent)
    return FALSE;

  if (object == NULL)
    return FALSE;

  event_tab = OBJECT_event(object)->event;
  func_id = event_tab[event_id];
  parent = OBJECT_event(object)->parent;
  class = OBJECT_class(object);

#if DEBUG_EVENT
    printf("GB_Raise(%p, %d, %s, %p)\n", object, event_id, class->event[event_id].name, param);
    printf("func_id = %d  parent = %p  parent class = %p\n", func_id, parent, parent->class);
    fflush(NULL);
#endif

  if (parent == NULL || !func_id)
    return FALSE;

  func_id--;

  if (OBJECT_is_class(parent))
    desc = &(((CLASS *)parent)->table[func_id].desc->method);
  else
    desc = &(parent->class->table[func_id].desc->method);

  EVENT_Last = object;

  call_method(parent, desc, param);

  if (RP->type == T_VOID)
    return 0;
  else
    return (RP->_boolean.value != 0 ? 1 : 0);
}


/*!------------------------------------------------------------

\HELP
gb-error
function

\IN
gambas

\TITLE
GB.Error

\DESC:fr

Dclenche une erreur.

\SYNTAX:fr

void <B>GB.Error</B>(char *<I>format</I>, ...)

\ARG:fr

<TR><TD><I>format</I></TD><TD>Chane de format correspondant au message d'erreur.</TD><TR>

\REMARK

<P>La syntaxe de cette fonction est quivalente  celle de la fonction C <I>printf</I>.</P>

<P>Vous spcifiez  la place d'une chane de format une des constantes suivantes, qui
correspondent  des messages d'erreurs prdfinis :</P>

<TABLE BORDER=1 CELLPADDING=2 WIDTH=400>
  <TR><TD WIDTH=160>GB_ERR_TYPE</TD><TD>Type de donnes incorrect.</TD></TR>
</TABLE><BR>

-------------------------------------------------------------*/

PUBLIC void GB_Error(char *error)
{
  GAMBAS_Error = error;
}


/*!------------------------------------------------------------

\HELP
gb-ref
function

\IN
gambas

\TITLE
GB.Ref

\DESC:fr

Incrmente le compteur de rfrences d'un objet.

\SYNTAX:fr

void <B>GB.Ref</B>(void *<I>objet</I>)

\ARG:fr

<TR><TD><I>objet</I></TD><TD>Pointeur sur un objet.</TD><TR>

\REMARK
Il est ncessaire de rfrencer tout objet dont on conserve
un pointeur pendant l'excution du programme interprt.

\SEE
_gb-unref_

-------------------------------------------------------------*/

PUBLIC void GB_Ref(void *object)
{
  #if DEBUG_ALLOC
  CLASS *save = CP;
  CP = NULL;
  #endif

  if (object != NULL)
    OBJECT_REF(object, "GB_Ref");

  #if DEBUG_ALLOC
  CP = save;
  #endif
}

/*!------------------------------------------------------------

\HELP
gb-unref
function

\IN
gambas

\TITLE
GB.Unref

\DESC:fr

Dcrmente le compteur de rfrences d'un objet. Si le compteur
de rfrence de l'objet devient nul, celui-ci est libr.

\SYNTAX:fr

void <B>GB.Unref</B>(void *<I>objet</I>)

\ARG:fr

<TR><TD><I>objet</I></TD><TD>Pointeur sur un objet.</TD><TR>

\SEE
_gb-detach_ _gb-ref_

-------------------------------------------------------------*/

PUBLIC void GB_Unref(void **object)
{
  #if DEBUG_ALLOC
  CLASS *save = CP;
  CP = NULL;
  #endif

  if (*object != NULL)
    OBJECT_UNREF(object, "GB_Unref");

  #if DEBUG_ALLOC
  CP = save;
  #endif
}


PUBLIC void GB_UnrefKeep(void **object, int delete)
{
  #if DEBUG_ALLOC
  CLASS *save = CP;
  CP = NULL;
  #endif

  if (*object != NULL)
  {
    if (delete)
      OBJECT_UNREF(object, "GB_UnrefKeep");
    else
      OBJECT_UNREF_KEEP(object, "GB_UnrefKeep");
  }

  #if DEBUG_ALLOC
  CP = save;
  #endif
}


/*!------------------------------------------------------------

\HELP
gb-detach
function

\IN
gambas

\TITLE
GB.Detach

\DESC:fr

Indique  l'interprteur que l'objet en question ne lancera
plus d'vnements.

\SYNTAX:fr

void <B>GB.Detach</B>(void *<I>objet</I>)

\ARG:fr

<TR><TD><I>objet</I></TD><TD>Pointeur sur un objet.</TD><TR>

\SEE
_gb-unref_

-------------------------------------------------------------*/

PUBLIC void GB_Detach(void *object)
{
  if (object)
    OBJECT_detach(object);
}

PUBLIC void GB_Attach(void *object, void *parent, char *name)
{
  if (object)
    OBJECT_attach(object, parent, name);
}


PUBLIC void GB_StopEnum(void)
{
  EXEC_enum_stop = TRUE;
  GB_Return(T_NULL);
}

PUBLIC void *GB_GetEnum(void)
{
  return (void *)EXEC_enum_data;
}


PUBLIC void GB_Return(ulong type, ...)
{
  static void *jump[16] = {
    &&_VOID, &&_BOOLEAN, &&_BYTE, &&_SHORT, &&_INTEGER, &&_LONG, &&_FLOAT, &&_DATE,
    &&_STRING, &&_STRING, &&_VARIANT, &&_ARRAY, &&_STRUCT, &&_FUNCTION, &&_CLASS, &&_NULL
    };

  VALUE *ret = &TEMP;
  va_list args;
  long val;

  va_start(args, type);

  ret->type = type;
  if (TYPE_is_object(type))
    goto _OBJECT;
  else
    goto *jump[type];

_BOOLEAN:

  ret->_integer.value = va_arg(args, int) ? (-1) : 0;
  goto _CONV;

_BYTE:

  ret->_integer.value = va_arg(args, int);
  goto _CONV;

_SHORT:

  ret->_integer.value = va_arg(args, int);
  goto _CONV;

_INTEGER:

  ret->_integer.value = va_arg(args, long);
  goto _CONV;

_FLOAT:

  ret->_float.value = va_arg(args, double);
  goto _CONV;

_STRING:

  ret->type = T_CSTRING;
  ret->_string.addr = va_arg(args, char *);
  ret->_string.start = 0;
  if (ret->_string.addr != NULL)
    ret->_string.len = strlen(ret->_string.addr);
  else
    ret->_string.len = 0;
  goto _CONV;

_DATE:

  val = va_arg(args, long);
  DATE_from_time(ret, (time_t )val, 0);
  goto _CONV;

_OBJECT:

  ret->_object.object = va_arg(args, void *);
  goto _CONV;

_CONV:

  VALUE_conv(ret, GAMBAS_ReturnType);

_VOID:
_LONG:
_VARIANT:
_ARRAY:
_STRUCT:
_FUNCTION:
_CLASS:
_NULL:
}


PUBLIC void GB_ReturnInt(long val)
{
  GB_Return(T_INTEGER, val);
}


PUBLIC void GB_ReturnDate(long val)
{
  GB_Return(T_DATE, val);
}


PUBLIC void GB_ReturnBoolean(int val)
{
  GB_Return(T_BOOLEAN, (char)val);
}


PUBLIC void GB_ReturnObject(void *val)
{
  GB_Return(GAMBAS_ReturnType, val);
}


PUBLIC void GB_ReturnPtr(ulong type, void *value)
{
  if (type == T_VOID)
    return;

  VALUE_read(&TEMP, value, type);
  VALUE_conv(&TEMP, GAMBAS_ReturnType);
}


PUBLIC char *GB_NewString(long len)
{
  char *str;

  STRING_new(&str, NULL, len);
  return str;
}


PUBLIC char *GB_CopyString(char *copy, long len)
{
  char *str;

  if (copy == NULL)
    return NULL;

  STRING_new(&str, copy, (len <= 0) ? strlen(copy) : len);
  return str;
}


PUBLIC void GB_FreeString(char *str)
{
  STRING_free(&str);
}


PUBLIC char *GB_ToZeroString(GB_STRING src)
{
  static int current = 0;
  char *str;

  STRING_free(&gambas_zero_string[current]);
  str = GB_CopyString(src.addr, src.len);
  gambas_zero_string[current] = str;

  current++;
  if (current >= NBR_ZERO_STRING)
    current = 0;

  return str;
}


PUBLIC void GB_ReturnString(char *str, long len)
{
  TEMP.type = T_CSTRING;
  TEMP._string.addr = str;
  TEMP._string.start = 0;
  TEMP._string.len = (len <= 0) ? (str ? strlen(str) : 0) : len;

  if (TEMP._string.len == 0)
    TEMP._string.addr = 0;

  VALUE_conv(&TEMP, GAMBAS_ReturnType);
}


PUBLIC void GB_ReturnZeroString(char *str)
{
  TEMP.type = T_CSTRING;
  TEMP._string.addr = str;
  TEMP._string.start = 0;
  TEMP._string.len = str ? strlen(str) : 0;

  if (TEMP._string.len == 0)
    TEMP._string.addr = 0;

  VALUE_conv(&TEMP, GAMBAS_ReturnType);
}


PUBLIC void GB_ReturnNewString(char *str)
{
  TEMP.type = T_STRING;
  TEMP._string.addr = str;
  TEMP._string.start = 0;
  TEMP._string.len = STRING_length(str);

  VALUE_conv(&TEMP, GAMBAS_ReturnType);
}


PUBLIC void GB_ReturnCopyString(char *str, long len)
{
  TEMP.type = T_STRING;

  if (str != NULL && len <= 0)
    len = strlen(str);

  if (str == NULL || len == 0)
  {
    TEMP._string.addr = 0;
    TEMP._string.start = 0;
    TEMP._string.len = 0;
  }
  else
  {
    STRING_new(&TEMP._string.addr, str, len);
    TEMP._string.start = 0;
    TEMP._string.len = len;
  }

  VALUE_conv(&TEMP, GAMBAS_ReturnType);
}

PUBLIC void GB_ReturnNull(void)
{
  TEMP.type = T_NULL;

  VALUE_conv(&TEMP, GAMBAS_ReturnType);
}


PUBLIC void *GB_GetClass(void *object)
{
  return OBJECT_class(object);
}


PUBLIC void *GB_FindClass(char *name)
{
  return CLASS_find(name);
}


PUBLIC int GB_Is(void *object, void *class)
{
  CLASS *ob_class = OBJECT_class(object);

  return ((ob_class == class) || CLASS_inherits(ob_class, class));
}


PUBLIC long GB_GetReturnType(void)
{
  return GAMBAS_ReturnType;
}


PUBLIC void GB_LoadFile(char *path, char **addr, long *len)
{
  STREAM_load(path, addr, len);
}

PUBLIC void GB_ReleaseFile(char **addr, long len)
{
  FREE(addr, "GB_ReleaseFile");
}


PUBLIC int GB_IsMissing(int param)
{
  if ((param < 1) || (param > 32))
    return TRUE;
  else
    return (GAMBAS_MissingParam & (1 << (param - 1))) != 0;
}


PUBLIC void GB_Store(ulong type, void *src, void *dst)
{
  VALUE value;

  if (src != NULL)
  {
    VALUE_read(&value, src, type);
    VALUE_write(&value, dst, type);
  }
  else
  {
    VALUE_free(dst, type);
  }
}


PUBLIC void GB_StoreString(GB_STRING src, char **dst)
{
  GB_FreeString(*dst);
  *dst = GB_CopyString(src.addr, src.len);
}

PUBLIC void GB_StoreObject(void *object, void **dst)
{
  GB_Unref(dst);
  *dst = object;
  GB_Ref(object);
}


PUBLIC void GB_Watch(int fd, int flag, void *callback, long param)
{
  HOOK_DEFAULT(watch, WATCH_watch)(fd, flag, callback, param);
}


PUBLIC int GB_New(void **object, char *class_name, char *name, void *parent)
{
  CLASS *class;
  /*
  CLASS_DESC_METHOD *desc;
  VALUE *save_SP;
  int i;
  */

  class = CLASS_look(class_name);
  if (class == NULL || class->new == NULL)
    return TRUE;

  (*class->new)(object, class, name, parent);

  EXEC_special_inheritance(SPEC_NEW, class, *object, 0, TRUE);
  /*
  desc = &class->special[SPEC_NEW]->method;

  save_SP = SP;

  SP =  STACK_check(SP, nparam);

  for (i = 0; i < nparam; i++)
  {
    VALUE_read(SP, param, desc->signature[i]);
    param += TYPE_sizeof(desc->signature[i]);
    SP++;
  }

  EXEC_special(SPEC_NEW, class, *object, nparam, TRUE);

  SP = save_SP;
  */

  OBJECT_UNREF_KEEP(object, "GB_new");

  return FALSE;
}


PUBLIC int GB_GetFunction(GB_FUNCTION *_func, void *object, char *name, char *sign)
{
  GB_API_FUNCTION *func = (GB_API_FUNCTION *)_func;
  short len_min, nparam;
  TYPE *tsign;
  long index;
  CLASS *class;
  int kind;
  CLASS_DESC *desc;

  class = OBJECT_class(object);

  if (OBJECT_is_class(class))
  {
    class = (CLASS *)object;
    kind = CD_STATIC_METHOD;
  }
  else
  {
    kind = CD_METHOD;
  }

  index = CLASS_find_symbol(class, name);
  if (index == NO_SYMBOL)
    goto _NOT_FOUND;

  desc = class->table[index].desc;
  if (CLASS_DESC_get_type(desc) != kind)
    goto _NOT_FOUND;


  TYPE_signature_length(sign, &len_min, &nparam);

  ALLOC(&tsign, nparam * sizeof(TYPE), "GB_GetFunction");

  tsign = TYPE_transform_signature(&tsign, sign, nparam);

  if (TYPE_compare_signature(desc->method.signature, desc->method.npmax, tsign, nparam))
  {
    GB_Error("Parameters do not match");
    goto _NOT_FOUND;
  }

  func->object = object;
  func->desc = &desc->method;

  FREE(&tsign, "GB_GetFunction");
  return 0;

_NOT_FOUND:

  func->object = NULL;
  func->desc = NULL;
  return 1;
}


PUBLIC void GB_Call(GB_FUNCTION *_func, char *param)
{
  GB_API_FUNCTION *func = (GB_API_FUNCTION *)_func;

  call_method(func->object, func->desc, param);
}


PUBLIC void GB_CheckObject(void *object)
{
  if (object == NULL)
    THROW(E_NULL);

  if ((*((OBJECT *)object)->class->check)(object))
    THROW(E_IOBJECT);
}
