/*

  hash.c

  Implmentation d'une table de hachage

  Ce code est une adaptation du code des tables de
  hachage de la GLIB 1.2.8

*/

#define __HASH_C

#include "common.h"
#include "memory.h"
#include "hash.h"


#define NODE_key(htable, node) (((char *)node) + sizeof(HASH_NODE) + htable->s_value)
#define NODE_value(node) ((void *)((char *)node) + sizeof(HASH_NODE))

PRIVATE void hash_table_resize(HASH_TABLE *hash_table);
PRIVATE HASH_NODE **hash_table_lookup_node(HASH_TABLE *hash_table, char *key, int len);
PRIVATE HASH_NODE *hash_node_new(HASH_TABLE *hash_table, char *key, int len);
PRIVATE void hash_node_destroy(HASH_NODE *hash_node);
PRIVATE void hash_nodes_destroy(HASH_NODE *hash_node);


PRIVATE const long primes[] =
{
  11, 19, 37, 73, 109, 163, 251, 367, 557, 823, 1237, 1861, 2777, 4177, 6247, 9371,
  14057, 21089, 31627, 47431, 71143, 106721, 160073, 240101, 360163, 540217, 810343,
  1215497, 1823231, 2734867, 4102283, 6153409, 9230113, 13845163
};

PRIVATE const long nprimes = sizeof (primes) / sizeof (primes[0]);

PRIVATE long spaced_primes_closest(long num)
{
  int i;

  for (i = 0; i < nprimes; i++)
    if (primes[i] > num)
      return primes[i];

  return primes[nprimes - 1];
}


/*
G_LOCK_DEFINE_PRIVATE (g_hash_global);

PRIVATE GMemChunk *node_mem_chunk = NULL;
PRIVATE HASH_NODE *node_free_list = NULL;
*/

PUBLIC void HASH_TABLE_create(HASH_TABLE **hash, size_t s_value)
{
  HASH_TABLE *hash_table;
  /*int i;*/

  ALLOC(&hash_table, sizeof(HASH_TABLE), "HASH_TABLE_create");

  hash_table->size = HASH_TABLE_MIN_SIZE;
  hash_table->nnodes = 0;
  hash_table->s_value = s_value;
  /*hash_table->frozen = FALSE;*/

  ALLOC(&hash_table->nodes, sizeof(HASH_NODE) * hash_table->size, "HASH_TABLE_create");

  /*
  for (i = 0; i < hash_table->size; i++)
    hash_table->nodes[i] = NULL;
  */

  *hash = hash_table;
}


PUBLIC void HASH_TABLE_delete(HASH_TABLE **hash)
{
  int i;
  HASH_TABLE *hash_table = *hash;
  
  if (hash_table == NULL)
    return;

  for (i = 0; i < hash_table->size; i++)
    hash_nodes_destroy(hash_table->nodes[i]);

  FREE(&hash_table->nodes, "HASH_TABLE_delete");
  FREE(hash, "HASH_TABLE_delete");
}


PUBLIC long HASH_TABLE_size(HASH_TABLE *hash_table)
{
  return hash_table->nnodes;
}


PRIVATE ulong key_hash(char *key, int len)
{
  int i;
  ulong hash = 0;

  if (len >= 0)
  {
    for (i = 0; i < len; i++)
      hash = (hash << 4) + (hash ^ key[i]);
  }
  else
  {
    for(i = 0; ; i++)
    {
      if (!key[i])
        break;
      hash = (hash << 4) + (hash ^ key[i]);
    }
  }

  return hash;
}


PRIVATE boolean key_equal(char *key_comp, char *key, int len)
{
  return (!strncmp(key_comp, key, len) && (!key_comp[len]));
}


PRIVATE INLINE HASH_NODE **hash_table_lookup_node(HASH_TABLE *hash_table, char *key, int len)
{
  HASH_NODE **node;

  node = &hash_table->nodes[key_hash(key, len) % hash_table->size];

  while (*node && !key_equal(NODE_key(hash_table, *node), key, len))
    node = &(*node)->next;

  return node;
}


PUBLIC void *HASH_TABLE_lookup(HASH_TABLE *hash_table, char *key, int len)
{
  HASH_NODE *node;

  node = *hash_table_lookup_node(hash_table, key, len);

  return node ? NODE_value(node) : NULL;
}


PUBLIC void *HASH_TABLE_insert(HASH_TABLE *hash_table, char *key, int len)
{
  HASH_NODE **node;

  node = hash_table_lookup_node(hash_table, key, len);

  if (*node)
    return NODE_value(*node);

  /*hash_table_resize(hash_table);*/

  *node = hash_node_new(hash_table, key, len);

  hash_table->nnodes++;

  /*if (!hash_table->frozen)*/

  return NODE_value(*node);
}


PUBLIC void HASH_TABLE_remove(HASH_TABLE *hash_table, char *key, int len)
{
  HASH_NODE **node, *dest;

  node = hash_table_lookup_node(hash_table, key, len);

  if (*node)
  {
    dest = *node;
    (*node) = dest->next;
    hash_node_destroy(dest);
    hash_table->nnodes--;

    /*if (!hash_table->frozen)*/
    hash_table_resize(hash_table);
  }
}

/*
PUBLIC boolean HASH_TABLE_lookup_extended (HASH_TABLE *hash_table,
         gconstpointer  lookup_key,
         gpointer  *orig_key,
         gpointer  *value)
{
  HASH_NODE *node;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  node = *g_hash_table_lookup_node (hash_table, lookup_key);

  if (node)
    {
      if (orig_key)
 *orig_key = node->key;
      if (value)
 *value = node->value;
      return TRUE;
    }
  else
    return FALSE;
}

*/

/*
void
g_hash_table_freeze (HASH_TABLE *hash_table)
{
  g_return_if_fail (hash_table != NULL);

  hash_table->frozen++;
}

void
g_hash_table_thaw (HASH_TABLE *hash_table)
{
  g_return_if_fail (hash_table != NULL);
  
  if (hash_table->frozen)
    if (!(--hash_table->frozen))
      g_hash_table_resize (hash_table);
}
*/

/*
guint
g_hash_table_foreach_remove (HASH_TABLE *hash_table,
        GHRFunc  func,
        gpointer  user_data)
{
  HASH_NODE *node, *prev;
  guint i;
  guint deleted = 0;
  
  g_return_val_if_fail (hash_table != NULL, 0);
  g_return_val_if_fail (func != NULL, 0);

  for (i = 0; i < hash_table->size; i++)
    {
    restart:
      
      prev = NULL;

      for (node = hash_table->nodes[i]; node; prev = node, node = node->next)
 {
   if ((* func) (node->key, node->value, user_data))
     {
       deleted += 1;

       hash_table->nnodes -= 1;
       
       if (prev)
  {
    prev->next = node->next;
    g_hash_node_destroy (node);
    node = prev;
  }
       else
  {
    hash_table->nodes[i] = node->next;
    g_hash_node_destroy (node);
    goto restart;
  }
     }
 }
    }
  
  if (!hash_table->frozen)
    g_hash_table_resize (hash_table);
  
  return deleted;
}
*/

/*
PUBLIC void *HASH_TABLE_first(HASH_TABLE *hash_table)
{
  hash_table->enum_i = -1;
  hash_table->enum_node = NULL;

  return HASH_TABLE_next(hash_table);
}
*/

PUBLIC void *HASH_TABLE_next(HASH_TABLE *hash_table, long *index, HASH_NODE **node)
{
  HASH_NODE *enum_node = *node;
  long enum_i = *index;

  if (enum_node)
    enum_node = enum_node->next;

  while (enum_node == NULL)
  {
    enum_i++;

    if (enum_i >= hash_table->size)
      break;

    enum_node = hash_table->nodes[enum_i];
  };

  *node = enum_node;
  *index = enum_i;

  if (enum_node)
    return NODE_value(enum_node);
  else
    return NULL;
}

/*
void HASH_TABLE_foreach (HASH_TABLE *hash_table,
        GHFunc   func,
        gpointer   user_data)
{
  HASH_NODE *node;
  gint i;
  
  g_return_if_fail (hash_table != NULL);
  g_return_if_fail (func != NULL);

  for (i = 0; i < hash_table->size; i++)
    for (node = hash_table->nodes[i]; node; node = node->next)
      (* func) (node->key, node->value, user_data);
}
*/

PRIVATE void hash_table_resize(HASH_TABLE *hash_table)
{
  HASH_NODE **new_nodes;
  HASH_NODE *node;
  HASH_NODE *next;
  char *node_key;
  double nodes_per_list;
  long hash_val;
  long new_size;
  long i;

  nodes_per_list = hash_table->nnodes / hash_table->size;
  
  if ((nodes_per_list > 0.3 || hash_table->size <= HASH_TABLE_MIN_SIZE) &&
      (nodes_per_list < 3.0 || hash_table->size >= HASH_TABLE_MAX_SIZE))
    return;
  
  new_size = MinMax(spaced_primes_closest(hash_table->nnodes), HASH_TABLE_MIN_SIZE, HASH_TABLE_MAX_SIZE);

  ALLOC(&new_nodes, new_size * sizeof(HASH_NODE *), "hash_table_resize");

  for (i = 0; i < hash_table->size; i++)
    for (node = hash_table->nodes[i]; node; node = next)
      {
        next = node->next;

        node_key = NODE_key(hash_table, node);
        hash_val = key_hash(node_key, -1) % new_size;

        node->next = new_nodes[hash_val];
        new_nodes[hash_val] = node;
      }

  FREE(&hash_table->nodes, "hash_table_resize");
  hash_table->nodes = new_nodes;
  hash_table->size = new_size;
}


PRIVATE HASH_NODE *hash_node_new(HASH_TABLE *hash_table, char *key, int len)
{
  HASH_NODE *hash_node;
  int size;
  char *node_key;
  
  size = sizeof(HASH_NODE) + hash_table->s_value + len + 1;
  ALLOC(&hash_node, size, "hash_node_new");
  
  hash_node->next = NULL;

  memset(NODE_value(hash_node), 0, hash_table->s_value);

  node_key = NODE_key(hash_table, hash_node);
  strncpy(node_key, key, len);
  node_key[len] = 0;

  return hash_node;
}


PRIVATE void hash_node_destroy(HASH_NODE *hash_node)
{
  FREE(&hash_node, "hash_node_destroy");
}

PRIVATE void hash_nodes_destroy(HASH_NODE *hash_node)
{
  HASH_NODE *node = hash_node;
  HASH_NODE *next;

  for(;;)
  {
    if (node == NULL)
      return;

    next = node->next;
    FREE(&node, "hash_nodes_destroy");
    node = next;
  }
}
