/***************************************************************************

  library.c

  GAMBAS plug-in management routines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __LIBRARY_C

#include "common.h"
#include "common_buffer.h"
#include "error.h"
#include "memory.h"

#include <dlfcn.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>

#include "error.h"

#include "class.h"
#include "exec.h"
#include "event.h"
#include "stack.h"
#include "file.h"
#include "project.h"
#include "gambas_api.h"

#include "Object.h"

#include "library.h"


/*#define DEBUG*/


PRIVATE LIBRARY *Library = NULL;


PUBLIC void LIBRARY_add(char *path)
{
  LIBRARY *lib;
  int len;

  LIBRARY_init();

  lib = ARRAY_add(&Library);
  lib->name = FILE_get_name(path);

  if (FILE_is_relative(path))
    path = FILE_cat(PROJECT_exec_path, path, NULL);

  len = strlen(path);

  STRING_new(&lib->path, NULL, len + 3);
  strcpy(lib->path, path);
  strcpy(&lib->path[len], ".so");

  lib->handle = NULL;
  lib->persistent = FALSE;
}


PRIVATE void *get_symbol(LIBRARY *lib, char *symbol)
{
  void *sym;

  sym = dlsym(lib->handle, symbol);
  if (sym == NULL)
  {
    strcpy(COMMON_buffer, dlerror());
    dlclose(lib->handle);
    lib->handle = NULL;
    THROW(E_LIBRARY, lib->path, COMMON_buffer);
  }

  return sym;
}


PRIVATE void copy_interface(long *src, long *dst)
{
  for(;;)
  {
    if (*src == 0)
      return;

    *dst++ = *src++;
  }
}


PRIVATE void load_library(LIBRARY *lib)
{
  int (*func)();
  GB_INTERFACE *iface;

#ifdef DEBUG
  clock_t t = clock();
  printf("Loading library %s ", lib->path);
  fflush(NULL);
#endif

  lib->handle = dlopen(lib->path, RTLD_LAZY);
  if (lib->handle == NULL)
    THROW(E_LIBRARY, lib->path, dlerror());

  func = get_symbol(lib, LIB_INIT);

  /* Interface de Gambas */

  iface = get_symbol(lib, LIB_GAMBAS);
  copy_interface((long *)GAMBAS_Api, (long *)iface);

  /* Initialisation */

  lib->persistent = (boolean)(*func)();

#ifdef DEBUG
  printf("in %g s\n", ((double)(clock() - t) / CLOCKS_PER_SEC));
#endif
}


PRIVATE void unload_library(LIBRARY *lib)
{
  void (*gambas_exit)();

  if (lib->handle == NULL)
    return;

  gambas_exit = dlsym(lib->handle, LIB_EXIT);
  if (gambas_exit != NULL)
    (*gambas_exit)();

  if (!lib->persistent)
    dlclose(lib->handle);

  STRING_free(&lib->path);
  lib->handle = NULL;

#ifdef DEBUG
  printf("Unloading library %s\n", lib->name);
#endif
}


PUBLIC void LIBRARY_load_all(void)
{
  int i;

  for (i = 0; i < ARRAY_count(Library); i++)
    load_library(&Library[i]);
}


PUBLIC void LIBRARY_init(void)
{
  if (Library == NULL)
    ARRAY_create(&Library);
}


PUBLIC void LIBRARY_exit()
{
  int i;

  for (i = 0; i < ARRAY_count(Library); i++)
    unload_library(&Library[i]);

  ARRAY_delete(&Library);
}


PUBLIC LIBRARY *LIBRARY_find(char *name)
{
  int i;

  for (i = 0; i < ARRAY_count(Library); i++)
  {
    if (strcmp(Library[i].name, name) == 0)
      return &Library[i];
  }

  return NULL;
}


PUBLIC boolean LIBRARY_get_interface(char *name, long version, void *iface)
{
  LIBRARY *lib;
  char symbol[32];
  int i, len;
  char c;

  lib = LIBRARY_find(name);
  if (lib == NULL)
    return TRUE;

  len = strlen(name);
  for (i = 0; i < len; i++)
  {
    c = toupper(name[i]);
    if (!isalnum(c))
      c = '_';

    symbol[i] = c;
  }

  sprintf(&symbol[len], "_%ld", version);

  copy_interface((long *)get_symbol(lib, symbol), (long *)iface);

  return FALSE;
}

