/***************************************************************************

  memory.c

  Memory management routines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __MEMORY_C

#include "common.h"
#include "error.h"
#include "debug.h"
#include "memory.h"

#include <malloc.h>

#define DEBUG

#define MAGIC 0x44554B45L


#ifdef DEBUG

PUBLIC long MEMORY_Eco = 0;

PRIVATE long Count = 0;
PRIVATE boolean Init = FALSE;
PRIVATE long SizeAlloc = 0;
PRIVATE long SizeAllocMax = 1023;
PRIVATE long NbrAlloc = 0;

typedef
  struct _MEMORY {
    long magic;
    long size;
    struct _MEMORY *next;
    struct _MEMORY *prev;
    char data[0];
    }
  MEMORY;

#if DEBUG_ALLOC

PRIVATE MEMORY *Alloc = NULL;

#endif

PRIVATE void add_alloc(long size)
{
  long rsize;
  long tsize;
  boolean neg;

  neg = (size < 0);
  size = abs(size);

  tsize = (1 << 24);

  do
  {
    rsize = tsize;
    tsize >>= 1;
  }
  while (tsize > size);

  if (neg)
  {
    SizeAlloc -= rsize;
/*     printf("- %d\n", rsize); */
  }
  else
  {
    SizeAlloc += rsize;
/*     printf("+ %d\n", rsize); */
  }

  if (SizeAlloc > SizeAllocMax)
  {
    SizeAllocMax = (SizeAlloc + 1023) & ~1023;
    /*printf("ALLOC MAX = %ld Ko\n", SizeAllocMax >> 10);*/
  }
}

#endif

#if DEBUG_ALLOC
PUBLIC void MEMORY_alloc(void *p_ptr, size_t size, char *src)
#else
PUBLIC void MEMORY_alloc(void *p_ptr, size_t size)
#endif
{
  char **ptr = (char **)p_ptr;
  MEMORY *alloc;

  size += sizeof(MEMORY);
  alloc = (MEMORY *)malloc(size);
  if (alloc)
  {
    memset(alloc, 0, size);
    *((long *)alloc) = MAGIC;
    *ptr  = alloc->data;
    Count++;
  }
  else
  {
    *ptr = NULL;
    THROW(E_MEMORY);
  }

#ifdef DEBUG
  if (!Init)
  {
    atexit(MEMORY_check);
    Init = TRUE;
  }
#endif

  alloc->size = size;

#ifdef DEBUG
  add_alloc(size);
  NbrAlloc++;
#endif

#if DEBUG_ALLOC

  alloc->prev = NULL;
  alloc->next = Alloc;
  if (Alloc)
    Alloc->prev = alloc;

  Alloc = alloc;

  TRACE_where();
  printf("%s: MEMORY_alloc(%d) -> %p\n", src, size, *ptr);
  fflush(stdout);

  /*MEMORY_verify();*/

#endif
}


#if DEBUG_ALLOC
PUBLIC void MEMORY_alloc_zero(void *p_ptr, size_t size, char *src)
#else
PUBLIC void MEMORY_alloc_zero(void *p_ptr, size_t size)
#endif
{
  void *alloc;

  ALLOC(p_ptr, size, src);

  alloc = *((void **)p_ptr);

  if (alloc)
    memset(alloc, 0, size);
}


PUBLIC void MEMORY_alloc_chunk(void *p_ptr, size_t size)
{
  void **ptr = (void **)p_ptr;

  *ptr = memalign(size, size);

  if (!*ptr)
    THROW(E_MEMORY);

  add_alloc(size);

#if DEBUG_ALLOC
  printf("MEMORY_alloc_chunk() -> %p\n", *ptr);
  fflush(stdout);
#endif
}


#if DEBUG_ALLOC
PUBLIC void MEMORY_realloc(void *p_ptr, size_t size, char *src)
#else
PUBLIC void MEMORY_realloc(void *p_ptr, size_t size)
#endif
{
  char **ptr = (char **)p_ptr;
  MEMORY *alloc, *old_alloc;

#if DEBUG_ALLOC

  TRACE_where();
  printf("%s: MEMORY_realloc(%p, %d) -> %p\n", src, *ptr, size, &alloc[4]);
  fflush(stdout);

#endif

  if (!*ptr) goto BAD;

  old_alloc = (MEMORY *)(*ptr - sizeof(MEMORY));
  if (*((long *)old_alloc) != MAGIC) goto BAD;

#ifdef DEBUG
  add_alloc(size - old_alloc->size);
  NbrAlloc++;
#endif

  alloc = (MEMORY *)realloc(old_alloc, size + sizeof(MEMORY));
  alloc->size = size;

#if DEBUG_ALLOC

  if (alloc->prev)
    alloc->prev->next = alloc;
  if (alloc->next)
    alloc->next->prev = alloc;

  if (Alloc == old_alloc)
    Alloc = alloc;

#endif

  if (!*ptr)
    THROW(E_MEMORY);
  else
    *ptr = alloc->data;

  return;

BAD:

  ERROR_panic("MEMORY_realloc: Bad pointer 0x%08X", (long)(*ptr));

}


#if DEBUG_ALLOC
PUBLIC void MEMORY_free(void *p_ptr, char *src)
#else
PUBLIC void MEMORY_free(void *p_ptr)
#endif
{
  char **ptr = (char **)p_ptr;
  MEMORY *alloc;

#if DEBUG_ALLOC

  /*MEMORY_verify();*/

  TRACE_where();
  printf("%s: MEMORY_free(%p)\n", src, *ptr);
  fflush(stdout);

#endif

  if (!*ptr) goto BAD;

  alloc = (MEMORY *)(*ptr - sizeof(MEMORY));
  if (alloc->magic != MAGIC) goto BAD;
  alloc->magic = 0;

#ifdef DEBUG
  memset(alloc->data, '#', alloc->size - sizeof(MEMORY));
#endif

#if DEBUG_ALLOC

  if (alloc->prev)
    alloc->prev->next = alloc->next;
  if (alloc->next)
    alloc->next->prev = alloc->prev;
  if (alloc == Alloc)
    Alloc = alloc->next;

#endif

#ifdef DEBUG
  add_alloc(-alloc->size);
  NbrAlloc++;
#endif

  free(alloc);
  *ptr = NULL;

  Count--;
  return;

BAD:

  ERROR_panic("MEMORY_free: Bad pointer 0x%08X", (long)(*ptr));
}


#if DEBUG_ALLOC

PUBLIC void MEMORY_verify(void)
{
  MEMORY *alloc;
  boolean bad = FALSE;

  for (alloc = Alloc; alloc != NULL; alloc = alloc->next)
  {
    if (alloc->magic != MAGIC)
    {
      bad = TRUE;
      break;
    }
  }

  if (!bad)
    return;

  printf("\n---\n");
  for (alloc = Alloc; alloc != NULL; alloc = alloc->next)
  {
    printf("%p 0x%08lX %ld\n", alloc->data, (unsigned long)alloc->magic, alloc->size);
    if (alloc->magic != MAGIC)
      bad = TRUE;
  }
  printf("---\n\n");

  if (bad)
    ERROR_panic("MEMORY_verify: corrupted memory");
}

#endif

PUBLIC void MEMORY_check_ptr(void *ptr)
{
  MEMORY *alloc = (MEMORY *)((char *)ptr - sizeof(MEMORY));

  if (alloc->magic != MAGIC)
  {
    printf("%p CORRUPTED !\n", ptr);
    fflush(NULL);
  }
}


PUBLIC void MEMORY_check(void)
{
  if (Count)
  {
#if DEBUG_ALLOC

    printf ("\n--------\n");

    while (Alloc != NULL)
    {
      printf("%p\n", Alloc->data);
      Alloc = Alloc->next;
    }

#endif

    ERROR_panic("MEMORY_check: Forgot to free %d memory blocks.", Count);
  }
/*
#ifdef DEBUG
  printf("%ld calls to malloc() / realloc() / free()\n", NbrAlloc);
  if (MEMORY_Eco)
    printf("%d calls could be saved\n", MEMORY_Eco);
#endif
*/
}
