/***************************************************************************

  project.c

  Project loader

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __PROJECT_C

#include "common.h"
#include "memory.h"
#include "error.h"

#include <unistd.h>

#include "limit.h"
#include "buffer.h"
#include "file.h"
#include "stream.h"
#include "library.h"
#include "archive.h"
#include "exec.h"

#include "project.h"

PUBLIC char *PROJECT_path;
PUBLIC char *PROJECT_exec_path;

PUBLIC char *PROJECT_name;
PUBLIC char *PROJECT_startup;

PUBLIC int PROJECT_argc;
PUBLIC char **PROJECT_argv;

PRIVATE char *project_buffer;

PRIVATE char *project_ptr;
PRIVATE int project_line;


PRIVATE void project_name(char *name, int len)
{
  if (len == 0)
    THROW(E_PROJECT, project_line, "Bad project name");

  name[len] = 0;
  PROJECT_name = name;
}


PRIVATE void project_library(char *name, int len)
{
  const char *delim = ",";
  char *lib;

  name[len] = 0;

  lib = strtok(name, delim);
  while (lib != NULL)
  {
    LIBRARY_add(lib);
    lib = strtok(NULL, delim);
  }
}


PRIVATE void project_exec(char *name, int len)
{
  if (len == 0)
    THROW(E_PROJECT, project_line, "Bad project startup class");

  name[len] = 0;
  PROJECT_startup = name;
}


PRIVATE void project_command(char *line, long len)
{
  static PROJECT_COMMAND command[] = {
    { "PROJECT", project_name },
    { "LIBRARY", project_library },
    { "STARTUP", project_exec },
    { NULL }
    };

  PROJECT_COMMAND *pc;
  char cmd[32];
  int len_cmd;

  for (len_cmd = 0; len_cmd < len; len_cmd++)
  {
    if (line[len_cmd] == '=')
      break;
  }

  if (len_cmd >= len || len_cmd >= sizeof(cmd) || len_cmd == 0)
    THROW(E_PROJECT, project_line, "Syntax error");

  strncpy(cmd, line, len_cmd);

  for (pc = command; ; pc++)
  {
    if (pc->command == NULL)
      THROW(E_PROJECT, project_line, "Unknown command");

    if (strncasecmp(pc->command, cmd, len_cmd) == 0)
    {
      (*pc->func)(&line[len_cmd + 1], len - len_cmd - 1);
      break;
    }
  }
}


PRIVATE void project_analyze(char *addr, long len)
{
  char *end = &addr[len];
  char c;
  char *start;


  project_ptr = addr;
  project_line = 1;
  start = project_ptr;

  for(;;)
  {
    if (project_ptr >= end)
      break;

    c = *project_ptr++;

    if (c == '\n')
    {
      project_line++;
      start = project_ptr;
      continue;
    }

    if (c == '#')
    {
      while ((project_ptr < end) && c != '\n')
        c = *project_ptr++;
      project_ptr--;
      continue;
    }

    if (c <= ' ')
      continue;

    project_ptr--;
    start = project_ptr;

    while ((project_ptr < end) && (c != '\n'))
      c = *project_ptr++;

    project_command(start, project_ptr - start - 1);
    project_ptr--;
  }

  if (PROJECT_name[0] == 0)
    THROW(E_PROJECT, project_line, "No project name");

  if (PROJECT_startup[0] == 0)
    THROW(E_PROJECT, project_line, "No startup class");
}


void PROJECT_init(char *exec, char *file)
{
  long len;
  char *path;

  /* chemin d'installation de Gambas */

  path = FILE_readlink("/usr/bin/gbx");

  if (path == NULL)
  {
    if (FILE_is_absolute(exec))
      path = exec;
    else
      path = FILE_getcwd(exec);
  }

  if (path == NULL)
    goto _PANIC_GBX;

  STRING_new(&PROJECT_exec_path, FILE_get_dir(path), -1);


  /* fichier project */

  if (EXEC_arch)
  {
    ARCH_init(file);

    file = FILE_set_ext(FILE_get_name(file), ".project");
    STRING_new(&PROJECT_path, file, -1);
    file = PROJECT_path;
  }
  else
  {
    if (*file == '/')
    {
      STRING_new(&PROJECT_path, FILE_get_dir(file), -1);
    }
    else
    {
      if (*file == '.' && file[1] == '/')
        file += 2;

      path = FILE_getcwd(file);
      if (path == NULL)
        goto _PANIC;

      path = FILE_get_dir(path);
      chdir(path);

      path = FILE_getcwd(NULL);
      if (path == NULL)
        goto _PANIC;

      STRING_new(&PROJECT_path, path, -1);
    }
  }

  STREAM_load(file, &project_buffer, &len);

  project_analyze(project_buffer, len);

  LIBRARY_load_all();
  return;

_PANIC:

  ERROR_panic("Cannot initialize project: %s", strerror(errno));

_PANIC_GBX:

  ERROR_panic("Cannot find Gambas installation path.\nMaybe you should reinstall it.");
}


void PROJECT_exit(void)
{
  if (project_buffer)
    FREE(&project_buffer, "PROJECT_exit");

  if (EXEC_arch)
    ARCH_exit();

  STRING_free(&PROJECT_path);
  STRING_free(&PROJECT_exec_path);
}
