/***************************************************************************

  subr_conv.c

  Type and conversion subroutines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "common.h"

#include "value.h"
#include "subr.h"
#include "local.h"

#include "String.h"
#include "Date.h"
#include "Number.h"


/*!------------------------------------------------------------

\HELP
null
function

\IN
conv

\TITLE
Null?

\DESC:fr

Dtermine si une expression vaut <I>Null</I>

\SYNTAX:fr

<I>retour</I> = <B>Null?(</B><I>expression</I><B>)</B>

\ARG:fr

<TR><TD><I>expression</I></TD><TD>n'importe quelle expression</TD><TR>
<TR><TD><I>retour</I></TD><TD><I>True</I> si l'expression vaut <I>Null</I>, <I>False</I> dans le cas contraire</TD><TR>

\EXAMPLE

hCol AS Collection

PRINT Null?(hCol)
True

hCol = NEW Collection
PRINT Null?(hCol)
False

\SAME
IsNull

\SEE
_type_

-------------------------------------------------------------*/

PUBLIC void SUBR_is_type(void)
{
  static void *jump[] = {
    &&_BAD, &&_BOOLEAN, &&_BYTE, &&_SHORT, &&_INTEGER, &&_LONG, &&_FLOAT, &&_DATE,
    &&_STRING, &&_BAD, &&_BAD, &&_BAD, &&_BAD, &&_BAD, &&_BAD, &&_NULL,
    &&_OBJECT, &&_NUMBER
    };

  boolean test;
  TYPE type;

  SUBR_ENTER_PARAM(1);

  VARIANT_undo(PARAM);

  type = EXEC_code & 0x3F;
  goto *jump[type];

_BOOLEAN:
_BYTE:
_SHORT:
_INTEGER:
_LONG:
_FLOAT:
_DATE:

  test = PARAM->type == type;
  goto _END;

_STRING:

  test = PARAM->type == T_STRING || PARAM->type == T_CSTRING;
  if (test)
    goto _END;
  else
    goto _NULL;

_NULL:

  test = VALUE_is_null(PARAM);
  goto _END;

_OBJECT:

  test = TYPE_is_object_null(PARAM->type);
  goto _END;

_NUMBER:

  test = TYPE_is_number(PARAM->type);
  goto _END;

_BAD:

  THROW(E_ILLEGAL);

_END:

  RELEASE_STRING(PARAM);

  RETURN->type = T_BOOLEAN;
  RETURN->_integer.value = test ? -1 : 0;

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
cbool
function

\IN
conv

\TITLE
CBool

\DESC:fr

Convertit une expression en <I>Boolean</I>

<P>La conversion d'un nombre en boolen retourne </I>False</I> lorsque ce
nombre vaut 0, et </I>True</I> dans tous les autres cas.</P>
<P>La conversion d'une chane de caractres en boolen retourne <I>False</I>
lorsque la chane est de longueur nulle, et <I>True</I> dans les autres cas.</P>
<P>La conversion d'une rfrence d'objet en boolen retourne <I>False</I>
lorsque cette rfrence est nulle, et <I>True</I> dans le cas contraire.</P>

\SYNTAX:fr

<I>retour</I> = <B>CBOOL(</B><I>expression</I><B>)</B>

\ARG:fr

<TR><TD><I>expression</I></TD><TD>n'importe quelle expression</TD><TR>
<TR><TD><I>retour</I></TD><TD>L'expression convertie</TD><TR>

\EXAMPLE

PRINT CBOOL(0)
False

PRINT CBOOL(1)
True


\SEE
_type_ _cbool_ _cbyte_ _cdate_ _cfloat_ _cint_ _clong_ _cshort_ _cstr_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
cbyte
function

\IN
conv

\TITLE
CByte

\DESC:fr

Convertit une expression en <I>Byte</I> (octet non sign).

<P>Les valeurs numriques en dehors de la plage 0...255 sont
tronques.</P>

<P>Les chanes de caractres sont interprtes selon le format amricain.
Pour convertir une chane de caractres en nombre en utilisant le format
de votre langue, il faut utiliser la fonction <B>VAL</B>.</P>

\SYNTAX:fr

<I>retour</I> = <B>CByte(</B><I>expression</I><B>)</B>

\ARG:fr

<TR><TD><I>expression</I></TD><TD>n'importe quelle expression</TD><TR>
<TR><TD><I>retour</I></TD><TD>L'expression convertie</TD><TR>

\EXAMPLE

PRINT CByte(-1); CByte(1972); CByte(16.2)
 255 180 16.2

\SEE
_type_ _cbool_ _cbyte_ _cdate_ _cfloat_ _cint_ _clong_ _cshort_ _cstr_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
cint
function

\IN
conv

\TITLE
CInt

\DESC:fr

Convertit une expression en <I>Integer</I> (entier sign 32 bits).

<P>Les valeurs numriques en dehors de la plage des entiers 32 bits sont
tronques.</P>

<P>Les chanes de caractres sont interprtes selon le format amricain.
Pour convertir une chane de caractres en nombre en utilisant le format
de votre langue, il faut utiliser la fonction <B>VAL</B>.</P>

\SYNTAX:fr

<I>retour</I> = <B>CInt(</B><I>expression</I><B>)</B>

\ARG:fr

<TR><TD><I>expression</I></TD><TD>n'importe quelle expression</TD><TR>
<TR><TD><I>retour</I></TD><TD>L'expression convertie</TD><TR>

\EXAMPLE

PRINT CInt(Pi); CInt("-334")
 3 -334

\SEE
_type_ _cbool_ _cbyte_ _cdate_ _cfloat_ _cint_ _clong_ _cshort_ _cstr_

-------------------------------------------------------------*/

PUBLIC void SUBR_conv(void)
{
  SUBR_ENTER_PARAM(1);

  VALUE_conv(PARAM, EXEC_code & 0x3F);

  SUBR_LEAVE();
}


PUBLIC void SUBR_type(void)
{
  SUBR_ENTER_PARAM(1);

  /* pas bon ! Ne respecte pas la transitivit de type */

  /*if (TYPE_is_object(PARAM->type))
  {
    type = TYPE_get_name((TYPE)OBJECT_class(PARAM->_object.object));
    STRING_new_constant_value(RETURN, type, -1);
  }
  else
  */

  if (VALUE_is_null(PARAM))
    RETURN->_integer.value = T_NULL;
  else if (PARAM->type == T_CSTRING)
    RETURN->_integer.value = T_STRING;
  else if (PARAM->type == T_VARIANT)
    RETURN->_integer.value = PARAM->_variant.vtype;
  else
    RETURN->_integer.value = PARAM->type;

  RETURN->type = T_INTEGER;

  SUBR_LEAVE();
}


PUBLIC void SUBR_str(void)
{
  char *addr;
  long len;

  SUBR_ENTER_PARAM(1);

  VALUE_to_string(PARAM, &addr, &len);
  STRING_new_value(RETURN, addr, len);

  SUBR_LEAVE();
}


PUBLIC void SUBR_val(void)
{
  char *addr;
  long len;
  VALUE result;

  SUBR_ENTER_PARAM(1);

  SUBR_check_string(PARAM);

  VALUE_get_string(PARAM, &addr, &len);

  VALUE_from_string(&result, addr, len);

  RELEASE_STRING(PARAM);
  *RETURN = result;

  SUBR_LEAVE();
}


PUBLIC void SUBR_format(void)
{
  int fmt_type;
  char *format = NULL;
  long len = 0;
  DATE_SERIAL *date;
  char *str;
  long len_str;

  SUBR_ENTER();

  if (NPARAM == 1)
    fmt_type = LF_STANDARD;
  else
  {
    if (TYPE_is_string(PARAM[1].type))
    {
      fmt_type = LF_USER;
      VALUE_get_string(&PARAM[1], &format, &len);
    }
    else if (TYPE_is_integer(PARAM[1].type))
    {
      fmt_type = PARAM[1]._integer.value;
    }
    else
      THROW(E_TYPE, TYPE_get_name(T_INTEGER), TYPE_get_name(PARAM[1].type));
  }

  if (PARAM->type == T_DATE)
  {
    date = DATE_split(PARAM);
    if (LOCAL_format_date(date, fmt_type, format, len, &str, &len_str))
      THROW(E_FORMAT);
  }
  else
  {
    VALUE_conv(PARAM, T_FLOAT);
    if (LOCAL_format_number(PARAM->_float.value, fmt_type, format, len, &str, &len_str, TRUE))
      THROW(E_FORMAT);
  }

  if (NPARAM >= 2)
    RELEASE_STRING(&PARAM[1]);

  STRING_new_value(RETURN, str, len_str);

  SUBR_LEAVE();
}



/*!------------------------------------------------------------

\HELP
hex
function

\IN
conv

\TITLE
Hex$

\DESC:fr

Convertit un nombre en hexadcimal.

\SYNTAX:fr

<I>retour</I> = <B>Hex$(</B><I>nombre</I> [ <B>,</B> <I>prcision</I> ] <B>)</B>

\ARG:fr

<TR><TD><I>expression</I></TD><TD>un nombre entier</TD><TR>
<TR><TD><I>prcision</I></TD><TD>nombre de chiffres  retourner (par dfaut, les chiffres superflus ne sont pas affichs)</TD><TR>
<TR><TD><I>retour</I></TD><TD>la reprsentation en hexadcimal du nombre</TD><TR>

\EXAMPLE

Print Hex$(-1)
FFFFFFFF

Print Hex$(1972)
7B4

Print Hex$(1972, 6)
0007B4

\SAME
Hex

\SEE
_bin_ _hex_ _str_

-------------------------------------------------------------*/

PUBLIC void SUBR_hex(void)
{
  int prec = 0;

  SUBR_ENTER();

  VALUE_conv(PARAM, T_INTEGER);

  if (NPARAM == 2)
  {
    VALUE_conv(&PARAM[1], T_INTEGER);

    prec = PARAM[1]._integer.value;

    if (prec < 1 || prec > 16)
      THROW(E_ARG);
  }

  NUMBER_int_to_string((unsigned long)(PARAM->_integer.value), prec, 16, RETURN);

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
bin
function

\IN
conv

\TITLE
Bin$

\DESC:fr

Convertit un nombre en binaire.

\SYNTAX:fr

<I>retour</I> = <B>Hex$(</B><I>nombre</I> [ <B>,</B> <I>prcision</I> ] <B>)</B>

\ARG:fr

<TR><TD><I>expression</I></TD><TD>un nombre entier</TD><TR>
<TR><TD><I>prcision</I></TD><TD>nombre de chiffres  retourner (par dfaut, les chiffres superflus ne sont pas affichs)</TD><TR>
<TR><TD><I>retour</I></TD><TD>la reprsentation en binaire du nombre</TD><TR>

\EXAMPLE

Print Bin$(-1)
11111111111111111111111111111111

Print Bin$(1972)
11110110100

Print Bin$(1972, 14)
00011110110100

\SAME
Bin

\SEE
_bin_ _hex_ _str_

-------------------------------------------------------------*/

PUBLIC void SUBR_bin(void)
{
  int prec = 0;

  SUBR_ENTER();

  VALUE_conv(PARAM, T_INTEGER);

  if (NPARAM == 2)
  {
    VALUE_conv(&PARAM[1], T_INTEGER);

    prec = PARAM[1]._integer.value;

    if (prec < 1 || prec > 64)
      THROW(E_ARG);
  }

  NUMBER_int_to_string((unsigned long)(PARAM->_integer.value), prec, 2, RETURN);

  SUBR_LEAVE();
}


