/***************************************************************************

  subr_file.c

  The file and input/output subroutines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/


#include "common.h"
#include "subr.h"
#include "file.h"
#include "stream.h"
#include "gambas_api.h"

#include "String.h"
#include "CFile.h"


PRIVATE STREAM *get_stream(VALUE *value, boolean can_default)
{
  STREAM *stream;

  if (TYPE_is_integer(value->type) && can_default)
  {
    switch(value->_integer.value)
    {
      case 0:
        return CFILE_STREAM(&CFILE_in);
      case 1:
        return CFILE_STREAM(&CFILE_out);
      case 2:
        return CFILE_STREAM(&CFILE_err);
      default:
        THROW(E_ILLEGAL);
    }
  }

  if (VALUE_is_null(value))
    THROW(E_NULL);

  VALUE_conv(value, (TYPE)CLASS_File);
  stream = CFILE_STREAM(value->_object.object);

  if (stream->type == NULL)
    THROW(E_CLOSED);

  return stream;
}

#define get_path SUBR_get_string

/*!------------------------------------------------------------

\HELP
open
function

\IN
file

\TITLE
Open

\DESC:fr

Ouvre un fichier afin d'y lire ou bien y crire des donnes.

\SYNTAX

<B>Open</B> <I>chemin</I> <B>For</B> <I>mode</I> <B>As #</B> <I>fichier</I>

\ARG

<TR><TD><I>chemin</I></TD><TD>Emplacement du fichier.</TD><TR>
<TR><TD><I>mode</I></TD><TD>Mode d'ouverture du fichier.</TD><TR>
<TR><TD><I>fichier</I></TD><TD>objet de type <I>File</I> correspondant au fichier ouvert.</TD><TR>

\REMARK

Le paramtre <I>mode</I> peut prendre l'une des quatre valeurs suivantes :
<UL>
  <LI><B>Read</B> pour ouvrir le fichier en lecture seule.
  <LI><B>Write</B> pour ouvrir le fichier en lecture/criture.
  <LI><B>Create</B> pour crer le fichier puis l'ouvrir en lecture/criture.
      Si le fichier existait dj, il est cras.
  <LI><B>Append</B> pour ouvrir le fichier en lecture/criture en se positionnant
      pralablement  sa fin.
</UL>

\EXAMPLE

hFic As File

Open "/home/gambas/.Xdefaults" For Read As #hFic
...
Close #hFic

\SEE
_close_ _flush_ _input_ _open_ _print_ _read_ _write_

-------------------------------------------------------------*/

PUBLIC void SUBR_open(void)
{
  CFILE *file;

  SUBR_ENTER_PARAM(2);

  SUBR_check_integer(&PARAM[1]);

  file = CFILE_new();

  STREAM_open(CFILE_STREAM(file), get_path(PARAM), PARAM[1]._integer.value);

  OBJECT_put(RETURN, file);

  SUBR_LEAVE();
}


PUBLIC void SUBR_close(void)
{
  SUBR_ENTER_PARAM(1);

  VALUE_conv(&PARAM[0], (TYPE)CLASS_File);

  STREAM_close(get_stream(PARAM, FALSE));

  SUBR_LEAVE();
}


PUBLIC void SUBR_flush(void)
{
  SUBR_ENTER_PARAM(1);

  STREAM_flush(get_stream(PARAM, TRUE));

  SUBR_LEAVE();
}


PUBLIC void SUBR_print(void)
{
  int i;
  STREAM *stream;
  char *addr;
  long len;

  SUBR_ENTER();

  if (NPARAM < 1)
    THROW(E_NEPARAM);

  stream = get_stream(PARAM, TRUE);

  for (i = 1; i < NPARAM; i++)
  {
    PARAM++;
    VALUE_to_string(PARAM, &addr, &len);
    STREAM_write(stream, addr, len);
    RELEASE_STRING(PARAM);
  }

  SUBR_LEAVE();
}


PUBLIC void SUBR_linput(void)
{
  STREAM *stream;
  char *addr;

  SUBR_ENTER_PARAM(1);

  stream = get_stream(PARAM, TRUE);

  STREAM_line_input(stream, &addr);

  RETURN->type = T_STRING;
  RETURN->_string.addr = addr;
  RETURN->_string.start = 0;
  RETURN->_string.len = STRING_length(addr);

  SUBR_LEAVE();
}


PUBLIC void SUBR_input(void)
{
  STREAM *stream;
  char *addr;

  SUBR_ENTER_PARAM(1);

  stream = get_stream(PARAM, TRUE);

  STREAM_input(stream, &addr);

  VALUE_from_string(RETURN, addr, STRING_length(addr));

  if (RETURN->type == T_NULL)
  {
    RETURN->type = T_STRING;
    RETURN->_string.addr = addr;
    RETURN->_string.start = 0;
    RETURN->_string.len = STRING_length(addr);
  }
  else
    STRING_free(&addr);

  SUBR_LEAVE();
}


PUBLIC void SUBR_eof(void)
{
  STREAM *stream;

  SUBR_ENTER_PARAM(1);
  stream = get_stream(PARAM, FALSE);

  RETURN->type = T_BOOLEAN;
  RETURN->_boolean.value = STREAM_eof(stream) ? -1 : 0;

  SUBR_LEAVE();
}


PUBLIC void SUBR_lof(void)
{
  STREAM *stream;

  SUBR_ENTER_PARAM(1);
  stream = get_stream(PARAM, FALSE);

  RETURN->type = T_INTEGER;
  STREAM_lof(stream, &(RETURN->_integer.value));

  SUBR_LEAVE();
}


PUBLIC void SUBR_seek(void)
{
  STREAM *stream;
  long pos;
  long whence = SEEK_SET;

  SUBR_ENTER();

  stream = get_stream(PARAM, FALSE);

  VALUE_conv(&PARAM[1], T_INTEGER);
  pos = PARAM[1]._integer.value;

  if (NPARAM == 3)
  {
    VALUE_conv(&PARAM[2], T_INTEGER);
    whence = PARAM[2]._integer.value;
    if (whence != SEEK_SET && whence != SEEK_CUR && whence != SEEK_END)
      THROW(E_ARG);
  }

  RETURN->type = T_INTEGER;
  RETURN->_integer.value = STREAM_tell(stream);

  STREAM_seek(stream, pos, (int)whence);

  SUBR_LEAVE();
}


PUBLIC void SUBR_read(void)
{
  STREAM *stream;
  long len;

  SUBR_ENTER();

  stream = get_stream(PARAM, TRUE);

  if (NPARAM == 3)
  {
    VALUE_conv(&PARAM[2], T_INTEGER);
    len = PARAM[2]._integer.value;
  }
  else
    len = 0;

  STREAM_read_type(stream, PARAM[1].type, RETURN, len);

  RELEASE_STRING(&PARAM[1]);

  SUBR_LEAVE();
}


PUBLIC void SUBR_write(void)
{
  STREAM *stream;
  long len;

  SUBR_ENTER();

  stream = get_stream(PARAM, TRUE);

  if (NPARAM == 3)
  {
    VALUE_conv(&PARAM[2], T_INTEGER);
    len = PARAM[2]._integer.value;
  }
  else
    len = 0;

  STREAM_write_type(stream, PARAM[1].type, &PARAM[1], len);

  RELEASE_STRING(&PARAM[1]);

  SUBR_LEAVE();
}


PUBLIC void SUBR_stat(void)
{
  SUBR_ENTER_PARAM(1);

  FILE_stat(get_path(PARAM));

  RETURN->type = T_INTEGER;
  RETURN->_integer.value = FILE_stat_info.type;

  SUBR_LEAVE();
}


PUBLIC void SUBR_exist(void)
{
  boolean exist;

  SUBR_ENTER_PARAM(1);

  exist = FILE_exist(get_path(PARAM));

  RETURN->type = T_BOOLEAN;
  RETURN->_integer.value = exist ? -1 : 0;

  SUBR_LEAVE();
}


PUBLIC void SUBR_dir()
{
  char *path;
  char *pattern;
  long len_pattern;

  SUBR_ENTER();

  if (NPARAM >= 1)
  {
    path = get_path(PARAM);

    if (NPARAM >= 2)
      SUBR_get_string_len(&PARAM[1], &pattern, &len_pattern);
    else
    {
      pattern = NULL;
      len_pattern = 0;
    }

    FILE_dir_first(path, pattern, len_pattern);
  }

  path = FILE_dir_next();
  if (path == NULL)
    STRING_void_value(RETURN);
  else
    STRING_new_constant_value(RETURN, path, -1);

  if (NPARAM >= 2)
    RELEASE_STRING(&PARAM[1]);

  SUBR_LEAVE();
}


PUBLIC void SUBR_kill(void)
{
  SUBR_ENTER_PARAM(1);

  FILE_unlink(get_path(PARAM));

  SUBR_LEAVE();
}


PUBLIC void SUBR_mkdir(void)
{
  SUBR_ENTER_PARAM(1);

  FILE_mkdir(get_path(PARAM));

  SUBR_LEAVE();
}


PUBLIC void SUBR_rmdir(void)
{
  SUBR_ENTER_PARAM(1);

  FILE_rmdir(get_path(PARAM));

  SUBR_LEAVE();
}


PUBLIC void SUBR_rename(void)
{
  SUBR_ENTER_PARAM(2);

  FILE_rename(SUBR_get_string(&PARAM[0]), SUBR_get_string(&PARAM[1]));

  SUBR_LEAVE();
}


