/***************************************************************************

  subr_math.c

  Mathematical routines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "common.h"

#include <math.h>

#include "value.h"
#include "subr.h"
#include "mathext.h"


PRIVATE MATH_FUNC MathFunc[] = {
  NULL, frac, log, exp, sqrt, sin, cos, tan, atan, asin, acos,
  deg, rad, log10, sinh, cosh, tanh, asinh, acosh, atanh
  };


/* Arithmtique */

#define SMT_NAME  SUBR_add
#define SMT_TYPE  2
#define SMT_OP    +=
#define SMT_DATE

#include "subr_math_temp.h"


#define SMT_NAME  SUBR_sub
#define SMT_TYPE  2
#define SMT_OP    -=
#define SMT_DATE

#include "subr_math_temp.h"


#define SMT_NAME  SUBR_mul
#define SMT_TYPE  2
#define SMT_OP    *=

#include "subr_math_temp.h"


#define SMT_NAME  SUBR_div
#define SMT_TYPE  2
#define SMT_OP    /=
#define SMT_TEST_ZERO

#include "subr_math_temp.h"


#define SMT_NAME  SUBR_neg
#define SMT_TYPE  1
#define SMT_OP    -

#include "subr_math_temp.h"


#define SMT_NAME    SUBR_quo
#define SMT_TYPE    3
#define SMT_RESULT  quot

#include "subr_math_temp.h"


#define SMT_NAME    SUBR_rem
#define SMT_TYPE    3
#define SMT_RESULT  rem

#include "subr_math_temp.h"



/* Logique */

#define SMT_NAME  SUBR_and
#define SMT_TYPE  2
#define SMT_OP    &=
#define SMT_INTEGER

#include "subr_math_temp.h"


#define SMT_NAME  SUBR_or
#define SMT_TYPE  2
#define SMT_OP    |=
#define SMT_INTEGER

#include "subr_math_temp.h"


#define SMT_NAME  SUBR_xor
#define SMT_TYPE  2
#define SMT_OP    ^=
#define SMT_INTEGER

#include "subr_math_temp.h"


/*
#define SMT_NAME  SUBR_not
#define SMT_TYPE  1
#define SMT_OP    ~
#define SMT_INTEGER

#include "subr_math_temp.h"
*/

#define SMT_NAME  SUBR_pow
#define SMT_TYPE  2
#define SMT_FUNC  pow
#define SMT_FLOAT

#include "subr_math_temp.h"


/*!------------------------------------------------------------

\HELP
abs
function

\IN
math

\TITLE
Abs

\DESC:fr

Retourne la valeur absolue du nombre pass en argument

\SYNTAX

<I>retour</I> = <B>Abs(</B><I>nombre</I><B>)</B>

\ARG

<TR><TD><I>nombre</I></TD><TD>Un nombre quelconque</TD><TR>
<TR><TD><I>retour</I></TD><TD>La valeur absolue de ce nombre</TD><TR>

\EXAMPLE

Print Abs(-3.1416)
 3,1416

Print Abs(0)
 0

Print Abs(3.1416)
 3,1416

\SEE
_abs_ _sgn_

-------------------------------------------------------------*/

#define SMT_NAME          SUBR_abs
#define SMT_TYPE          1
#define SMT_FUNC_INTEGER  labs
#define SMT_FUNC_FLOAT    fabs

#include "subr_math_temp.h"


#define SMT_NAME          SUBR_int
#define SMT_TYPE          1
#define SMT_FUNC_INTEGER  0+
#define SMT_FUNC_FLOAT    floor

#include "subr_math_temp.h"


#define SMT_NAME          SUBR_fix
#define SMT_TYPE          1
#define SMT_FUNC_INTEGER  0+
#define SMT_FUNC_FLOAT    fix

#include "subr_math_temp.h"


/*!------------------------------------------------------------

\HELP
sgn
function

\IN
math

\TITLE
Sgn

\DESC:fr

Retourne le signe du nombre pass en argument.

<P>
+1 signifie que le nombre est positif.<BR>
0 signifie que le nombre est nul.<BR>
-1 signifie que le nombre est ngatif.<BR>
</P>

\SYNTAX

<I>retour</I> = <B>Sgn(</B><I>nombre</I><B>)</B>

\ARG

<TR><TD><I>nombre</I></TD><TD>Un nombre quelconque</TD><TR>
<TR><TD><I>retour</I></TD><TD>Le signe du nombre</TD><TR>

\EXAMPLE

Print Sgn(-3.1416)
-1

Print Sgn(0)
 0

Print Sgn(3.1416)
+1

\SEE
_abs_ _sgn_

-------------------------------------------------------------*/

#define SMT_NAME          SUBR_sgn
#define SMT_TYPE          1
#define SMT_FUNC_INTEGER  lsgn
#define SMT_FUNC_FLOAT    fsgn

#include "subr_math_temp.h"


/*!------------------------------------------------------------

\HELP
pi
function

\IN
math

\TITLE
Pi

\DESC:fr

Retourne le nombre Pi, ventuellement multipli par une valeur
passe en argument.

\SYNTAX

<I>retour</I> = <B>Pi</B> [ <B>(</B><I>nombre</I><B>)</B> ]

\ARG

<TR><TD><I>nombre</I></TD><TD>Un nombre quelconque (par dfaut 1)</TD><TR>
<TR><TD><I>retour</I></TD><TD>Pi multipli par le nombre prcdent</TD><TR>

\EXAMPLE

Print Pi
 3,14159265359

Print Pi(1.0 / 3)
 1,047197551197

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/

void SUBR_pi(void)
{
  SUBR_ENTER();

  if (NPARAM == 0)
  {
    RETURN->type = T_FLOAT;
    RETURN->_float.value = M_PI;
  }
  else
  {
    VALUE_conv(PARAM, T_FLOAT);
    RETURN->type = T_FLOAT;
    RETURN->_float.value = M_PI * PARAM->_float.value;
  }

  SUBR_LEAVE();
}


void SUBR_randomize(void)
{
  SUBR_ENTER_PARAM(0);

  randomize();

  RETURN->type = T_VOID;

  SUBR_LEAVE();
}


void SUBR_rnd(void)
{
  double min = 0.0, max = 1.0;

  SUBR_ENTER();

  if (NPARAM >= 1)
  {
    VALUE_conv(&PARAM[0], T_FLOAT);
    max = PARAM->_float.value;
  }

  if (NPARAM == 2)
  {
    min = max;
    VALUE_conv(&PARAM[1], T_FLOAT);
    max = PARAM[1]._float.value;
  }

  RETURN->type = T_FLOAT;
  RETURN->_float.value = (rnd() * (max - min)) + min;

  SUBR_LEAVE();
}


void SUBR_round(void)
{
  long val = 0;
  double power;

  SUBR_ENTER();

  if (NPARAM == 2)
    val = SUBR_get_integer(&PARAM[1]);

  power = pow(10, val);

  VALUE_conv(&PARAM[0], T_FLOAT);

  RETURN->type = T_FLOAT;
  /*RETURN->_float.value = rint(PARAM->_float.value / power) * power;*/
  RETURN->_float.value = rint(PARAM->_float.value / power) * power;

  SUBR_LEAVE();
}


PUBLIC void SUBR_math(void)
{
  SUBR_ENTER_PARAM(1);

  if (TYPE_is_variant(PARAM->type))
    VARIANT_undo(PARAM);

  if (!TYPE_is_number(PARAM->type))
    THROW(E_TYPE, "Number", TYPE_get_name(PARAM->type));

  VALUE_conv(PARAM, T_FLOAT);

  PARAM->_float.value = (*MathFunc[EXEC_code & 0x1F])(PARAM->_float.value);

  if (!finite(PARAM->_float.value))
    THROW(E_MATH);

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
sin
function

\IN
math

\TITLE
Sin

\DESC:fr

Retourne le sinus de l'angle pass en argument.

\SYNTAX

<I>retour</I> = <B>Sin(</B><I>argument</I><B>)</B>

\ARG

<TR><TD><I>argument</I></TD><TD>Un angle en radians de type <I>float</I></TD><TR>
<TR><TD><I>retour</I></TD><TD>Le sinus de cet angle</TD><TR>

\EXAMPLE

Print Sin(Pi / 3)
 0,866025403784

\REM

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
cos
function

\IN
math

\TITLE
Cos

\DESC:fr

Retourne le cosinus de l'angle pass en argument.

\SYNTAX

<I>retour</I> = <B>Cos(</B><I>argument</I><B>)</B>

\ARG

<TR><TD><I>argument</I></TD><TD>Un angle en radians de type <I>float</I></TD><TR>
<TR><TD><I>retour</I></TD><TD>Le cosinus de cet angle</TD><TR>

\EXAMPLE

Print Cos(Pi / 3)
 0,5

\REM

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
tan
function

\IN
math

\TITLE
Tan

\DESC:fr

Retourne la tangente de l'angle pass en argument.

\SYNTAX

<I>retour</I> = <B>Tan(</B><I>argument</I><B>)</B>

\ARG

<TR><TD><I>argument</I></TD><TD>Un angle en radians de type <I>float</I></TD><TR>
<TR><TD><I>retour</I></TD><TD>La tangente de cet angle</TD><TR>

\EXAMPLE

Print Tan(Pi / 3)
 1,732050807569

\REM

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
asn
function

\IN
math

\TITLE
Asn

\DESC:fr

Retourne l'angle dont le sinus est pass en argument

\SYNTAX

<I>retour</I> = <B>Asn(</B><I>argument</I><B>)</B>

\ARG

<TR><TD><I>argument</I></TD><TD>Un nombre compris entre -1 et +1</I></TD><TR>
<TR><TD><I>retour</I></TD><TD>L'angle en radians, compris entre -Pi/2 et +Pi/2</TD><TR>

\EXAMPLE

Print Asn(0.866025403784)
 1,047197551197

\SAME
Asin

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
acs
function

\IN
math

\TITLE
Acs

\DESC:fr

Retourne l'angle dont le cosinus est pass en argument

\SYNTAX

<I>retour</I> = <B>Cos(</B><I>argument</I><B>)</B>

\ARG

<TR><TD><I>argument</I></TD><TD>Un nombre compris entre -1 et +1</I></TD><TR>
<TR><TD><I>retour</I></TD><TD>L'angle en radians, compris entre 0 et Pi</TD><TR>

\EXAMPLE

Print Acs(0.5)
 1,047197551197

\SAME
Acos

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
atn
function

\IN
math

\TITLE
Atn

\DESC:fr

Retourne l'angle dont la tangente est pass en argument

\SYNTAX

<I>retour</I> = <B>Tan(</B><I>argument</I><B>)</B>

\ARG

<TR><TD><I>argument</I></TD><TD>Un nombre quelconque</I></TD><TR>
<TR><TD><I>retour</I></TD><TD>L'angle en radians, compris entre -Pi/2 et +Pi/2</TD><TR>

\EXAMPLE

Print Atn(1.732050807569)
 1,047197551197

\SAME
Atan

\SEE
_acs_ _asn_ _atn_ _cos_ _deg_ _pi_ _rad_ _sin_ _tan_

-------------------------------------------------------------*/


PUBLIC void SUBR_not(void)
{
  static void *jump[17] = {
    &&_VARIANT, &&_BOOLEAN, &&_INTEGER, &&_INTEGER, &&_INTEGER, &&_LONG, &&_FLOAT, &&_DATE,
    &&_STRING, &&_STRING, &&_ERROR, &&_ERROR, &&_ERROR, &&_ERROR, &&_ERROR, &&_NULL,
    &&_OBJECT
    };

  VALUE *P1;
  void *jump_end;
  TYPE type = EXEC_code & 0x1F;

  P1 = SP - 1;
  jump_end = &&_END;
  goto *jump[type];

_VARIANT:

  type = P1->type;

  if (TYPE_is_variant(type))
  {
    type = P1->_variant.vtype;
    if (TYPE_is_number_date(type))
    {
      VARIANT_undo(P1);
      jump_end = &&_VARIANT_END;
      goto *jump[type];
    }
  }
  else if (TYPE_is_object(type))
  {
    *PC |= T_OBJECT;
    goto _OBJECT;
  }
  else
  {
    *PC |= type;
    goto *jump[type];
  }

_BOOLEAN:

  P1->_integer.value = P1->_integer.value ? 0 : (-1);
  goto *jump_end;

_INTEGER:

  P1->_integer.value = ~P1->_integer.value;
  goto *jump_end;

_LONG:
_FLOAT:
_DATE:
  goto _ERROR;

_STRING:
_OBJECT:
_NULL:

  P1->_integer.value = VALUE_is_null(P1) ? (-1) : 0;
  P1->type = T_BOOLEAN;
  goto *jump_end;

_ERROR:

  THROW(E_TYPE, "Number, String or Object", TYPE_get_name(type));

_VARIANT_END:

  VALUE_conv(P1, T_VARIANT);

_END:

  SP--;
}
