/***************************************************************************

  subr_string.c

  The String management subroutines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "common.h"
#include "common_buffer.h"

#include <ctype.h>
#include <regex.h>

#include "pcode.h"
#include "value.h"
#include "subr.h"
#include "regexp.h"
#include "String.h"


PUBLIC void SUBR_cat(void)
{
  int i;
  long len, len_cat;
  char *str, *ptr;

  SUBR_ENTER();

  len_cat = 0;

  for (i = 0; i < NPARAM; i++)
  {
    VALUE_conv(&PARAM[i], T_STRING);
    len_cat += PARAM[i]._string.len;
  }

  STRING_new(&str, NULL, len_cat);
  ptr = str;

  for (i = 0; i < NPARAM; i++)
  {
    len = PARAM[i]._string.len;
    if (len > 0)
    {
      memcpy(ptr, PARAM[i]._string.addr + PARAM[i]._string.start, len);
      ptr += len;
      RELEASE_STRING(&PARAM[i]);
    }
  }

  RETURN->type = T_STRING;
  RETURN->_string.addr = str;
  RETURN->_string.start = 0;
  RETURN->_string.len = len_cat;

  SUBR_LEAVE();
}


PUBLIC void SUBR_file(void)
{
  int i;
  long length;
  char *addr;
  long len;
  char *str, *ptr;
  boolean slash;

  SUBR_ENTER();

  length = 0;
  slash = FALSE;

  for (i = 0; i < NPARAM; i++)
  {
    /*VALUE_conv(&PARAM[i], T_STRING);*/
    SUBR_get_string_len(&PARAM[i], &addr, &len);

    if (len > 0)
    {
      if (length > 0)
      {
        if (!slash && (addr[0] != '/'))
          length++;
      }

      slash = addr[len - 1] == '/';

      length += len;
    }

  }

  STRING_new(&str, NULL, length);
  ptr = str;

  for (i = 0; i < NPARAM; i++)
  {
    VALUE_get_string(&PARAM[i], &addr, &len);
    if (len > 0)
    {
      if ((ptr > str) && (ptr[-1] != '/') && (*addr != '/'))
        *ptr++ = '/';
      memcpy(ptr, addr, len);
      ptr += len;
    }
    RELEASE_STRING(&PARAM[i]);
  }

  RETURN->type = T_STRING;
  RETURN->_string.addr = str;
  RETURN->_string.start = 0;
  RETURN->_string.len = length;

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
left
function

\IN
string

\TITLE
Left$

\DESC:fr

Extrait les <I>n</I> premiers caractres d'une chane en
partant de la gauche.

\SYNTAX:fr

<I>retour</I> = <B>Left$(</B><I>chane</I> [ <B>,</B> <I>nombre</I> ] <B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>nombre</I></TD><TD>Le nombre de caractres  extraire (par dfaut, 1)</TD><TR>
<TR><TD><I>retour</I></TD><TD>La sous-chane compose des premiers caractres de <I>chane</I></TD><TR>

\EXAMPLE

Print Left$("Bonjour")
B

Print Left$("Bonjour", 3)
Bon

\SAME
Left

\SEE
_left_ _len_ _mid_ _right_ _trim_

-------------------------------------------------------------*/

PUBLIC void SUBR_left(void)
{
  long val;

  SUBR_ENTER();

  if (SUBR_check_string(PARAM))
    goto _FIN;

  if (NPARAM == 1)
    val = 1;
  else
  {
    VALUE_conv(&PARAM[1], T_INTEGER);
    val = PARAM[1]._integer.value;
  }

  if (val < 0)
    val += PARAM->_string.len;

  PARAM->_string.len = MinMax(val, 0, PARAM->_string.len);

_FIN:

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
right
function

\IN
string

\TITLE
Right$

\DESC:fr

Extrait les <I>n</I> derniers caractres d'une chane, en
partant de la droite.

\SYNTAX:fr

<I>retour</I> = <B>Right$(</B><I>chane</I> [ <B>,</B> <I>nombre</I> ] <B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>nombre</I></TD><TD>Le nombre de caractres  extraire (par dfaut, 1)</TD><TR>
<TR><TD><I>retour</I></TD><TD>La sous-chane compose des derniers caractres de <I>chane</I></TD><TR>

\EXAMPLE

Print Right$("Bonjour")
r

Print Right$("Bonjour", 3)
our

\SAME
Right

\SEE
_left_ _len_ _mid_ _right_ _trim_

-------------------------------------------------------------*/

PUBLIC void SUBR_right(void)
{
  long val;
  long new_len;

  SUBR_ENTER();

  if (SUBR_check_string(PARAM))
    goto _FIN;

  if (NPARAM == 1)
    val = 1;
  else
  {
    VALUE_conv(&PARAM[1], T_INTEGER);
    val = PARAM[1]._integer.value;
  }

  if (val < 0)
    val += PARAM->_string.len;

  new_len = MinMax(val, 0, PARAM->_string.len);

  PARAM->_string.start += PARAM->_string.len - new_len;
  PARAM->_string.len = new_len;

_FIN:

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
mid
function

\IN
string

\TITLE
Mid$

\DESC:fr

Extrait une sous-chane d'une chane de caractres.

\SYNTAX:fr

<I>retour</I> = <B>Mid$(</B><I>chane</I><B>,</B> <I>position</I>
[ <B>,</B><I>nombre</I> ] <B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>position</I></TD>
<TD>Position du premier caractre  extraire (1 correspondant au dbut de la chane)</TD><TR>
<TR><TD><I>nombre</I></TD>
<TD>Le nombre de caractres  extraire (par dfaut, jusqu' la fin de la chane)</TD><TR>
<TR><TD><I>retour</I></TD><TD>La sous-chane extraite</TD><TR>

\EXAMPLE

PRINT Mid$("Bonjour", 3, 4)
njour

PRINT Mid$("Bonjour", 5)
our

\SAME
Mid

\SEE
_left_ _len_ _mid_ _right_ _trim_

-------------------------------------------------------------*/


PUBLIC void SUBR_mid(void)
{
  long start;
  long len;

  SUBR_ENTER();

  if (SUBR_check_string(PARAM))
    goto FIN;

  VALUE_conv(&PARAM[1], T_INTEGER);
  start = PARAM[1]._integer.value - 1;

  if (start < 0)
    THROW(E_ARG);

  if (start >= PARAM->_string.len)
  {
    STRING_void_value(PARAM);
    goto FIN;
  }

  if (NPARAM == 2)
    len = PARAM->_string.len;
  else
  {
    VALUE_conv(&PARAM[2], T_INTEGER);
    len = PARAM[2]._integer.value;
  }

  if (len < 0)
    len = Max(0, PARAM->_string.len - start + len);

  PARAM->_string.start += start;
  PARAM->_string.len = MinMax(len, 0, PARAM->_string.len - start);

FIN:

  SUBR_LEAVE();
}

/*!------------------------------------------------------------

\HELP
len
function

\IN
string

\TITLE
Len

\DESC:fr

Retourne la longueur d'une chane de caractres.

\SYNTAX:fr

<I>retour</I> = <B>Len(</B><I>chane</I><B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>retour</I></TD><TD>Le nombre de caractres composant cette chane</TD><TR>

\EXAMPLE

PRINT Len("Bonjour")
7

\REM

\SEE
_left_ _len_ _mid_ _right_ _trim_

-------------------------------------------------------------*/


PUBLIC void SUBR_len(void)
{
  long len;

  SUBR_ENTER_PARAM(1);

  if (SUBR_check_string(PARAM))
    len = 0;
  else
  {
    len = PARAM->_string.len;
    RELEASE_STRING(PARAM);
  }

  RETURN->type = T_INTEGER;
  RETURN->_integer.value = PARAM->_string.len;

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
space
function

\IN
string

\TITLE
Space$

\DESC:fr

Fabrique une chane compose uniquement d'espaces.

\SYNTAX:fr

<I>retour</I> = <B>Space$(</B><I>nombre</I><B>)</B>

\ARG:fr

<TR><TD><I>nombre</I></TD><TD>nombre d'espaces composant la chane</TD><TR>
<TR><TD><I>retour</I></TD><TD>une chane constitue uniquement d'espaces</TD><TR>

\EXAMPLE

PRINT "'" & Space$(7) & "'"
'       '

\SAME
Space

\SEE
_space_ _string_

-------------------------------------------------------------*/

PUBLIC void SUBR_space(void)
{
  int len;

  SUBR_ENTER_PARAM(1);

  SUBR_check_integer(PARAM);
  len = PARAM->_integer.value;

  if (len < 0)
    THROW(E_ARG);

  if (len == 0)
  {
    STRING_void_value(RETURN);
  }
  else
  {
    STRING_new_value(RETURN, NULL, len);
    memset(RETURN->_string.addr, ' ', len);
  }

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
string
function

\IN
string

\TITLE
String$

\DESC:fr

Fabrique une chane contenant plusieurs fois le mme caractre.

\SYNTAX:fr

<I>retour</I> = <B>String$(</B><I>nombre</I><B>, </B><I>car</I><B>)</B>

\ARG:fr

<TR><TD><I>nombre</I></TD><TD>nombre de caractres de la chane</TD><TR>
<TR><TD><I>car</I></TD><TD>code ASCII du caractre  rpter</TD><TR>
<TR><TD><I>retour</I></TD><TD>une chane contenant <I>nombre</I> fois le caractre <I>car</I></TD><TR>

\EXAMPLE

PRINT String$(12, Asc("*"))
************

\SAME
String

\SEE
_space_ _string_

-------------------------------------------------------------*/

PUBLIC void SUBR_string(void)
{
  int len;
  int car;

  SUBR_ENTER_PARAM(2);

  SUBR_check_integer(PARAM);
  len = PARAM->_integer.value;

  SUBR_check_integer(&PARAM[1]);
  car = PARAM[1]._integer.value;
  if (car < 0 || car > 255)
    THROW(E_ARG);

  if (len < 0)
    THROW(E_ARG);

  if (len == 0)
  {
    STRING_void_value(RETURN);
  }
  else
  {
    STRING_new_value(RETURN, NULL, len);
    memset(RETURN->_string.addr, car, len);
  }

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
trim
function

\IN
string

\TITLE
Trim$

\DESC:fr

Retire les caractres non imprimables et les espaces des deux
extrmits d'une chane de caractres.

\SYNTAX:fr

<I>retour</I> = <B>Trim$(</B><I>chane</I><B><B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>retour</I></TD><TD>La chane de caractres tronque</I></TD><TR>

\EXAMPLE

PRINT "-" & Trim$("  Bonjour tout le monde  ") & "-"
-Bonjour tout le monde-

\SMAE
Trim

\SEE
_left_ _len_ _mid_ _right_ _trim_

-------------------------------------------------------------*/

PUBLIC void SUBR_trim(void)
{
  char *str;

  SUBR_ENTER_PARAM(1);

  if (SUBR_check_string(PARAM))
    goto _FIN;

  if (PARAM->_string.len > 0)
  {
    str = &PARAM->_string.addr[PARAM->_string.start];

    while (PARAM->_string.len > 0 && *str <= ' ')
    {
      PARAM->_string.start++;
      PARAM->_string.len--;
      str++;
    }

    while (PARAM->_string.len > 0 && str[PARAM->_string.len - 1] <= ' ')
    {
      PARAM->_string.len--;
    }
  }

_FIN:

  SUBR_LEAVE();
}



/*!------------------------------------------------------------

\HELP
upper
function

\IN
string

\TITLE
Upper$

\DESC:fr

Convertit une chane de caractres en majuscules.

\SYNTAX:fr

<I>retour</I> = <B>Upper$(</B><I>chane</I><B><B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>retour</I></TD><TD>La chane de caractres convertie en majuscules</I></TD><TR>

\EXAMPLE

PRINT Upper$("Bonjour")
BONJOUR

\SAME

Upper UCase$ UCase

\SEE
_lower_

-------------------------------------------------------------*/

/*!------------------------------------------------------------

\HELP
lower
function

\IN
string

\TITLE
Lower$

\DESC:fr

Convertit une chane de caractres en minuscules.

\SYNTAX:fr

<I>retour</I> = <B>Lower$(</B><I>chane</I><B><B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>Une chane de caractres</TD><TR>
<TR><TD><I>retour</I></TD><TD>La chane de caractres convertie en minuscules</I></TD><TR>

\EXAMPLE

PRINT Lower$("BonJOUR")
bonjour

\SAME
Lower LCase$ LCase

\SEE
_upper_

-------------------------------------------------------------*/

#define STRING_APPLY(func) \
  char *str, *ptr; \
  int len; \
  \
  SUBR_ENTER_PARAM(1); \
   \
  if (!SUBR_check_string(PARAM)) \
  { \
     \
    len = PARAM->_string.len; \
    if (len > 0) \
    { \
      STRING_copy_from_value(&str, PARAM); \
      \
      for (ptr = str; *ptr; ptr++) \
        *ptr = func(*ptr); \
        \
      RELEASE_STRING(PARAM); \
      \
      PARAM->type = T_STRING; \
      PARAM->_string.addr = str; \
      PARAM->_string.start = 0; \
      PARAM->_string.len = len; \
    } \
  } \
  \
  SUBR_LEAVE();


PUBLIC void SUBR_upper(void)
{
  STRING_APPLY(toupper);
}

PUBLIC void SUBR_lower(void)
{
  STRING_APPLY(tolower);
}

/*!------------------------------------------------------------

\HELP
chr
function

\IN
string

\TITLE
Chr$

\DESC:fr

Retourne une chane contenant le cararctre dont le code ASCII
est pass en argument

\SYNTAX:fr

<I>retour</I> = <B>Chr$(</B><I>code</I><B>)</B>

\ARG:fr

<TR><TD><I>code</I></TD><TD>code ASCII du caractre</TD><TR>
<TR><TD><I>retour</I></TD><TD>caractre</TD><TR>

\EXAMPLE

PRINT Chr$(65)
A

\SAME
Chr

\SEE
<A HREF="asc.html">Asc</A>

-------------------------------------------------------------*/

PUBLIC void SUBR_chr(void)
{
  int car;

  SUBR_ENTER_PARAM(1);

  SUBR_check_integer(PARAM);

  car = PARAM->_integer.value;
  if (car < 0 || car > 255)
    THROW(E_ARG);

  STRING_char_value(PARAM, car);

  SUBR_LEAVE();
}


/*!------------------------------------------------------------

\HELP
asc
function

\IN
string

\TITLE
Asc

\DESC:fr

Retourne le code ASCII d'un des caractres contenu dans une chane.

\SYNTAX:fr

<I>retour</I> = <B>Asc(</B><I>chane</I> [ <B>,</B> <I>position</I> ] <B>)</B>

\ARG:fr

<TR><TD><I>chane</I></TD><TD>chane de caractres</TD><TR>
<TR><TD><I>position</I></TD><TD>position du caractre  examiner (par dfaut, le premier de la chane)</TD><TR>
<TR><TD><I>retour</I></TD><TD>code ASCII du caract-re</TD><TR>

\EXAMPLE

PRINT Asc("Bonjour")
66

PRINT Asc("Bonjour", 4)
106

\SEE
<A HREF="chr.html">Chr$</A>

-------------------------------------------------------------*/

PUBLIC void SUBR_asc(void)
{
  int pos = 0;

  SUBR_ENTER();

  if (!SUBR_check_string(PARAM))
  {
    pos = 1;
    if (NPARAM == 2)
    {
      SUBR_check_integer(&PARAM[1]);
      pos = PARAM[1]._integer.value;
    }

    if (pos < 1 || pos > PARAM->_string.len)
      pos = 0;
    else
      pos = (unsigned char)PARAM->_string.addr[PARAM->_string.start + pos - 1];

    RELEASE_STRING(PARAM);
  }

  RETURN->type = T_INTEGER;
  RETURN->_integer.value = pos;

  SUBR_LEAVE();
}


PUBLIC void SUBR_instr(void)
{
  boolean right;
  long is, ip, ret;
  char *cs, *cp;
  long ls, lp;

  SUBR_ENTER();

  /* Knuth Morris Pratt un jour peut-tre ? */

  ret = 0;

  /* /!\ | et pas ||, car on veut que les deux fonctions soient excutes */

  if (SUBR_check_string(PARAM) | SUBR_check_string(&PARAM[1]))
    goto FOUND;

  lp = PARAM[1]._string.len;
  ls = PARAM->_string.len;

  right = ((EXEC_code >> 8) == CODE_RINSTR);

  if (lp > ls) goto FOUND;

  ls = ls - lp + 1; /* Longueur du dbut du texte o effectuer la recherche */

  is = 0;

  if (NPARAM == 3)
  {
    SUBR_check_integer(&PARAM[2]);
    is = PARAM[2]._integer.value;
  }

  if (is < 0)
    is = ls - is;
  else if (is == 0)
    is = right ? ls : 1;
  else if (is > ls)
    goto FOUND;

  is--;

  cp = PARAM[1]._string.addr + PARAM[1]._string.start;
  cs = PARAM->_string.addr + PARAM->_string.start + is;

  if (right)
  {
    for (; is >= 0; is--, cs--)
    {
      for (ip = 0; ip < lp; ip++)
      {
        if (cs[ip] != cp[ip])
          goto NEXT_R;
      }

      ret = is + 1;
      goto FOUND;

  NEXT_R:
    }
  }
  else
  {
    for (; is < ls; is++, cs++)
    {
      for (ip = 0; ip < lp; ip++)
      {
        if (cs[ip] != cp[ip])
          goto NEXT_L;
      }

      ret = is + 1;
      goto FOUND;

  NEXT_L:
    }
  }

FOUND:

  RELEASE_STRING(&PARAM[0]);
  RELEASE_STRING(&PARAM[1]);

  RETURN->type = T_INTEGER;
  RETURN->_integer.value = ret;

  SUBR_LEAVE();
}


PUBLIC void SUBR_like(void)
{
  char *pattern, *string;
  long len_pattern, len_string;
  boolean ret;

  SUBR_ENTER_PARAM(2);

  SUBR_get_string_len(&PARAM[0], &string, &len_string);
  SUBR_get_string_len(&PARAM[1], &pattern, &len_pattern);

  ret = REGEXP_match(pattern, len_pattern, string, len_string) ? -1 : 0;

  RELEASE_STRING(&PARAM[0]);
  RELEASE_STRING(&PARAM[1]);

  RETURN->type = T_BOOLEAN;
  RETURN->_boolean.value = ret;

  SUBR_LEAVE();
}


PRIVATE char *subst_buffer = NULL;

PRIVATE void subst_init(void)
{
  STRING_new(&subst_buffer, NULL, 0);
}

PRIVATE void subst_add(char *src, long len)
{
  long old_len = STRING_length(subst_buffer);

  STRING_extend(&subst_buffer, old_len + len);
  memcpy(&subst_buffer[old_len], src, len);
}


PRIVATE void subst_exit(void)
{
  subst_buffer = NULL;
}


PUBLIC void SUBR_subst(void)
{
  char *string;
  long len;
  char *add;
  long len_add;
  unsigned char c;
  int np;
  char *par;
  long len_par;

  unsigned char next_char(void) { len--; return (unsigned char)*string++; }

  SUBR_ENTER();

  subst_init();

  SUBR_get_string_len(&PARAM[0], &string, &len);

  if (len > 0)
  {
    add = string;
    len_add = 0;

    for(;;)
    {
      if (len == 0)
      {
        subst_add(add, len_add);
        break;
      }

      c = next_char();

      if (c != '&')
      {
        len_add++;
        continue;
      }

      subst_add(add, len_add);

      c = next_char();

      if (!isdigit(c))
      {
        len_add += 2;
        continue;
      }
      np = (c - '0');

      c = *string;

      if (isdigit(c))
      {
        np = np * 10 + c - '0';
        next_char();
      }

      if (np > 0 && np < NPARAM)
      {
        VALUE_to_string(&PARAM[np], &par, &len_par);
        subst_add(par, len_par);
      }

      add = string;
      len_add = 0;
    }
  }

  RELEASE_STRING(&PARAM[0]);

  RETURN->type = T_STRING;
  RETURN->_string.addr = subst_buffer;
  RETURN->_string.start = 0;
  RETURN->_string.len = STRING_length(subst_buffer);

  subst_exit();

  SUBR_LEAVE();
}

