/***************************************************************************

  trace.c

  The debugger

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __TRACE_C

#include "common.h"
#include <unistd.h>
#include <signal.h>

#include "error.h"
#include "type.h"
#include "debug.h"
#include "limit.h"
#include "common_buffer.h"
#include "stack.h"
#include "class.h"
#include "exec.h"
#include "local.h"
#include "Object.h"

#include "trace.h"

PUBLIC TRACE_GLOBAL TRACE = { 0 };

PRIVATE TRACE_BREAK *Breakpoint;
PRIVATE boolean Error;

PUBLIC void TRACE_break_on_next_line(void)
{
  TRACE.stop = TRUE;
  TRACE.fp = NULL;
  TRACE.bp = NULL;
}

PRIVATE void signal_user(int sig)
{
  signal(SIGUSR1, signal_user);

  /*printf("Process %d received SIGUSR1\n", getpid());
  fflush(NULL);*/

  TRACE.signaled = TRUE;
  if (EXEC_debug)
    TRACE_break_on_next_line();
}


PRIVATE boolean calc_line_from_position(CLASS *class, FUNCTION *func, PCODE *addr, ushort *line)
{
  int i;
  ushort pos = addr - func->code;
  ushort *post = func->debug->pos;

  if (func->debug)
  {
    for (i = 0; i < (func->debug->nline - 1); i++)
    {
      if (pos >= post[i] && pos < post[i + 1])
      {
        *line = i + func->debug->line;
        return FALSE;
      }
    }

    /*printf("pos = %d addr=%p func->code=%p\n", pos, addr, func->code);*/
  }

  return TRUE;
}


PRIVATE boolean calc_position_from_line(CLASS *class, ushort line, FUNCTION **function, PCODE **addr)
{
  int i;
  ushort pos, pos_after;
  FUNCTION *func;
  FUNC_DEBUG *debug;

  for (i = 0; i < class->load->n_func; i++)
  {
    func = &class->load->func[i];
    debug = func->debug;
    if (line >= debug->line && line < (debug->line + debug->nline))
      break;
  }

  if (i >= class->load->n_func)
    return TRUE;

  line -= debug->line;

  for(;;)
  {
    pos = debug->pos[line];
    pos_after = debug->pos[line + 1];
    if (pos != pos_after)
      break;

    line++;
    if (line >= debug->nline)
      return TRUE;
  }

  *function = func;
  *addr = &func->code[pos];

  /*printf("%s.%d -> %04X\n", class->name, line + debug->line, **addr);*/

  return FALSE;
}


PUBLIC void TRACE_init(void)
{
  ARRAY_create(&Breakpoint);

  /*
  TRACE.stop = FALSE;
  TRACE.fp = NULL;
  TRACE.bp = NULL;
  TRACE.signaled = FALSE;
  */

  signal(SIGUSR1, signal_user);
}

PUBLIC void TRACE_exit(void)
{
  ARRAY_delete(&Breakpoint);
}


PRIVATE int find_free_breakpoint(void)
{
  int i;
  char used[MAX_BREAKPOINT];

  memset(used, FALSE, MAX_BREAKPOINT);

  for (i = 0; i < ARRAY_count(Breakpoint); i++)
    used[Breakpoint[i].id - 1] = TRUE;

  for (i = 0; i < MAX_BREAKPOINT; i++)
    if (!used[i])
      return (i + 1);

  return 0;
}


PRIVATE boolean set_breakpoint(CLASS *class, ushort line)
{
  PCODE *addr;
  TRACE_BREAK *brk;
  int id;
  FUNCTION *func;

  if (ARRAY_count(Breakpoint) >= MAX_BREAKPOINT)
  {
    printf("Too many breakpoints (max = %d)\n", MAX_BREAKPOINT);
    return TRUE;
  }

  if (CLASS_is_native(class) || !class->debug)
  {
    printf("No debugging information\n");
    return TRUE;
  }

  if (calc_position_from_line(class, line, &func, &addr))
  {
    printf("Cannot calc position from line number\n");
    return TRUE;
  }

  if (!PCODE_is_breakpoint(*addr))
  {
    printf("Not a line beginning ?\n");
    return TRUE;
  }

  if (*addr & 0xFF)
  {
    printf("Breakpoint already set\n");
    return TRUE;
  }

  id = find_free_breakpoint();
  if (id == 0)
  {
    printf("Cannot create breakpoint\n");
    return TRUE;
  }

  brk = ARRAY_add(&Breakpoint);

  brk->id = id;
  /*brk->func = func;*/
  brk->addr = addr;
  brk->class = class;
  brk->line = line;

  *addr = PCODE_BREAKPOINT(id);

  return FALSE;
}


PRIVATE boolean unset_breakpoint(CLASS *class, ushort line)
{
  int i;

  for (i = 0; i < ARRAY_count(Breakpoint); i++)
  {
    if (Breakpoint[i].class == class && Breakpoint[i].line == line)
    {
      *(Breakpoint[i].addr) = PCODE_BREAKPOINT(0);
      ARRAY_remove(&Breakpoint, i);
      return FALSE;
    }
  }

  printf("Unknown breakpoint\n");
  return TRUE;
}


/*
PRIVATE boolean reset_breakpoint(int num)
{
  TRACE_BREAK *tb;

  if (num < 0 || num >= MAX_BREAKPOINT)
    return TRUE;

  tb = &Breakpoint[num];

  if (tb->addr == NULL)
    return TRUE;

  *(tb->addr) = tb->value;

  EXEC_stop_next = TRUE;
  Reset_breakpoint = num;

  return FALSE;
}


PUBLIC ushort TRACE_get_current_line(void)
{
  int i;
  ushort pos = PC - FP->code;

  if (FP->debug)
  {
    for (i = 0; i < (FP->debug->nline - 1); i++)
    {
      if (pos >= FP->debug->pos[i] && pos < FP->debug->pos[i + 1])
        return i + FP->debug->line;
    }
  }

  return 0;
}
*/


PRIVATE void command_quit(char *cmd)
{
  abort();
}

PRIVATE void command_go(char *cmd)
{
  TRACE.stop = FALSE;
}

PRIVATE void command_step(char *cmd)
{
  TRACE_break_on_next_line();
}

PRIVATE void command_next(char *cmd)
{
  TRACE.stop = TRUE;
  TRACE.fp = FP;
  TRACE.bp = BP;
}


PRIVATE void command_set_breakpoint(char *cmd)
{
  char class_name[64];
  CLASS *class;
  ushort line;

  if (sscanf(cmd, "+%64[^.].%hu", class_name, &line) != 2)
    printf("Syntax error\n");
  else
  {
    class = CLASS_get(class_name);
    if (class == NULL)
      printf("Unknown class\n");
    else
      set_breakpoint(class, line);
      /*printf("OK\n");*/
  }
}


PRIVATE void command_unset_breakpoint(char *cmd)
{
  char class_name[64];
  CLASS *class;
  ushort line;

  if (sscanf(cmd, "-%64[^.].%hu", class_name, &line) != 2)
    printf("Syntax error\n");
  else
  {
    class = CLASS_get(class_name);
    if (class == NULL)
      printf("Unknown class\n");
    else
      unset_breakpoint(class, line);
      /*printf("OK\n");*/
  }
}


PRIVATE void command_where(char *cmd)
{
  int i;
  STACK_CONTEXT *context;
  ushort line;

  if (CP)
    printf("%s\n", TRACE_get_current_position());

  for (i = 0; i < (STACK_frame_count - 1); i++)
  {
    context = &STACK_frame[i];

    if (context->pc)
    {
      line = 0;
      calc_line_from_position(context->cp, context->fp, context->pc, &line);
      printf("%s.%s.%d\n", context->cp->name, context->fp->debug->name, line);
    }
    else if (context->cp)
      printf("(Native code)\n");
  }
}


PRIVATE void to_string(VALUE *value, char **addr, long *len, boolean *more)
{
  static void *jump[16] = {
    &&_VOID, &&_BOOLEAN, &&_BYTE, &&_SHORT, &&_INTEGER, &&_LONG, &&_FLOAT, &&_DATE,
    &&_STRING, &&_STRING, &&_VARIANT, &&_ARRAY, &&_STRUCT, &&_FUNCTION, &&_CLASS, &&_NULL
    };

  *more = FALSE;

_conv:

  if (TYPE_is_object(value->type))
    goto _OBJECT;
  else
    goto *jump[value->type];

_NULL:

  *addr = "NULL";
  *len = 4;
  return;

_BOOLEAN:

  if (value->_boolean.value)
  {
    *addr = "TRUE";
    *len = 4;
  }
  else
  {
    *addr = "FALSE";
    *len = 5;
  }
  return;

_BYTE:
_SHORT:
_INTEGER:

  *len = sprintf(COMMON_buffer, "% ld", value->_integer.value);
  *addr = COMMON_buffer;

  return;

_DATE:

  LOCAL_format_date(DATE_split(value), LF_STANDARD, NULL, 0, addr, len);
  return;

_FLOAT:

  LOCAL_format_number(value->_float.value, LF_STANDARD, NULL, 0, addr, len, TRUE);
  return;

_STRING:

  *len = value->_string.len;
  *addr = value->_string.addr + value->_string.start;

  if( *len > 250)
    *len = sprintf(COMMON_buffer, "\"%.250s...\"", *addr);
  else
    *len = sprintf(COMMON_buffer, "\"%.*s\"", (int)*len, *addr);

  *addr = COMMON_buffer;
  return;

_OBJECT:

  if (VALUE_is_null(value))
    goto _NULL;

  *more = !CLASS_is_native(OBJECT_class(value->_object.object));

  *len = sprintf(COMMON_buffer, "%s %p", OBJECT_class(value->_object.object)->name, value->_object.object);
  *addr = COMMON_buffer;
  return;

_VARIANT:

  VARIANT_undo(value);
  goto _conv;

_VOID:

  *addr = "(void)";
  *len = 6;
  return;

_CLASS:

  {
    CLASS *class = value->_class.class;
    *more = (!CLASS_is_native(class) && class->load->n_stat > 0);

    *len = sprintf(COMMON_buffer, "%s %p", class->name, class);
    *addr = COMMON_buffer;
    return;
  }

_ARRAY:

  *len = sprintf(COMMON_buffer, "ARRAY %p", value->_array.addr);
  *addr = COMMON_buffer;
  return;

_LONG:
_STRUCT:
_FUNCTION:

  THROW(E_TYPE, TYPE_get_name(T_STRING), TYPE_get_name(value->type));
}


PRIVATE void print_value(char *name, int len, VALUE *value)
{
  char *pval;
  long lpval;
  boolean more;

  if (len == 0)
    len = strlen(name);

  to_string(value, &pval, &lpval, &more);
  printf("%c%.*s = %.*s\n", (more ? '+' : ' '), len, name, (int)lpval, pval);
}


PRIVATE void command_local(char *cmd)
{
  int i;
  VALUE value;
  int index;
  LOCAL_SYMBOL *lp;

  if (CP == NULL)
    return;

  if (OP)
  {
    value.type = T_OBJECT;
    value._object.class = CP;
    value._object.object = OP;
    print_value("ME", 0, &value);
  }
  else
  {
    value.type = T_CLASS;
    value._class.class = CP;
    print_value("CLASS", 0, &value);
  }

  if (FP == NULL)
    return;

  for (i = 0; i < FP->debug->n_local; i++)
  {
    lp = &FP->debug->local[i];
    index = lp->value;
    value = BP[index];
    print_value(lp->sym.name, lp->sym.len, &value);
  }
}


PRIVATE void command_print(char *cmd)
{
  OBJECT *object;
  CLASS *class;
  int i;
  VALUE value;
  GLOBAL_SYMBOL *gp;
  CLASS_VAR *var;
  char *addr;

  if (CP == NULL)
    return;

  if (sscanf(cmd, "?%p", &object) != 1)
  {
    printf("Syntax error\n");
    return;
  }

  if (object == NULL)
  {
    printf("Null pointer\n");
    return;
  }

  class = OBJECT_class(object);
  if (class == CLASS_Class)
  {
    class = (CLASS *)object;
    object = NULL;
  }
  else
  {
    value.type = T_CLASS;
    value._class.class = class;
    print_value("CLASS", 0, &value);
  }

  if (CLASS_is_native(class))
    return;

  if (object == NULL)
  {
    for (i = 0; i < class->load->n_global; i++)
    {
      gp = &class->load->global[i];

      if (CTYPE_get_kind(gp->ctype) != TK_VARIABLE)
        continue;

      if (!CTYPE_is_static(gp->ctype))
        continue;

      var = &class->load->stat[gp->value];
      addr = (char *)class->stat + var->pos;

      VALUE_class_read(class, &value, addr, var->type);
      print_value(gp->sym.name, gp->sym.len, &value);
    }
  }
  else
  {
    for (i = 0; i < class->load->n_global; i++)
    {
      gp = &class->load->global[i];

      if (CTYPE_get_kind(gp->ctype) != TK_VARIABLE)
        continue;

      if (CTYPE_is_static(gp->ctype))
        continue;

      /*if ((object == NULL) ^ (CTYPE_is_static(gp->ctype)))
        continue;*/

      if (!CTYPE_is_static(gp->ctype))
      {
        var = &class->load->dyn[gp->value];
        addr = (char *)object + var->pos;
      }
      else
      {
        var = &class->load->stat[gp->value];
        addr = (char *)class->stat + var->pos;
      }

      VALUE_class_read(class, &value, addr, var->type);
      print_value(gp->sym.name, gp->sym.len, &value);
    }
  }
}


PRIVATE void command_error(char *cmd)
{
  if (Error)
    ERROR_print_at(stdout);
}

PRIVATE void command_info(char *cmd)
{
  printf("w:\n");
  command_where(cmd);
  printf("l:\n");
  command_local(cmd);
  printf("e:\n");
  command_error(cmd);
}

PUBLIC void TRACE_main(boolean error)
{
  static TRACE_TYPE last_command = TC_NONE;

  static TRACE_COMMAND Command[] =
  {
    { "q", TC_NONE, command_quit, FALSE },
    { "n", TC_NEXT, command_next, FALSE },
    { "s", TC_STEP, command_step, FALSE },
    { "+", TC_NONE, command_set_breakpoint, TRUE },
    { "-", TC_NONE, command_unset_breakpoint, TRUE },
    { "g", TC_GO, command_go, FALSE },
    { "w", TC_NONE, command_where, TRUE },
    { "l", TC_NONE, command_local, TRUE },
    { "?", TC_NONE, command_print, TRUE },
    { "i", TC_NONE, command_info, TRUE },
    { "e", TC_NONE, command_error, TRUE },
    { NULL }
  };


  char cmd[256];
  int len;
  TRACE_COMMAND *tc;

  Error = error;

  do
  {
    printf("\n");
    fflush(NULL);

    if (CP == NULL)
      printf("[]:");
    else
      printf("[%s%s]:", TRACE_get_current_position(), Error ? "*" : "");

    if (EXEC_signal)
      kill(getppid(), SIGUSR1);

    fflush(NULL);

    *cmd = 0;
    fgets(cmd, sizeof(cmd), stdin);
    len = strlen(cmd);
    if (len > 0 && cmd[len - 1] == '\n')
      cmd[len - 1] = 0;

    /*printf("%s:\n", cmd);*/

    if (*cmd == 0)
    {
      if (last_command != TC_NONE)
      {
        for (tc = Command; tc->pattern; tc++)
        {
          if (tc->type == last_command)
          {
            (*tc->func)(cmd);
            break;
          }
        }
      }
    }
    else
    {
      for (tc = Command; tc->pattern; tc++)
      {
        if (strncmp(cmd, tc->pattern, strlen(tc->pattern)) == 0)
        {
          if (tc->type != TC_NONE)
            last_command = tc->type;
          (*tc->func)(cmd);
          break;
        }
      }
    }

    if (tc->pattern == NULL)
      printf("Unknown command\n");
  }
  while (tc->pattern == NULL || tc->loop);
}


PUBLIC void TRACE_breakpoint(int id)
{
  TRACE_main(FALSE);
}


PUBLIC char *TRACE_get_current_position(void)
{
#if DEBUG_ALLOC
  static char buffer[256];
#endif
  ushort line = 0;

  if (FP != NULL && FP->debug)
    calc_line_from_position(CP, FP, PC, &line);

#if DEBUG_ALLOC
  sprintf(buffer, "%s.%s.%d",
    CP ? CP->name : "?",
    (FP && FP->debug) ? FP->debug->name : "?",
    line);

  return buffer;
#else
  sprintf(COMMON_buffer, "%s.%s.%d",
    CP ? CP->name : "?",
    (FP && FP->debug) ? FP->debug->name : "?",
    line);

  return COMMON_buffer;
#endif
}


PUBLIC void TRACE_where(void)
{
  printf("%s: ", TRACE_get_current_position());
}


PUBLIC void TRACE_welcome(void)
{
  printf(TRACE_WELCOME);
}
