/***************************************************************************

  CDraw.cpp

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CDRAW_CPP

#include "common.h"

#include <qpainter.h>
#include <qpen.h>
#include <qbrush.h>

#include "gambas.h"

#include "CWidget.h"
#include "CWindow.h"
#include "CPicture.h"
#include "CDraw.h"

static CFONT draw_font;
static QPainter *DP = NULL;
static bool draw_on_window = false;

static void *CLASS_Window;
static void *CLASS_Picture;

static void end_painter(void)
{
  if (!DP)
    return;

  if (draw_on_window)
  {
    MyMainWindow *win = (MyMainWindow *)(DP->device());
    win->paintUnclip(false);
    draw_on_window = false;
  }

  delete DP;
  DP = NULL;
}


static void check_painter(void)
{
  if (!DP)
    GB.Error("No device");
}


BEGIN_METHOD_VOID(CDRAW_init)

  draw_font.klass = CLASS_Font;
  draw_font.font = new QFont;

  CLASS_Window = GB.FindClass("Window");
  CLASS_Picture = GB.FindClass("Picture");

END_METHOD


BEGIN_METHOD_VOID(CDRAW_exit)

  end_painter();

  delete draw_font.font;

END_METHOD


BEGIN_METHOD(CDRAW_begin, void *device)

  void *device = PARAM(device);

  end_painter();

  GB.CheckObject(device);

  if (GB.Is(device, CLASS_Window))
  {
    MyMainWindow *win = (MyMainWindow *)((CWIDGET *)device)->widget;
    win->paintUnclip(true);
    DP = new QPainter(win);
    draw_on_window = true;
  }
  else if (GB.Is(device, CLASS_Picture))
  {
    QPixmap *pix = ((CPICTURE *)device)->pixmap;
    DP = new QPainter(pix);
  }
  else
    GB.Error("Bad device");

END_METHOD


BEGIN_METHOD_VOID(CDRAW_end)

  end_painter();

END_METHOD


BEGIN_PROPERTY(CDRAW_background)

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnInteger(DP->backgroundColor().rgb() & 0xFFFFFF);
  else
    DP->setBackgroundColor(QColor(PROPERTY(long)));

END_PROPERTY


BEGIN_PROPERTY(CDRAW_transparent)

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnBoolean(DP->backgroundMode() == Qt::TransparentMode);
  else
    DP->setBackgroundMode(PROPERTY(char) ? Qt::TransparentMode : Qt::OpaqueMode);

END_PROPERTY


BEGIN_PROPERTY(CDRAW_invert)

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnBoolean(DP->rasterOp() == Qt::XorROP);
  else
    DP->setRasterOp(PROPERTY(char) ? Qt::XorROP : Qt::CopyROP);

END_PROPERTY


BEGIN_PROPERTY(CDRAW_foreground)

  QPen pen;

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnInteger(DP->pen().color().rgb() & 0xFFFFFF);
  else
  {
    pen = DP->pen();
    DP->setPen(QPen(QColor(PROPERTY(long)), pen.width(), pen.style()));
  }

END_PROPERTY


BEGIN_PROPERTY(CDRAW_line_width)

  QPen pen;

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnInteger(DP->pen().width());
  else
  {
    pen = DP->pen();
    DP->setPen(QPen(pen.color(), PROPERTY(long), pen.style()));
  }

END_PROPERTY


BEGIN_PROPERTY(CDRAW_line_style)

  QPen pen;

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnInteger(DP->pen().style());
  else
  {
    pen = DP->pen();
    DP->setPen(QPen(pen.color(), pen.width(), (Qt::PenStyle)PROPERTY(long)));
  }

END_PROPERTY


BEGIN_PROPERTY(CDRAW_fill_color)

  QBrush brush;

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnInteger(DP->brush().color().rgb() & 0xFFFFFF);
  else
  {
    brush = DP->brush();
    DP->setBrush(QBrush(QColor(PROPERTY(long)), brush.style()));
  }

END_PROPERTY


BEGIN_PROPERTY(CDRAW_fill_style)

  QBrush brush;

  check_painter();

  if (READ_PROPERTY)
    GB.ReturnInteger(DP->brush().style());
  else
  {
    brush = DP->brush();
    DP->setBrush(QBrush(brush.color(), (Qt::BrushStyle)PROPERTY(long)));
  }

END_PROPERTY


BEGIN_PROPERTY(CDRAW_font)

  CFONT *font;

  check_painter();

  if (READ_PROPERTY)
  {
    *(draw_font.font) = DP->font();
    GB.ReturnObject(&draw_font);
  }
  else
  {
    font = PROPERTY(CFONT *);

    if (font->font == 0)
      *(draw_font.font) = CFONT_TO_CWIDGET(font)->widget->font();
    else
      *(draw_font.font) = *(font->font);

    DP->setFont(*(draw_font.font));
  }

END_PROPERTY


BEGIN_METHOD(CDRAW_rect, long x; long y; long w; long h)

  check_painter();

  DP->drawRect(PARAM(x), PARAM(y), PARAM(w), PARAM(h));

END_METHOD


BEGIN_METHOD(CDRAW_ellipse, long x; long y; long w; long h; double start; double len)

  check_painter();

  if (GB.IsMissing(5) || GB.IsMissing(6))
    DP->drawEllipse(PARAM(x), PARAM(y), PARAM(w), PARAM(h));
  else
    DP->drawPie(PARAM(x), PARAM(y), PARAM(w), PARAM(h),
      (long)(PARAM(start) * 16 + 0.5), (long)(PARAM(len) * 16 + 0.5));

END_METHOD


BEGIN_METHOD(CDRAW_line, long x1; long y1; long x2; long y2)

  check_painter();

  DP->drawLine(PARAM(x1), PARAM(y1), PARAM(x2), PARAM(y2));

END_METHOD


BEGIN_METHOD(CDRAW_point, long x; long y)

  check_painter();

  DP->drawPoint(PARAM(x), PARAM(y));

END_METHOD


BEGIN_METHOD(CDRAW_picture, long x; long y; CPICTURE *pict; long sx; long sy; long sw; long sh)

  long sx, sy, sw, sh;

  check_painter();

  sx = GB.IsMissing(4) ? 0 : PARAM(sx);
  sy = GB.IsMissing(5) ? 0 : PARAM(sy);
  sw = GB.IsMissing(6) ? -1 : PARAM(sw);
  sh = GB.IsMissing(7) ? -1 : PARAM(sh);

  DP->drawPixmap(PARAM(x), PARAM(y), *(PARAM(pict)->pixmap), sx, sy, sw, sh);

END_METHOD


BEGIN_METHOD(CDRAW_text, long x; long y; GB_STRING text)

  check_painter();

  DP->drawText(PARAM(x), PARAM(y), GB.ToZeroString(PARAM(text)));

END_METHOD


BEGIN_METHOD(CDRAW_text_width, GB_STRING text)

  check_painter();

END_METHOD


BEGIN_METHOD(CDRAW_text_height, GB_STRING text)

  check_painter();

END_METHOD


GB_DESC CDrawDesc[] =
{
  GB_DECLARE("Draw", 0), GB_VIRTUAL_CLASS(),

  GB_STATIC_METHOD("_init", NULL, CDRAW_init, NULL),
  GB_STATIC_METHOD("_exit", NULL, CDRAW_exit, NULL),

  GB_STATIC_METHOD("Begin", NULL, CDRAW_begin, "(Device)o"),
  GB_STATIC_METHOD("End", NULL, CDRAW_end, NULL),

  //GB_STATIC_METHOD("Clip", NULL, CDRAW_clip, "(X)i(Y)i(Width)i(Height)i"),

  GB_STATIC_PROPERTY("BackColor", "i", CDRAW_background),
  GB_STATIC_PROPERTY("Background", "i", CDRAW_background),
  GB_STATIC_PROPERTY("Transparent", "b", CDRAW_transparent),
  GB_STATIC_PROPERTY("Invert", "b", CDRAW_invert),

  GB_STATIC_PROPERTY("ForeColor", "i", CDRAW_foreground),
  GB_STATIC_PROPERTY("Foreground", "i", CDRAW_foreground),

  GB_STATIC_PROPERTY("LineWidth", "i", CDRAW_line_width),
  GB_STATIC_PROPERTY("LineStyle", "i", CDRAW_line_style),

  GB_STATIC_PROPERTY("FillColor", "i", CDRAW_fill_color),
  GB_STATIC_PROPERTY("FillStyle", "i", CDRAW_fill_style),

  GB_STATIC_PROPERTY("Font", "Font", CDRAW_font),

  //GB_STATIC_METHOD("Arc", NULL, CDRAW_arc, "(X)i(Y)i(Width)i(Height)i(Start)e(Length)e"),
  //GB_STATIC_METHOD("Chord", NULL, CDRAW_chord, "(X)i(Y)i(Width)i(Height)i(Start)e(Length)e"),

  //GB_SATIC_METHOD("Circle", NULL, CDRAW_circle, "(X)i(Y)i(Radius)i[(Start)e(Length)e]"),
  GB_STATIC_METHOD("Ellipse", NULL, CDRAW_ellipse, "(X)i(Y)i(Width)i(Height)i[(Start)f(Length)f]"),
  GB_STATIC_METHOD("Line", NULL, CDRAW_line, "(X1)i(Y1)i(X2)i(Y2)i"),
  GB_STATIC_METHOD("Picture", NULL, CDRAW_picture, "(X)i(Y)i(Picture)Picture;[(SrcX)i(SrcY)i(SrcWidth)i(SrcHeight)i]"),
  GB_STATIC_METHOD("Point", NULL, CDRAW_point, "(X)i(Y)i"),
  GB_STATIC_METHOD("Rect", NULL, CDRAW_rect, "(X)i(Y)i(Width)i(Height)i"),
  GB_STATIC_METHOD("Text", NULL, CDRAW_text, "(X)i(Y)i(Text)s"),
  //GB_STATIC_METHOD("RectText", NULL, CDRAW_rect_text, "(X)i(Y)i(Width)i(Height)i(Text)s[(Alignment)i]"),

  GB_METHOD("TextWidth", "i", CDRAW_text_width, "(Text)s"),
  GB_METHOD("TextHeight", "i", CDRAW_text_height, "(Text)s"),

  GB_END_DECLARE
};


