/***************************************************************************

  CTextBox.cpp

  The TextBox class

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/


#define __CTEXTBOX_CPP

#include "common.h"

#include <qapplication.h>
#include <qlineedit.h>
#include <qcombobox.h>
#include <qlistbox.h>

#include "gambas.h"

#include "CTextBox.h"


static long EVENT_Change;
static long EVENT_Activate;
static long EVENT_Click;

#define MAX_LEN 32767


/***************************************************************************

  TextBox

***************************************************************************/

PRIVATE bool get(void *_object, QLineEdit **wid, bool error = true)
{
  QComboBox *combo;

  if (TEXTBOX->isA("QLineEdit"))
  {
    *wid = TEXTBOX;
    return false;
  }

  combo = COMBOBOX;
  if (!combo->editable())
  {
    if (error)
      GB.Error("ComboBox is read-only");
    return true;
  }

  *wid = combo->lineEdit();
  return false;
}

#define GET_TEXT_BOX() \
  QLineEdit *textbox; \
  if (get(_object, &textbox)) \
    return;



BEGIN_METHOD(CTEXTBOX_new, void *parent)

  QLineEdit *wid = new QLineEdit(CONTAINER(PARAM(parent)));

  CWIDGET_new(wid, (void *)_object, "TextBox");

  QObject::connect(wid, SIGNAL(textChanged(const QString &)), &CTextBox::manager, SLOT(event_change()));
  QObject::connect(wid, SIGNAL(returnPressed()), &CTextBox::manager, SLOT(event_activate()));

  OBJECT(CTEXTBOX)->widget.background = QColorGroup::Base;

  wid->setAlignment(Qt::AlignVCenter | Qt::AlignLeft);
  wid->show();

END_METHOD


BEGIN_METHOD_VOID(CTEXTBOX_clear)

  TEXTBOX->clear();

END_METHOD


BEGIN_METHOD(CTEXTBOX_insert, GB_STRING text)

  GET_TEXT_BOX();

  textbox->insert(QString(GB.ToZeroString(PARAM(text))));

END_METHOD


BEGIN_PROPERTY(CTEXTBOX_text)

  if (READ_PROPERTY)
    GB.ReturnCopyString(TEXTBOX->text(), 0);
  else
    TEXTBOX->setText(GB.ToZeroString(PROPERTY(GB_STRING)));

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_length)

  GB.ReturnInteger(TEXTBOX->text().length());

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_alignment)

  if (READ_PROPERTY)
    GB.ReturnInteger(TEXTBOX->alignment() | Qt::AlignVCenter);
  else
    TEXTBOX->setAlignment(PROPERTY(long) & 0x7);

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_pos)

  GET_TEXT_BOX();

  if (READ_PROPERTY)
    GB.ReturnInteger(textbox->cursorPosition());
  else
    textbox->setCursorPosition(PROPERTY(long));

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_read_only)

  if (READ_PROPERTY)
    GB.ReturnBoolean(TEXTBOX->isReadOnly());
  else
    TEXTBOX->setReadOnly(PROPERTY(char) != 0);

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_border)

  GET_TEXT_BOX();

  if (READ_PROPERTY)
    GB.ReturnBoolean(textbox->frame());
  else
    textbox->setFrame(PROPERTY(char) != 0);

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_password)

  GET_TEXT_BOX();

  if (READ_PROPERTY)
    GB.ReturnBoolean(textbox->echoMode() != QLineEdit::Normal);
  else
    textbox->setEchoMode(PROPERTY(char) != 0 ? QLineEdit::Password : QLineEdit::Normal);

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_max_length)

  long max;

  GET_TEXT_BOX();

  if (READ_PROPERTY)
  {
    max = textbox->maxLength();
    GB.ReturnInteger(max >= MAX_LEN ? 0 : max);
  }
  else
  {
    max = PROPERTY(long);
    if (max < 1 || max > MAX_LEN)
      max = MAX_LEN;

    textbox->setMaxLength(max);
  }

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_selection)

  RETURN_SELF();

END_PROPERTY


/***************************************************************************

  .TextBox.Selection

***************************************************************************/

BEGIN_PROPERTY(CTEXTBOX_sel_text)

  GET_TEXT_BOX();

  if (READ_PROPERTY)
    GB.ReturnString(textbox->markedText(), 0);
  else
    textbox->insert(GB.ToZeroString(PROPERTY(GB_STRING)));

END_PROPERTY


PRIVATE void set_selection(QLineEdit *textbox, int start, int length)
{
  if (start < 0 || start >= (int)textbox->text().length())
  {
    start = textbox->cursorPosition();
    length = 0;
  }

  textbox->setCursorPosition(start);

  if (length <= 0)
    textbox->deselect();
  else
    textbox->setSelection(start, length);
}

PRIVATE void get_selection(QLineEdit *textbox, int *start, int *length)
{
  *start = textbox->cursorPosition();
  if (!textbox->hasMarkedText())
    *length = 0;
  else
    *length = textbox->markedText().length();
}


BEGIN_PROPERTY(CTEXTBOX_sel_length)

  int start, length;

  GET_TEXT_BOX();

  get_selection(textbox, &start, &length);

  GB.ReturnInteger(length);

END_PROPERTY


BEGIN_PROPERTY(CTEXTBOX_sel_start)

  int start, length;

  GET_TEXT_BOX();

  get_selection(textbox, &start, &length);
  GB.ReturnInteger(start);

END_PROPERTY


BEGIN_METHOD_VOID(CTEXTBOX_sel_clear)

  GET_TEXT_BOX();

  textbox->deselect();

END_METHOD


BEGIN_METHOD(CTEXTBOX_sel_select, long start; long length)

  GET_TEXT_BOX();

  set_selection(textbox, PARAM(start), PARAM(length));

END_METHOD


BEGIN_METHOD_VOID(CTEXTBOX_sel_all)

  GET_TEXT_BOX();

  textbox->selectAll();

END_METHOD


/***************************************************************************

  ComboBox

***************************************************************************/


BEGIN_METHOD(CCOMBOBOX_new, void *parent)

  QComboBox *wid = new QComboBox(TRUE, CONTAINER(PARAM(parent)));

  QObject::connect(wid, SIGNAL(textChanged(const QString &)), &CTextBox::manager, SLOT(event_change()));
  QObject::connect(wid, SIGNAL(activated(int)), &CTextBox::manager, SLOT(event_click()));

  CWIDGET_new(wid, (void *)_object, "ComboBox");

  OBJECT(CCOMBOBOX)->widget.background = QColorGroup::Base;

  wid->setInsertionPolicy(QComboBox::NoInsertion);
  wid->show();

END_METHOD


BEGIN_METHOD_VOID(CCOMBOBOX_clear)

  COMBOBOX->clear();
  COMBOBOX->clearEdit();

END_METHOD


BEGIN_PROPERTY(CCOMBOBOX_text)

  if (READ_PROPERTY)
    GB.ReturnString(COMBOBOX->currentText(), 0);
  else
  {
    QString text(GB.ToZeroString(PROPERTY(GB_STRING)));

    if (COMBOBOX->editable())
      COMBOBOX->lineEdit()->setText(text);

    // A faire lorsque la combo s'ouvre

    for (int i = 0; i < COMBOBOX->count(); i++)
    {
      if (COMBOBOX->text(i) == text)
      {
        COMBOBOX->setCurrentItem(i);
        return;
      }
    }

    // Il semble que ceci vide le texte de la combo !

    COMBOBOX->setCurrentItem(-1);
  }

END_PROPERTY


BEGIN_PROPERTY(CCOMBOBOX_length)

  GB.ReturnInteger(COMBOBOX->currentText().length());

END_PROPERTY


BEGIN_PROPERTY(CCOMBOBOX_read_only)

  if (READ_PROPERTY)
    GB.ReturnBoolean(!COMBOBOX->editable());
  else
  {
    if (PROPERTY(char) == 0)
    {
      if (!COMBOBOX->editable())
      {
        COMBOBOX->setEditable(true);
        QObject::connect(COMBOBOX->lineEdit(), SIGNAL(returnPressed()), &CTextBox::manager, SLOT(event_activate()));
      }
    }
    else
    {
      COMBOBOX->setEditable(false);
    }

  }

END_PROPERTY


BEGIN_METHOD(CCOMBOBOX_get, long index)

  GB.ReturnString(COMBOBOX->text(PARAM(index)), 0);

END_METHOD


BEGIN_METHOD(CCOMBOBOX_set, long index; GB_STRING item)

  COMBOBOX->changeItem(QString(GB.ToZeroString(PARAM(item))), PARAM(index));

END_METHOD


BEGIN_METHOD(CCOMBOBOX_add, GB_STRING item; long pos)

  COMBOBOX->insertItem(QString(GB.ToZeroString(PARAM(item))), GB.IsMissing(2) ? -1 : PARAM(pos));

END_METHOD


BEGIN_METHOD(CCOMBOBOX_remove, long pos)

  COMBOBOX->removeItem(PARAM(pos));

END_METHOD


BEGIN_METHOD_VOID(CCOMBOBOX_sort)

  COMBOBOX->listBox()->sort();

END_METHOD


BEGIN_PROPERTY(CCOMBOBOX_count)

  GB.ReturnInteger(COMBOBOX->count());

END_PROPERTY


BEGIN_PROPERTY(CCOMBOBOX_index)

  if (READ_PROPERTY)
    GB.ReturnInteger(COMBOBOX->currentItem());
  else
    COMBOBOX->setCurrentItem(PROPERTY(long));

END_PROPERTY


BEGIN_PROPERTY(CCOMBOBOX_current)

  int current = COMBOBOX->currentItem();

  if (current < 0)
    GB.ReturnNull();
  else
    GB.ReturnString(COMBOBOX->text(current), 0);

END_PROPERTY


BEGIN_PROPERTY(CCOMBOBOX_mouse)

  if (READ_PROPERTY)
    GB.ReturnInteger(COMBOBOX->cursor().shape());
  else
  {
    if (COMBOBOX->editable())
      COMBOBOX->lineEdit()->setCursor(PROPERTY(long));

    COMBOBOX->setCursor(PROPERTY(long));
  }

END_METHOD






/***************************************************************************

  class CTextBox

***************************************************************************/

CTextBox CTextBox::manager;

void CTextBox::event_change(void)
{
  RAISE_EVENT(EVENT_Change, NULL);
}


void CTextBox::event_activate(void)
{
  RAISE_EVENT(EVENT_Activate, NULL);
}


void CTextBox::event_click()
{
  RAISE_EVENT(EVENT_Click, NULL);
}


/***************************************************************************

  Descriptions

***************************************************************************/

GB_DESC CTextBoxSelectionDesc[] =
{
  GB_DECLARE(".TextBox.Selection", 0), GB_VIRTUAL_CLASS(),

  GB_PROPERTY("Text", "s", CTEXTBOX_sel_text),
  GB_PROPERTY_READ("Length", "i", CTEXTBOX_sel_length),
  GB_PROPERTY_READ("Start", "i", CTEXTBOX_sel_start),

  GB_METHOD("Clear", NULL, CTEXTBOX_sel_clear, NULL),
  GB_METHOD("Select", NULL, CTEXTBOX_sel_select, "(Start)i(Length)i"),
  GB_METHOD("SelectAll", NULL, CTEXTBOX_sel_all, NULL),

  GB_END_DECLARE
};


GB_DESC CTextBoxDesc[] =
{
  GB_DECLARE("TextBox", sizeof(CTEXTBOX)), GB_INHERITS("Control"),

  GB_METHOD("_new", NULL, CTEXTBOX_new, "(Parent)Container;"),

  GB_PROPERTY("Text", "s", CTEXTBOX_text),
  GB_PROPERTY("Alignment", "i<Align,Left,Center,Right>", CTEXTBOX_alignment),
  GB_PROPERTY_READ("Length", "i", CTEXTBOX_length),
  GB_PROPERTY("Pos", "i", CTEXTBOX_pos),
  GB_PROPERTY("ReadOnly", "b", CTEXTBOX_read_only),
  GB_PROPERTY("Border", "b", CTEXTBOX_border),
  GB_PROPERTY("Password", "b", CTEXTBOX_password),
  GB_PROPERTY("MaxLength", "i", CTEXTBOX_max_length),

  GB_PROPERTY_READ("Selection", ".TextBox.Selection", CTEXTBOX_selection),

  GB_CONSTANT("_Properties", "s", CTEXTBOX_PROPERTIES),
  GB_CONSTANT("_DefaultEvent", "s", "KeyPress"),

  GB_METHOD("Clear", NULL, CTEXTBOX_clear, NULL),
  GB_METHOD("Insert", NULL, CTEXTBOX_insert, "(Text)s"),

  GB_EVENT("Change", NULL, NULL, &EVENT_Change),
  GB_EVENT("Activate", NULL, NULL, &EVENT_Activate),

  GB_END_DECLARE
};


GB_DESC CComboBoxDesc[] =
{
  GB_DECLARE("ComboBox", sizeof(CCOMBOBOX)), GB_INHERITS("Control"),

  GB_METHOD("_new", NULL, CCOMBOBOX_new, "(Parent)Container;"),
  GB_METHOD("_get", "s", CCOMBOBOX_get, "(Index)i"),
  GB_METHOD("_put", NULL, CCOMBOBOX_set, "(Index)i(Item)s"),

  //

  GB_PROPERTY("Text", "s", CCOMBOBOX_text),
  GB_PROPERTY_READ("Length", "i", CCOMBOBOX_length),
  GB_PROPERTY("Pos", "i", CTEXTBOX_pos),
  GB_PROPERTY("ReadOnly", "b", CCOMBOBOX_read_only),
  GB_PROPERTY("Password", "b", CTEXTBOX_password),
  GB_PROPERTY("MaxLength", "i", CTEXTBOX_max_length),

  GB_PROPERTY_READ("Selection", ".TextBox.Selection", CTEXTBOX_selection),

  GB_CONSTANT("_Properties", "s", CCOMBOBOX_PROPERTIES),
  GB_CONSTANT("_DefaultEvent", "s", "Click"),

  GB_METHOD("Clear", NULL, CCOMBOBOX_clear, NULL),
  GB_METHOD("Insert", NULL, CTEXTBOX_insert, "(Text)s"),
  GB_METHOD("SelectAll", NULL, CTEXTBOX_sel_all, NULL),

  GB_EVENT("Change", NULL, NULL, &EVENT_Change),
  GB_EVENT("Activate", NULL, NULL, &EVENT_Activate),

  //

  GB_METHOD("Add", NULL, CCOMBOBOX_add, "(Item)s[(Index)i]"),
  GB_METHOD("Sort", NULL, CCOMBOBOX_sort, NULL),
  GB_METHOD("Remove", NULL, CCOMBOBOX_remove, "(Index)i"),

  /*
  GB_METHOD("Find", "i", CCOMBOBOX_find, 1, 1, "s"),
  GB_METHOD("Select", NULL, CCOMBOBOX_select, 1, 1, "i"),
  GB_METHOD("Invert", NULL, CCOMBOBOX_invert, 0, 0, NULL),
  */

  GB_PROPERTY_READ("Count", "i", CCOMBOBOX_count),
  GB_PROPERTY_READ("Current", "s", CCOMBOBOX_current),
  GB_PROPERTY("Index", "i", CCOMBOBOX_index),

  GB_PROPERTY("Mouse", "i", CCOMBOBOX_mouse),

  GB_EVENT("Click", NULL, NULL, &EVENT_Click),

  GB_END_DECLARE
};


