/***************************************************************************

  gambas.h

  Copyright (C) 2000 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

/*

  Special methods :

  _get      array read operator on object
  _put      array write operator on object
  _new      constructor
  _free     destructor
  _first    starting an enumeration
  _next     next iteration of enumeration

  Method signature :

  BOOLEAN     b
  BYTE        c
  SHORT       h
  INTEGER     i
  LONG        l
  FLOAT       f
  DATE        d
  STRING      s
  VARIANT     v
  OBJECT      o
  CLASS       Class;

*/

#ifndef __GAMBAS_H
#define __GAMBAS_H

/* Version */

#define GB_VERSION  1


/* Symbol Types */

#define GB_PROPERTY_ID          'p'
#define GB_METHOD_ID            'm'
#define GB_CONSTANT_ID          'C'
#define GB_EVENT_ID             ':'
#define GB_ENUM_ID              '#'
#define GB_STATIC_PROPERTY_ID   'P'
#define GB_STATIC_METHOD_ID     'M'

#define GB_HOOK_NEW_ID          ((char *)1)
#define GB_HOOK_FREE_ID         ((char *)2)
#define GB_HOOK_REF_ID          ((char *)3)
#define GB_HOOK_UNREF_ID        ((char *)4)
#define GB_HOOK_CHECK_ID        ((char *)5)
#define GB_INHERITS_ID          ((char *)16)


#ifndef __GAMBAS_INSIDE

/* Datatypes for return values and variants */

#define GB_T_VOID         0
#define GB_T_BOOLEAN      1
#define GB_T_BYTE         2
#define GB_T_SHORT        3
#define GB_T_INTEGER      4
#define GB_T_LONG         5
#define GB_T_FLOAT        6
#define GB_T_DATE         7
#define GB_T_STRING       8
#define GB_T_VARIANT      10
#define GB_T_NULL         15
#define GB_T_OBJECT       16


/* Constants for Stat() function */

#define GB_STAT_FILE        0
#define GB_STAT_DIRECTORY   1
#define GB_STAT_DEVICE      2
#define GB_STAT_PIPE        3
#define GB_STAT_SOCKET      4
#define GB_STAT_LINK        5


/* Constants for Message() function */

#define GB_MSG_OK           1
#define GB_MSG_CANCEL       2
#define GB_MSG_YES          3
#define GB_MSG_NO           4
#define GB_MSG_ABORT        5
#define GB_MSG_RETRY        6
#define GB_MSG_IGNORE       7

#define GB_MSG_USER_DEFINED         0
#define GB_MSG_OK_ONLY              1
#define GB_MSG_OK_CANCEL            2
#define GB_MSG_YES_NO               3
#define GB_MSG_YES_NO_CANCEL        4
#define GB_MSG_RETRY_IGNORE_CANCEL  5
#define GB_MSG_RETRY_ABORT          6
#define GB_MSG_RETRY_IGNORE_ABORT   7

#define GB_MSG_INFORMATION  1
#define GB_MSG_WARNING      2
#define GB_MSG_CRITICAL     3
#define GB_MSG_QUESTION     4


/* Structure passed to GB_Message() */

typedef
  struct {
    char *text;
    char *title;
    short icon;
    short type;
    char *button[3];
    short bydefault;
    }
  GAMBAS_MESSAGE;


/* Variant datatype */

typedef
  union {
    long type;
    struct { long type; long value; } _boolean;
    struct { long type; long value; } _byte;
    struct { long type; long value; } _short;
    struct { long type; long value; } _integer;
    struct { long type; long long value; } _long;
    struct { long type; double value; } _float;
    struct { long type; long long value; } _date;
    struct { long type; char *value; } _string;
    struct { long type; void *value; } _object;
    }
  GB_VARIANT;

/* String Datatype */

typedef
  struct {
    char *addr;
    long len;
    }
  GB_STRING;



/* Predefined errors */

#define GB_ERR_TYPE   ((char *)6)
#define GB_ERR_ARG    ((char *)20)


/* Declare macros */

#define GB_DECLARE(name, size) \
  { name, (long)GB_VERSION, (long)size }

#define GB_END_DECLARE  { 0 }


/* Hook macros */

#define GB_HOOK_NEW(hook)    { GB_HOOK_NEW_ID, (long)hook }
#define GB_HOOK_FREE(hook)   { GB_HOOK_FREE_ID, (long)hook }
#define GB_HOOK_REF(hook)    { GB_HOOK_REF_ID, (long)hook }
#define GB_HOOK_UNREF(hook)  { GB_HOOK_UNREF_ID, (long)hook }
#define GB_HOOK_CHECK(hook)  { GB_HOOK_CHECK_ID, (long)hook }

#define GB_VIRTUAL_CLASS()  \
  GB_HOOK_NEW(NULL), GB_HOOK_FREE(NULL), GB_HOOK_REF(NULL), GB_HOOK_UNREF(NULL)

/* Symbol macros */

#define GB_CONSTANT(symbol, type, value) \
  { "C" symbol, (long)type, (long)value }

#define GB_PROPERTY(symbol, type, proc) \
  { "p" symbol, (long)type, (long)proc }

#define GB_PROPERTY_READ(symbol, type, proc) \
  { "r" symbol, (long)type, (long)proc }

/*#define GB_PROPERTY_WRITE(symbol, type, proc) \
  { "p" symbol, (long)type, (long)0, (long)proc }*/

#define GB_METHOD(symbol, type, exec, signature) \
  { "m" symbol, (long)type, (long)signature, (long)(exec) }

#define GB_EVENT(symbol, type, signature, id) \
  { "::" symbol, (long)type, (long)signature, (long)id }

#define GB_STATIC_PROPERTY(symbol, type, proc) \
  { "P" symbol, (long)type, (long)proc }

#define GB_STATIC_PROPERTY_READ(symbol, type, proc) \
  { "R" symbol, (long)type, (long)proc }

/*#define GB_STATIC_PROPERTY_WRITE(symbol, type, proc) \
  { "P" symbol, (long)type, (long)0, (long)proc }*/

#define GB_STATIC_METHOD(symbol, type, exec, signature) \
  { "M" symbol, (long)type, (long)signature, (long)(exec) }

#define GB_INHERITS(symbol) \
  { GB_INHERITS_ID, (long)symbol }


/* Method declaration */

#define BEGIN_METHOD(_name, par) \
typedef \
  struct { \
    par; \
    } \
    _##_name; \
\
void _name(void *_object, void *_param) \
{ \
 _##_name *_p = (_##_name *)_param; \
 /*printf( #_name " < { %p }\n" , ((GAMBAS_OBJECT *)DEBUG_addr)->klass);*/

#define PARAM(par) ((_p)->par)

#define BEGIN_METHOD_VOID(_name) \
void _name(void *_object, void *_param) { \
  /*printf( #_name " < { %p }\n" , ((GAMBAS_OBJECT *)DEBUG_addr)->klass);*/

#define END_METHOD }

#define CALL_METHOD_VOID(_name) _name(_object, NULL)

/* Property declaration */

#define BEGIN_PROPERTY(_name) \
void _name(void *_object, void *_param) { \
  /*printf( #_name " < { %p }\n" , ((GAMBAS_OBJECT *)DEBUG_addr)->klass);*/

#define END_PROPERTY }

#define READ_PROPERTY  (_param == NULL)

#define PROPERTY(_type) (*((_type *)_param))


/* Common macros */

#define OBJECT(type) ((type *)_object)
#define RETURN_SELF()  GB.ReturnObject(_object)

#endif /* __GAMBAS_INSIDE */


/* declare events */

#define DECLARE_EVENT(_event) static long _event


/* hooks */

typedef
  enum {
    GB_HOOK_MAIN = 1,
    GB_HOOK_LOOP = 2,
    GB_HOOK_WAIT = 3,
    GB_HOOK_MESSAGE = 4,
    GB_HOOK_BEEP = 5,
    GB_HOOK_WATCH = 6
    }
  GB_HOOK;

/* watch callbacks */

enum {
  GB_WATCH_NONE = 0,
  GB_WATCH_READ = 1,
  GB_WATCH_WRITE = 2
  };

typedef
  void (*GB_WATCH_CALLBACK)(int, int, long);

/* gambas object */

typedef
  struct {
    long klass;
    long ref;
    }
  GB_OBJECT;

/* gambas method */

typedef
  struct {
    void *_reserved1;
    void *_reserved2;
    }
  GB_FUNCTION;

/* description */

typedef
  struct {
    char *name;
    long val1;
    long val2;
    long val3;
    long val4;
    long val5;
    }
  GB_DESC;

typedef
  void GB_METHOD_FUNC(void *, void *);

typedef
  void GB_PROPERTY_FUNC(void *, void *);

#define DECLARE_METHOD(_method) GB_METHOD_FUNC _method
#define DECLARE_PROPERTY(_property) GB_PROPERTY_FUNC _property

/* gambas interface */

typedef
  struct {
    long version;

    int (*GetInterface)(char *, long, void *);

    void (*Declare)(void *, ...);
    void *(*Hook)(int, void *);

    int (*GetFunction)(GB_FUNCTION *, void *, char *, char *);
    void (*Call)(GB_FUNCTION *, int, ...);
    int (*Raise)(void *, int, void *);
    int (*CanRaise)(void *, int);

    void (*Error)(char *, ...);

    void *(*GetClass)(void *);
    void *(*FindClass)(char *);
    int (*Is)(void *, void *);
    void (*Ref)(void *);
    void (*Unref)(void **);
    void (*UnrefKeep)(void **, int);
    void (*Detach)(void *);
    void (*Attach)(void *, void *, char *);
    int (*New)(void **, char *, char *, void *);
    void (*CheckObject)(void *);

    void *(*GetEnum)(void);
    void (*StopEnum)(void);

    void (*Return)(unsigned long, ...);
    void (*ReturnInteger)(long);
    void (*ReturnBoolean)(int);
    void (*ReturnDate)(long);
    void (*ReturnObject)(void *);
    void (*ReturnNull)(void);
    void (*ReturnPtr)(unsigned long, void *);

    void (*ReturnString)(const char *, long);
    void (*ReturnZeroString)(const char *);
    void (*ReturnCopyString)(const char *, long);
    void (*ReturnNewString)(char *);

    char *(*NewString)(long);
    char *(*CopyString)(char *, long);
    void (*FreeString)(char *);
    char *(*ToZeroString)(GB_STRING);

    int (*IsMissing)(int);
    unsigned long (*GetReturnType)(void);

    void (*LoadFile)(char *, char **, long *);
    void (*ReleaseFile)(char **, long);

    void (*Store)(long, void *, void *);
    void (*StoreString)(GB_STRING, char **);
    void (*StoreObject)(void *, void **);

    void (*Watch)(int, int, void *, long);
    }
  GB_INTERFACE;


/*  #if defined(__cplusplus)*/

#endif
